-----------------------------------------------------------------------------------------------
-- Client Lua Script for SteadyPlates
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------

-----------------------------------------------------------------------------------------------
-- Copyright 2014, 2015 (c) Jos_eu. All rights reserved
-- This file is part of SteadyPlates.
--
-- SteadyPlates is free software: you can redistribute it and/or modify it under the terms of 
-- the GNU General Public License as published by the Free Software Foundation, either version 3
-- of the License, or (at your option) any later version.
--
-- SteadyPlates is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
-- without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License along with Foobar. 
-- If not, see http://www.gnu.org/licenses/
-----------------------------------------------------------------------------------------------

require "Window"
require "GameLib"
require "GroupLib"
require "Apollo"

local type, tostring, string = type, tostring, string
local ipairs, pairs, select = ipairs, pairs, select

local Dev = Apollo.GetPackage("Jos_eu:DevTools-1.3").tPackage
local t2s, o2s, mt2s, explode, PrintObj, SPrint, CPrint = Dev.t2s, Dev.o2s, Dev.mt2s, Dev.explode, Dev.PrintObj, Dev.SPrint, Dev.CPrint

-----------------------------------------------------------------------------------------------
-- SteadyPlates Module Definition
-----------------------------------------------------------------------------------------------
local SteadyPlates = {} 
local SteadyPlatesInst = nil

local fNameplatesHelperVerifyVisibilityOptions = nil
local fNameplatesUpdateNameplateRewardInfo = nil
local fNameplatesOnUnitOcclusionChanged = nil
local fNameplatesOnUnitCreated = nil
local fNameplatesOnUnitDestroyed = nil
local fCheckDrawDistance = nil
local fOnTargetUnitChanged = nil
local fDrawHealth = nil
local fDrawCastBar = nil
local fDrawHealthShieldBar = nil
local fDrawName = nil
local fDrawGuild = nil
local fOnFrame = nil
local fHelperCalculateConValue = nil
local fOnWorldLocationOnScreen = nil
local fUpdateNameplateVisibility = nil
local fColorNameplate = nil
local fOnPreloadUnitCreateTimer = nil

local aNameplates = nil
local aPerspectivePlates = nil
local bPerspectivePlates = false
local aLibJSON = nil
local aGeminiColor = nil

local uPlayerUnit = nil
local bPlayerInCombat = false
local nUIScale = 1

-- locally cache function references to be used in OnFameHook for loop
local fnDrawHealth = nil
local fnDrawCastBar = nil
local fnDrawTargeting = nil
local fnDrawThreat = nilt
local fnDrawInterruptArmor = nil
local fnDrawCleanse = nil
local fnDrawCluster = nil
local fnHelperChangeHealthText = nil
local fnHelperEnsureNameplateOnScreen = nil
local fnDrawCertainDeath = nil
local fnDrawRewards = nil
local fnColorNameplate = nil
local fnDrawVulnerable = nil
local fnHelperCalculateConValue = nil
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
-- e.g. local kiExampleVariableMax = 999
--local kcrSelectedText = ApolloColor.new("UI_BtnTextHoloPressedFlyby")
--local kcrNormalText = ApolloColor.new("UI_BtnTextHoloNormal")
 
local karDisposition = 
{
	tTextColors =
	{
		[Unit.CodeEnumDisposition.Hostile] 	= ApolloColor.new("DispositionHostile"),
		[Unit.CodeEnumDisposition.Neutral] 	= ApolloColor.new("DispositionNeutral"),
		[Unit.CodeEnumDisposition.Friendly] = ApolloColor.new("DispositionFriendly"),
	},

	tTargetPrimary =
	{
		[Unit.CodeEnumDisposition.Hostile] 	= "CRB_Nameplates:sprNP_BaseSelectedRed",
		[Unit.CodeEnumDisposition.Neutral] 	= "CRB_Nameplates:sprNP_BaseSelectedYellow",
		[Unit.CodeEnumDisposition.Friendly] = "CRB_Nameplates:sprNP_BaseSelectedGreen",
	},

	tTargetSecondary =
	{
		[Unit.CodeEnumDisposition.Hostile] 	= "CRB_Nameplates:sprNp_Target_HostileSecondary",
		[Unit.CodeEnumDisposition.Neutral] 	= "CRB_Nameplates:sprNp_Target_NeutralSecondary",
		[Unit.CodeEnumDisposition.Friendly] = "CRB_Nameplates:sprNp_Target_FriendlySecondary",
		[10 + Unit.CodeEnumDisposition.Hostile] 	= "CRB_Nameplates:sprNp_Target_HostilePrimary",
		[10 + Unit.CodeEnumDisposition.Neutral] 	= "CRB_Nameplates:sprNp_Target_NeutralPrimary",
		[10 + Unit.CodeEnumDisposition.Friendly] = "CRB_Nameplates:sprNp_Target_FriendlyPrimary",
		[20 + Unit.CodeEnumDisposition.Hostile] 	= "CRB_Nameplates:sprNp_TargetSecondaryHostile",
		[20 + Unit.CodeEnumDisposition.Neutral] 	= "CRB_Nameplates:sprNp_TargetSecondaryNeutral",
		[20 + Unit.CodeEnumDisposition.Friendly] = "CRB_Nameplates:sprNp_TargetSecondaryFriendly",
		[30 + Unit.CodeEnumDisposition.Hostile] 	= "CRB_Nameplates:sprNp_TargetPrimaryHostile",
		[30 + Unit.CodeEnumDisposition.Neutral] 	= "CRB_Nameplates:sprNp_TargetPrimaryNeutral",
		[30 + Unit.CodeEnumDisposition.Friendly] = "CRB_Nameplates:sprNp_TargetPrimaryFriendly",
	},

	tHealthBar =
	{
		[Unit.CodeEnumDisposition.Hostile] 	= "CRB_Nameplates:sprNP_RedProg",
		[Unit.CodeEnumDisposition.Neutral] 	= "CRB_Nameplates:sprNP_YellowProg",
		[Unit.CodeEnumDisposition.Friendly] = "CRB_Nameplates:sprNP_GreenProg",
	},
	
	tHealthBarColor =
	{
		[Unit.CodeEnumDisposition.Hostile] 	= "xkcdRuby",
		[Unit.CodeEnumDisposition.Neutral] 	= "xkcdAmber",
		[Unit.CodeEnumDisposition.Friendly] = "xkcdAlgaeGreen",
	},

	tHealthTextColor =
	{
		[Unit.CodeEnumDisposition.Hostile] 	= "ffff8585",
		[Unit.CodeEnumDisposition.Neutral] 	= "ffffdb57",
		[Unit.CodeEnumDisposition.Friendly] = "ff9bff80",
	},
}

local ktHealthBarSprites =
{
	"sprNp_Health_FillGreen",
	"sprNp_Health_FillOrange",
	"sprNp_Health_FillRed"
}

local karClassIcons =
{
	[0] = "Icon_Windows_UI_CRB_Questlog_Call", --could not get class
	[GameLib.CodeEnumClass.Stalker] = "Icon_Windows_UI_CRB_Stalker",
	[GameLib.CodeEnumClass.Spellslinger] = "Icon_Windows_UI_CRB_Spellslinger",
	[GameLib.CodeEnumClass.Esper] = "Icon_Windows_UI_CRB_Esper",
	[GameLib.CodeEnumClass.Medic] = "Icon_Windows_UI_CRB_Medic",
	[GameLib.CodeEnumClass.Engineer] = "Icon_Windows_UI_CRB_Engineer",
	[GameLib.CodeEnumClass.Warrior] = "Icon_Windows_UI_CRB_Warrior",
}

local karThrottleUnitNames = 
{
	["enUS"] = {
		"Coldburrow Hunter",
		"Coldburrow Minion",
		"Coldburrow Shaman",
		"Coldburrow Warrior",
	},
	["deDE"] = {
		"Frostbau-Schamane",
		"Frostbau-Diener",
		"Frostbau-Krieger",
		"Frostbau-J\195\164ger",
	}
}

local karConColors =  -- differential value, color
{
	{-4, ApolloColor.new("ConTrivial")},
	{-3, ApolloColor.new("ConInferior")},
	{-2, ApolloColor.new("ConMinor")},
	{-1, ApolloColor.new("ConEasy")},
	{0, ApolloColor.new("ConAverage")},
	{1, ApolloColor.new("ConModerate")},
	{2, ApolloColor.new("ConTough")},
	{3, ApolloColor.new("ConHard")},
	{4, ApolloColor.new("ConImpossible")}
}

local kcrScalingHex 	= "ffffbf80"
local kcrScalingCColor 	= CColor.new(1.0, 191/255, 128/255, 0.7)

local kcrUnflaggedGroupmate = ApolloColor.new("DispositionFriendlyUnflaggedDull")
local kcrUnflaggedGuildmate = ApolloColor.new("DispositionGuildmateUnflagged")
local kcrUnflaggedAlly = ApolloColor.new("DispositionFriendlyUnflagged")
local kcrFlaggedAlly = ApolloColor.new("DispositionFriendly")
local kcrUnflaggedEnemyWhenUnflagged = ApolloColor.new("DispositionNeutral")
local kcrFlaggedEnemyWhenUnflagged = ApolloColor.new("DispositionPvPFlagMismatch")
local kcrUnflaggedEnemyWhenFlagged = ApolloColor.new("DispositionPvPFlagMismatch")
local kcrFlaggedEnemyWhenFlagged = ApolloColor.new("DispositionHostile")
local kcrDeadColor = ApolloColor.new("crayGray")

local kcrDefaultTaggedColor = ApolloColor.new("crayGray")

local knConColors = 9 -- needs to match the count of colors in karConColors

local knTargetRange = 40000 -- the distance^2 that normal nameplates should draw within (max targeting range)

local knDefaultLeft, knDefaultTop, knDefaultRight, knDefaultBottom  = -150, -93, 150, 30 -- -156, -66, 144, 30
local knMarkerSideOffset = 10
local knMarkerSideOffsetMod = 5

local karThreatMarkerOffsets =
{
	--Left = { nLeft = 56, nTop = -2, nRight = 84, nBottom = 26 },
	--Top = { nLeft = 110, nTop = -26, nRight = 138, nBottom = 2 }
	Left = { nLeft = -90, nTop = -2, nRight = -62, nBottom = 26 },
	Top = { nLeft = -50, nTop = 24, nRight = -22, nBottom = 52 }
}

local karClassNames = 
{
	[GameLib.CodeEnumClass.Esper] = "Esper",
	[GameLib.CodeEnumClass.Warrior] = "Warrior",
	[GameLib.CodeEnumClass.Stalker] = "Stalker",
	[GameLib.CodeEnumClass.Medic] = "Medic",
	[GameLib.CodeEnumClass.Spellslinger] = "Spellslinger",
	[GameLib.CodeEnumClass.Engineer] = "Engineer",
}

local karDispositionNames =
{
	[Unit.CodeEnumDisposition.Hostile] = "Hostile",
	[Unit.CodeEnumDisposition.Neutral] = "Neutral",
	[Unit.CodeEnumDisposition.Friendly] = "Friendly",
}

-- valid color values are:
	--   standard names, like "red"
  --   Xkcd names, see  http://xkcd.com/color/rgb/ 
  --   ARGB in hex format, <alpha><red><green><blue>, example "ff4169e1" = royal blue (4169e1) with 100% opacity (ff)
local karColorSets =
{
	Common =
	{
		Vulnerable = { Name = nil, HealthBar = "xkcdFuchsia" },
		Shield = { Name = nil, HealthBar = "cyan" },
		Absorb = { Name = nil, HealthBar = "ffff8c00" },
	},
	Self = 
	{
		NoPvP = { Name = "ff32cd00", HealthBar = "ff32cd32" },
		PvP = { Name = "ff01ff07", HealthBar = "ff00ff00" },
		--[[
		NoPvP = { Name = "xkcdAlgaeGreen", HealthBar = "xkcdAlgaeGreen" },
		PvP = { Name = "ff01ff07", HealthBar = "ff21c36f" },
		]]
	},
	NonPlayer =
	{
		[karDispositionNames[Unit.CodeEnumDisposition.Hostile]] 	= { Name = "ffff0000", HealthBar = "ffdc143c" },
		[karDispositionNames[Unit.CodeEnumDisposition.Neutral]] 	= { Name = "ffffd700", HealthBar = "ffffca00" },
		[karDispositionNames[Unit.CodeEnumDisposition.Friendly]] = { Name = "ff00ff00", HealthBar = "ff32cd32" },
	},
	Pet = -- includes "Pet", "Esper Pet" and "Scanner"
	{
		[karDispositionNames[Unit.CodeEnumDisposition.Hostile]] 	= { Name = "ffdc143c", HealthBar = "ffdc143c" },
		[karDispositionNames[Unit.CodeEnumDisposition.Neutral]] 	= { Name = "ffffd700", HealthBar = "ffffca00" },
		[karDispositionNames[Unit.CodeEnumDisposition.Friendly]] = { Name = "ff32cd32", HealthBar = "ff32cd32" },
	},
	Simple =
	{
		[karDispositionNames[Unit.CodeEnumDisposition.Hostile]] 	= { Name = "ffdc143c", HealthBar = "ffdc143c" },
		[karDispositionNames[Unit.CodeEnumDisposition.Neutral]] 	= { Name = "ffffd700", HealthBar = "ffffca00" },
		[karDispositionNames[Unit.CodeEnumDisposition.Friendly]] = { Name = "ff32cd32", HealthBar = "ff32cd32" },
	},
	Player =
	{
		Hostile	= { Name = "fff9de29", HealthBar = "fffa2a55" },
		HostilePvPMe	= { Name = "ffd18045", HealthBar = "fffa2a55" },
		HostilePvPHim	= { Name = "ffd18045", HealthBar = "fffa2a55" },
		HostilePvPBoth	= { Name = "ffff0500", HealthBar = "fffa2a55" },
		Friendly = { Name = "ff21c36f", HealthBar = "ff21c36f" },
		FriendlyPvPMe = { Name = "ff01ff07", HealthBar = "ff21c36f" },
		FriendlyPvPHim = { Name = "ff01ff07", HealthBar = "ff21c36f" },
		FriendlyPvPBoth = { Name = "ff01ff07", HealthBar = "ff21c36f" },
		Group = { Name = "ff6495ed", HealthBar = "ff87e005" },
		GroupPvPMe = { Name = "ff00bfff", HealthBar = "ff87e005" },
		GroupPvPHim = { Name = "ff00bfff", HealthBar = "ff87e005" },
		GroupPvPBoth = { Name = "ff00bfff", HealthBar = "ff87e005" },
		Guild = { Name = "ffba55d3", HealthBar = "ff32cd32" },
		GuildPvPMe = { Name = "ffff00ff", HealthBar = "ff32cd32" },
		GuildPvPHim = { Name = "ffff00ff", HealthBar = "ff32cd32" },
		GuildPvPBoth = { Name = "ffff00ff", HealthBar = "ff32cd32" },
		
		--[[
		Hostile	= { Name = "fff9de29", HealthBar = "fffa2a55" },
		HostilePvPMe	= { Name = "ffd18045", HealthBar = "fffa2a55" },
		HostilePvPHim	= { Name = "ffd18045", HealthBar = "fffa2a55" },
		HostilePvPBoth	= { Name = "ffff0500", HealthBar = "fffa2a55" },
		Friendly = { Name = "ff00aed7", HealthBar = "ff21c36f" },
		FriendlyPvPMe = { Name = "ff01ff07", HealthBar = "ff21c36f" },
		FriendlyPvPHim = { Name = "ff01ff07", HealthBar = "ff21c36f" },
		FriendlyPvPBoth = { Name = "ff01ff07", HealthBar = "ff21c36f" },
		Group = { Name = "ff569eb0", HealthBar = "ff87fd05" },
		GroupPvPMe = { Name = "ff01ff07", HealthBar = "ff87fd05" },
		GroupPvPHim = { Name = "ff01ff07", HealthBar = "ff87fd05" },
		GroupPvPBoth = { Name = "ff01ff07", HealthBar = "ff87fd05" },
		Guild = { Name = "ffba55d3", HealthBar = "ff32cd32" },
		GuildPvPMe = { Name = "ff01ff07", HealthBar = "ff32cd32" },
		GuildPvPHim = { Name = "ff01ff07", HealthBar = "ff32cd32" },
		GuildPvPBoth = { Name = "ff01ff07", HealthBar = "ff32cd32" },
		]]
	},
	PlayerClass = 
	{
		[karClassNames[GameLib.CodeEnumClass.Esper]] = 
		{
			Hostile	= { Name = "fff9de29", HealthBar = "fffa2a55" },
			HostilePvPMe	= { Name = "ffd18045", HealthBar = "fffa2a55" },
			HostilePvPHim	= { Name = "ffd18045", HealthBar = "fffa2a55" },
			HostilePvPBoth	= { Name = "ffff0500", HealthBar = "fffa2a55" },
			Friendly = { Name = "ff21c36f", HealthBar = "ff21c36f" },
			FriendlyPvPMe = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			FriendlyPvPHim = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			FriendlyPvPBoth = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			Group = { Name = "ff6495ed", HealthBar = "ff87e005" },
			GroupPvPMe = { Name = "ff00bfff", HealthBar = "ff87e005" },
			GroupPvPHim = { Name = "ff00bfff", HealthBar = "ff87e005" },
			GroupPvPBoth = { Name = "ff00bfff", HealthBar = "ff87e005" },
			Guild = { Name = "ffba55d3", HealthBar = "ff32cd32" },
			GuildPvPMe = { Name = "ffff00ff", HealthBar = "ff32cd32" },
			GuildPvPHim = { Name = "ffff00ff", HealthBar = "ff32cd32" },
			GuildPvPBoth = { Name = "ffff00ff", HealthBar = "ff32cd32" },
		},
		[karClassNames[GameLib.CodeEnumClass.Warrior]] = 
		{
			Hostile	= { Name = "fff9de29", HealthBar = "fffa2a55" },
			HostilePvPMe	= { Name = "ffd18045", HealthBar = "fffa2a55" },
			HostilePvPHim	= { Name = "ffd18045", HealthBar = "fffa2a55" },
			HostilePvPBoth	= { Name = "ffff0500", HealthBar = "fffa2a55" },
			Friendly = { Name = "ff21c36f", HealthBar = "ff21c36f" },
			FriendlyPvPMe = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			FriendlyPvPHim = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			FriendlyPvPBoth = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			Group = { Name = "ff6495ed", HealthBar = "ff87e005" },
			GroupPvPMe = { Name = "ff00bfff", HealthBar = "ff87e005" },
			GroupPvPHim = { Name = "ff00bfff", HealthBar = "ff87e005" },
			GroupPvPBoth = { Name = "ff00bfff", HealthBar = "ff87e005" },
			Guild = { Name = "ffba55d3", HealthBar = "ff32cd32" },
			GuildPvPMe = { Name = "ffff00ff", HealthBar = "ff32cd32" },
			GuildPvPHim = { Name = "ffff00ff", HealthBar = "ff32cd32" },
			GuildPvPBoth = { Name = "ffff00ff", HealthBar = "ff32cd32" },
		},
		[karClassNames[GameLib.CodeEnumClass.Stalker]] = 
		{
			Hostile	= { Name = "fff9de29", HealthBar = "fffa2a55" },
			HostilePvPMe	= { Name = "ffd18045", HealthBar = "fffa2a55" },
			HostilePvPHim	= { Name = "ffd18045", HealthBar = "fffa2a55" },
			HostilePvPBoth	= { Name = "ffff0500", HealthBar = "fffa2a55" },
			Friendly = { Name = "ff21c36f", HealthBar = "ff21c36f" },
			FriendlyPvPMe = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			FriendlyPvPHim = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			FriendlyPvPBoth = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			Group = { Name = "ff6495ed", HealthBar = "ff87e005" },
			GroupPvPMe = { Name = "ff00bfff", HealthBar = "ff87e005" },
			GroupPvPHim = { Name = "ff00bfff", HealthBar = "ff87e005" },
			GroupPvPBoth = { Name = "ff00bfff", HealthBar = "ff87e005" },
			Guild = { Name = "ffba55d3", HealthBar = "ff32cd32" },
			GuildPvPMe = { Name = "ffff00ff", HealthBar = "ff32cd32" },
			GuildPvPHim = { Name = "ffff00ff", HealthBar = "ff32cd32" },
			GuildPvPBoth = { Name = "ffff00ff", HealthBar = "ff32cd32" },
		},
		[karClassNames[GameLib.CodeEnumClass.Medic]] = 
		{
			Hostile	= { Name = "fff9de29", HealthBar = "fffa2a55" },
			HostilePvPMe	= { Name = "ffd18045", HealthBar = "fffa2a55" },
			HostilePvPHim	= { Name = "ffd18045", HealthBar = "fffa2a55" },
			HostilePvPBoth	= { Name = "ffff0500", HealthBar = "fffa2a55" },
			Friendly = { Name = "ff21c36f", HealthBar = "ff21c36f" },
			FriendlyPvPMe = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			FriendlyPvPHim = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			FriendlyPvPBoth = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			Group = { Name = "ff6495ed", HealthBar = "ff87e005" },
			GroupPvPMe = { Name = "ff00bfff", HealthBar = "ff87e005" },
			GroupPvPHim = { Name = "ff00bfff", HealthBar = "ff87e005" },
			GroupPvPBoth = { Name = "ff00bfff", HealthBar = "ff87e005" },
			Guild = { Name = "ffba55d3", HealthBar = "ff32cd32" },
			GuildPvPMe = { Name = "ffff00ff", HealthBar = "ff32cd32" },
			GuildPvPHim = { Name = "ffff00ff", HealthBar = "ff32cd32" },
			GuildPvPBoth = { Name = "ffff00ff", HealthBar = "ff32cd32" },
		},
		[karClassNames[GameLib.CodeEnumClass.Spellslinger]] = 
		{
			Hostile	= { Name = "fff9de29", HealthBar = "fffa2a55" },
			HostilePvPMe	= { Name = "ffd18045", HealthBar = "fffa2a55" },
			HostilePvPHim	= { Name = "ffd18045", HealthBar = "fffa2a55" },
			HostilePvPBoth	= { Name = "ffff0500", HealthBar = "fffa2a55" },
			Friendly = { Name = "ff21c36f", HealthBar = "ff21c36f" },
			FriendlyPvPMe = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			FriendlyPvPHim = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			FriendlyPvPBoth = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			Group = { Name = "ff6495ed", HealthBar = "ff87e005" },
			GroupPvPMe = { Name = "ff00bfff", HealthBar = "ff87e005" },
			GroupPvPHim = { Name = "ff00bfff", HealthBar = "ff87e005" },
			GroupPvPBoth = { Name = "ff00bfff", HealthBar = "ff87e005" },
			Guild = { Name = "ffba55d3", HealthBar = "ff32cd32" },
			GuildPvPMe = { Name = "ffff00ff", HealthBar = "ff32cd32" },
			GuildPvPHim = { Name = "ffff00ff", HealthBar = "ff32cd32" },
			GuildPvPBoth = { Name = "ffff00ff", HealthBar = "ff32cd32" },
		},
		[karClassNames[GameLib.CodeEnumClass.Engineer]] = 
		{
			Hostile	= { Name = "fff9de29", HealthBar = "fffa2a55" },
			HostilePvPMe	= { Name = "ffd18045", HealthBar = "fffa2a55" },
			HostilePvPHim	= { Name = "ffd18045", HealthBar = "fffa2a55" },
			HostilePvPBoth	= { Name = "ffff0500", HealthBar = "fffa2a55" },
			Friendly = { Name = "ff21c36f", HealthBar = "ff21c36f" },
			FriendlyPvPMe = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			FriendlyPvPHim = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			FriendlyPvPBoth = { Name = "ff01ff07", HealthBar = "ff21c36f" },
			Group = { Name = "ff6495ed", HealthBar = "ff87e005" },
			GroupPvPMe = { Name = "ff00bfff", HealthBar = "ff87e005" },
			GroupPvPHim = { Name = "ff00bfff", HealthBar = "ff87e005" },
			GroupPvPBoth = { Name = "ff00bfff", HealthBar = "ff87e005" },
			Guild = { Name = "ffba55d3", HealthBar = "ff32cd32" },
			GuildPvPMe = { Name = "ffff00ff", HealthBar = "ff32cd32" },
			GuildPvPHim = { Name = "ffff00ff", HealthBar = "ff32cd32" },
			GuildPvPBoth = { Name = "ffff00ff", HealthBar = "ff32cd32" },
		},
	},
}
	
local ktWeaponUnitNamePattern = {
enUS = "(%S+ %S+)'s Weapon",                        
deDE = "Waffe von (%S+ %S+)",
frFR = "Arme de (%S+ %S+)",
koKR = "(%S+ %S+)'s Weapon"
}
	
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function SteadyPlates:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here

    return o
end

function SteadyPlates:Init()
	local bHasConfigureFunction = true
	local strConfigureButtonText = "SteadyPlates"
	local tDependencies = {
		"Nameplates", "RewardIcons",
	}
	
	self.settingStatus = nil
	
	-- load default settings, they will be adjusted later on, if there are saved settings available
	self:LoadDefaultSettings()
	
	self.settingStatus = "defaults"
	
  Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

function SteadyPlates:ShouldOverrideHealthShieldBar()
	return self.bReplaceHealthText or self.bHideShieldAbsorbLoss or self.bUseFixedHealthShieldBarRatio or self.bHealthBarShieldAbove or self.bHideShieldProgressEdgeGlow or self.bOverwriteHealthBarColor or self.bOverwriteShieldBarColor or self.bOverwriteAbsorbBarColor
end

function SteadyPlates:LoadDefaultSettings()
	self.bShowThreatMarker = true
	self.bShowInterruptArmorMarker = true
	self.bEnsureTargetNameplateOnScreen = true
	self.bDisableHiddenCheckIfHostile = false
	self.strInterruptArmorMarkerSprite = "sprResourceBar_InterruptCircleOrange" --"sprMap_QuestMarker"
	self.strThreatMarkerSprite = "Icon_Windows_UI_CRB_Rival"
	self.bBubblesNeverHideNameplates = false
	self.bEnsureCombatNameplatesOnScreen = false
	self.bThreatMarkerLeft = false
	self.nHealthBarTextFriendly = 0
	self.nHealthBarTextNeutral = 0
	self.nHealthBarTextHostile = 0
	self.bShowOnMouseOver = true
	self.bShowClusterMarker = true
	self.bShowCastBarTargetOnlyFriendly = true
	self.bShowCastBarTargetOnlyNeutral = true
	self.bShowCastBarTargetOnlyHostile = false
	self.bRemoveHealthBarBackground = false
	self.bHideShieldAbsorbLoss = false
	self.bReplaceHealthText = true
	self.bReplaceHealthTextFriendlyDeficit = true
	self.bReplaceHealthTextHideShield = false
	self.bReplaceHealthTextPercentages = true
	self.bHideShieldProgressEdgeGlow = true
	self.bHideMyNameWhenNotTargeted = true
	self.bHidePetsHostile = false
	self.bHidePetsFriendly = false
	self.bHidePetsNeutral = false
	self.bHideNonPlayerHostile = false
	self.bHideNonPlayerFriendly = false
	self.bHideNonPlayerNeutral = false
	self.bShowHealthBarTargetOnlyFriendly = false
	self.bShowHealthBarTargetOnlyNeutral = false
	self.bShowHealthBarTargetOnlyHostile = false
	self.bWotWThrottle = false
	self.bShowCleanseMarker = true
	self.strCleanseMarkerSprite = "Icon_Windows_UI_CRB_Adventure_Malgrave_Fatigue"
	self.bCleanseMarkerGroupOnly = true
	self.bInterruptArmorMarkerUsePlayerClass = false
	self.bNameplateClickTargetsUnit = true
	self.bExtendCertainDeath = true
	self.bBubblesAlwaysHideNameplates = false
	self.bShowCastBarNameFriendly = true
	self.bShowCastBarNameNeutral = true
	self.bShowCastBarNameHostile = true
	self.bHideMyCastBar = false
	self.bAdjustCertainDeath = true
	self.bHideHostilePlayersOutOfPvP = false
	self.nTargetMarkerArrowModifier = 30
	self.bShowTargetMarkerArrow = true
	self.bThreatMarkerTankMode = false
	self.bUseCustomHealthTextFormat = false
	self.strCustomHealthTextFormat = "h+ | h%%"
	self.bHideTitle = false
	self.nEnsureMode = 1
	self.nEnsureResetDelay = 25
	self.bHealthBarShieldAbove = false
	self.bShowThreatMarkerGroupOnly = false
	self.bShowThreatMarkerInstanceOnly = false
	self.bOverwriteHealthBarColor = true
	self.bOverwriteNameColor = true
	self.bOverwriteColorPlayerPvPSpecific = false
	self.bOverwriteColorPlayerClassSpecific = false
	self.bOverwriteShieldBarColor = true
	self.bOverwriteAbsorbBarColor = true
	self.bDimmHealthInCombatFriendly = false
	self.bAnchorMyNameplateToFeet = false
	self.bDisableOcclusionForTarget = false
	self.bDisableOcclusionForGroupMembers = false
	self.bShowSmallClassIconLeftToName = false
	self.bShowPlayerWeaponNameplate = true
	self.bShowPlayerWeaponIcon = true
		
	self:LoadDefaultColorSets()
	self.bHealthShieldBarOverride = self:ShouldOverrideHealthShieldBar()
end

function SteadyPlates:LoadDefaultColorSets()
	self.tColorSets = Dev.CopyTable(karColorSets)
end

local function ImportColorSetValues(tFrom, tTo)
	if tFrom and tFrom.Name then
		tTo.Name = tFrom.Name
	end
	if tFrom and tFrom.HealthBar then
		tTo.HealthBar = tFrom.HealthBar
	end
end

local function ImportColorSetDispositions(tFrom, tTo)
	if tFrom and tTo then
		ImportColorSetValues(tFrom[karDispositionNames[Unit.CodeEnumDisposition.Hostile]], tTo[karDispositionNames[Unit.CodeEnumDisposition.Hostile]])
		ImportColorSetValues(tFrom[karDispositionNames[Unit.CodeEnumDisposition.Neutral]], tTo[karDispositionNames[Unit.CodeEnumDisposition.Neutral]])
		ImportColorSetValues(tFrom[karDispositionNames[Unit.CodeEnumDisposition.Friendly]], tTo[karDispositionNames[Unit.CodeEnumDisposition.Friendly]])
	end
end

local function ImportColorSetPlayer(tFrom, tTo)
	if tFrom and tTo then
		ImportColorSetValues(tFrom.Hostile, tTo.Hostile)
		ImportColorSetValues(tFrom.HostilePvPMe, tTo.HostilePvPMe)
		ImportColorSetValues(tFrom.HostilePvPHim, tTo.HostilePvPHim)
		ImportColorSetValues(tFrom.HostilePvPBoth, tTo.HostilePvPBoth)
		ImportColorSetValues(tFrom.Friendly, tTo.Friendly)
		ImportColorSetValues(tFrom.FriendlyPvPMe, tTo.FriendlyPvPMe)
		ImportColorSetValues(tFrom.FriendlyPvPHim, tTo.FriendlyPvPHim)
		ImportColorSetValues(tFrom.FriendlyPvPBoth, tTo.FriendlyPvPBoth)
		ImportColorSetValues(tFrom.Group, tTo.Group)
		ImportColorSetValues(tFrom.GroupPvPMe, tTo.GroupPvPMe)
		ImportColorSetValues(tFrom.GroupPvPHim, tTo.GroupPvPHim)
		ImportColorSetValues(tFrom.GroupPvPBoth, tTo.GroupPvPBoth)
		ImportColorSetValues(tFrom.Guild, tTo.Guild)
		ImportColorSetValues(tFrom.GuildPvPMe, tTo.GuildPvPMe)
		ImportColorSetValues(tFrom.GuildPvPHim, tTo.GuildPvPHim)
		ImportColorSetValues(tFrom.GuildPvPBoth, tTo.GuildPvPBoth)
	end
end

local function ImportColorSetCommon(tFrom, tTo)
	if tFrom and tTo then
		ImportColorSetValues(tFrom.Vulnerable, tTo.Vulnerable)
		ImportColorSetValues(tFrom.Shield, tTo.Shield)
		ImportColorSetValues(tFrom.Absorb, tTo.Absorb)
	end
end

local function ImportColorSetSelf(tFrom, tTo)
	if tFrom and tTo then
		ImportColorSetValues(tFrom.NoPvP, tTo.NoPvP)
		ImportColorSetValues(tFrom.PvP, tTo.PvP)
	end
end

function SteadyPlates:ImportColorSets(tColorSets)
	local strSet
	local tColorSetsNew = Dev.CopyTable(self.tColorSets)
	
	strSet = "Common"
	if tColorSets[strSet] then
		ImportColorSetCommon(tColorSets[strSet], tColorSetsNew[strSet])
	end
	strSet = "Self"
	if tColorSets[strSet] then
		ImportColorSetSelf(tColorSets[strSet], tColorSetsNew[strSet])
	end
	strSet = "NonPlayer"
	if tColorSets[strSet] then
		ImportColorSetDispositions(tColorSets[strSet], tColorSetsNew[strSet])
	end
	strSet = "Pet"
	if tColorSets[strSet] then
		ImportColorSetDispositions(tColorSets[strSet], tColorSetsNew[strSet])
	end
	strSet = "Simple"
	if tColorSets[strSet] then
		ImportColorSetDispositions(tColorSets[strSet], tColorSetsNew[strSet])
	end
	strSet = "Player"
	if tColorSets[strSet] then
		ImportColorSetPlayer(tColorSets[strSet], tColorSetsNew[strSet])
	end
	strSet = "PlayerClass"
	if tColorSets[strSet] then
		local tFrom, tTo = tColorSets[strSet], tColorSetsNew[strSet]
		if tFrom and tTo then
			ImportColorSetPlayer(tFrom[karClassNames[GameLib.CodeEnumClass.Esper]], tTo[karClassNames[GameLib.CodeEnumClass.Esper]])
			ImportColorSetPlayer(tFrom[karClassNames[GameLib.CodeEnumClass.Warrior]], tTo[karClassNames[GameLib.CodeEnumClass.Warrior]])
			ImportColorSetPlayer(tFrom[karClassNames[GameLib.CodeEnumClass.Stalker]], tTo[karClassNames[GameLib.CodeEnumClass.Stalker]])
			ImportColorSetPlayer(tFrom[karClassNames[GameLib.CodeEnumClass.Medic]], tTo[karClassNames[GameLib.CodeEnumClass.Medic]])
			ImportColorSetPlayer(tFrom[karClassNames[GameLib.CodeEnumClass.Spellslinger]], tTo[karClassNames[GameLib.CodeEnumClass.Spellslinger]])
			ImportColorSetPlayer(tFrom[karClassNames[GameLib.CodeEnumClass.Engineer]], tTo[karClassNames[GameLib.CodeEnumClass.Engineer]])
		end
	end
	
	return tColorSetsNew
end

-----------------------------------------------------------------------------------------------
-- SteadyPlates Events
-----------------------------------------------------------------------------------------------

function SteadyPlates:OnLoad()
  -- load our form file
	--Apollo.LoadSprites("SP_Nameplates.xml", "SP_Nameplates") 
	Apollo.LoadSprites("SteadyPlates_Sprites.xml", "SP_Nameplates") 
	self.xmlDoc = XmlDoc.CreateFromFile("SteadyPlates.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	
	local strCancel = Apollo.GetString(1)
	if strCancel == "Abbrechen" then
		self.arThrottleUnitNames = karThrottleUnitNames["deDE"]
		self.strWeaponUnitNamePattern = ktWeaponUnitNamePattern["deDE"]
	elseif strCancel == "Annuler" then
		self.arThrottleUnitNames = nil --karThrottleUnitNames["frFR"]
		self.strWeaponUnitNamePattern = ktWeaponUnitNamePattern["frFR"]
	elseif strCancel == "Cancel" then
		self.arThrottleUnitNames = karThrottleUnitNames["enUS"]
		self.strWeaponUnitNamePattern = ktWeaponUnitNamePattern["enUS"]
	else
		self.arThrottleUnitNames = nil
		self.strWeaponUnitNamePattern = nil
	end	
end

function SteadyPlates:OnDependencyError(strDependency, strError)
  Print("[SteadyPlates]: Could not find addon dependency '"..strDependency.."! SteadyPlates may throw errors or won't work at all!")
	return true
end

function SteadyPlates:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		Apollo.RegisterSlashCommand("spl", "OnSlashCommand", self)
		Apollo.RegisterSlashCommand("steadyplates", "OnSlashCommand", self)
	
		aNameplates = Apollo.GetAddon("Nameplates")
		if not aNameplates then
				Print("[SteadyPlates]: Could not find Carbine nameplates. SteadyPlates will be disabled!")
				return
		end
		
		-- locally cache function references to be used in for loop
		fnDrawHealth = SteadyPlates.DrawHealthHook
		fnDrawCastBar = SteadyPlates.DrawCastBarHook
		fnDrawTargeting = SteadyPlates.DrawTargeting
		fnDrawThreat = SteadyPlates.DrawThreat
		fnDrawInterruptArmor = SteadyPlates.DrawInterruptArmor
		fnDrawCleanse = SteadyPlates.DrawCleanse
		fnDrawCluster = SteadyPlates.DrawCluster
		fnHelperChangeHealthText = SteadyPlates.HelperChangeHealthText
		fnHelperEnsureNameplateOnScreen = SteadyPlates.HelperEnsureNameplateOnScreen
		fnDrawCertainDeath = SteadyPlates.DrawCertainDeath
		fnDrawRewards = aNameplates.DrawRewards
		fnColorNameplate = SteadyPlates.ColorNameplateHook
		fnDrawVulnerable = aNameplates.DrawVulnerable
		fnHelperCalculateConValue = SteadyPlates.HelperCalculateConValueHook
		
		local tMax = Apollo.GetDisplaySize()
		nUIScale = tMax.Scale
		
		fHelperCalculateConValue = aNameplates.HelperCalculateConValue
		aNameplates.HelperCalculateConValue = SteadyPlates.HelperCalculateConValueHook
	
		fNameplatesHelperVerifyVisibilityOptions = aNameplates.HelperVerifyVisibilityOptions
		aNameplates.HelperVerifyVisibilityOptions = SteadyPlates.HelperVerifyVisibilityOptionsHook
		
		fNameplatesUpdateNameplateRewardInfo = aNameplates.UpdateNameplateRewardInfo
		aNameplates.UpdateNameplateRewardInfo = SteadyPlates.UpdateNameplateRewardInfoHook
		
		fNameplatesOnUnitOcclusionChanged = aNameplates.OnUnitOcclusionChanged
		aNameplates.OnUnitOcclusionChanged = SteadyPlates.OnUnitOcclusionChangedHook
		
		fCheckDrawDistance = aNameplates.CheckDrawDistance
		aNameplates.CheckDrawDistance = SteadyPlates.CheckDrawDistanceHook
		
		fOnTargetUnitChanged = aNameplates.OnTargetUnitChanged
		aNameplates.OnTargetUnitChanged = SteadyPlates.OnTargetUnitChangedHook
		
		fDrawHealth = aNameplates.DrawHealth
		aNameplates.DrawHealth = SteadyPlates.DrawHealthHook
		
		fDrawCastBar = aNameplates.DrawCastBar
		aNameplates.DrawCastBar = SteadyPlates.DrawCastBarHook
		
		fDrawHealthShieldBar = aNameplates.DrawHealthShieldBar
		aNameplates.DrawHealthShieldBar = SteadyPlates.DrawHealthShieldBarHook
		
		fDrawName = aNameplates.DrawName
		aNameplates.DrawName = SteadyPlates.DrawNameHook
		
		fDrawGuild = aNameplates.DrawGuild
		aNameplates.DrawGuild = SteadyPlates.DrawGuildHook
		
		fOnFrame = aNameplates.OnFrame
		aNameplates.OnFrame = SteadyPlates.OnFrameHook
		
		fNameplatesOnUnitCreated = aNameplates.OnUnitCreated
		aNameplates.OnUnitCreated = SteadyPlates.OnUnitCreatedHook
		
		fOnWorldLocationOnScreen = aNameplates.OnWorldLocationOnScreen
		aNameplates.OnWorldLocationOnScreen = SteadyPlates.OnWorldLocationOnScreenHook
		
		fUpdateNameplateVisibility = aNameplates.UpdateNameplateVisibility
		aNameplates.UpdateNameplateVisibility = SteadyPlates.UpdateNameplateVisibilityHook
		
		fColorNameplate = aNameplates.ColorNameplate
		aNameplates.ColorNameplate = SteadyPlates.ColorNameplateHook
		
		fOnPreloadUnitCreateTimer = aNameplates.OnPreloadUnitCreateTimer
		aNameplates.OnPreloadUnitCreateTimer = SteadyPlates.OnPreloadUnitCreateTimerHook
		
		--fNameplatesOnUnitDestroyed = aNameplates.OnUnitDestroyed
		--aNameplates.OnUnitDestroyed = SteadyPlates.OnUnitDestroyedHook
		
		Apollo.RegisterTimerHandler("SteadyPlates_ChangedWorldTimer", "OnChangedWorldTimer", self)
		Apollo.CreateTimer("SteadyPlates_ChangedWorldTimer", 0.5, false)
		
		--self.unitTargetChangeRequestUnit = nil
		--Apollo.RegisterTimerHandler("SteadyPlates_TargetChangeRequestTimer", "OnTargetChangeRequestTimer", self)
		--Apollo.CreateTimer("SteadyPlates_TargetChangeRequestTimer", 0.150, false)
		--Apollo.StopTimer("SteadyPlates_TargetChangeRequestTimer")
		
		Apollo.RegisterEventHandler("MouseOverUnitChanged", "OnMouseOverUnitChanged", self)
		Apollo.RegisterEventHandler("UnitEnteredCombat", "OnEnteredCombat", self)
		Apollo.RegisterEventHandler("ChangeWorld", "OnChangeWorld", self)
		Apollo.RegisterEventHandler("CharacterCreated", "OnCharacterCreated", self)
		--Apollo.RegisterEventHandler("VarChange_ZoneName", "OnVarChange_ZoneName", self)
		--Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)
		--Apollo.RegisterEventHandler("UnitDestroyed", "OnUnitDestroyed", self)
		Apollo.RegisterEventHandler("KeyBindingUpdated", "OnKeyBindingUpdated", self)
		Apollo.RegisterEventHandler("KeyBindingReceived", "OnKeyBindingUpdated", self)
		
		--Apollo.RegisterEventHandler("TargetThreatListUpdated",  "OnTargetThreatListUpdated",  self)
		--Apollo.RegisterEventHandler("TargetUnitChanged", "OnTargetUnitChanged", self)
		
		--self.wndTargetTest = Apollo.LoadForm(self.xmlDoc, "EnsureHelper", "InWorldHudStratum", self)
		--self.wndTargetTest:Show(false, true)
		
		aPerspectivePlates = Apollo.GetAddon("PerspectivePlates")
		if aPerspectivePlates then
			bPerspectivePlates = true
			Event_FireGenericEvent("GenericEvent_PerspectivePlates_RegisterOffsets", knDefaultLeft, knDefaultTop, knDefaultRight, knDefaultBottom)
		end
		
		uPlayerUnit = GameLib.GetPlayerUnit()
		--self.unitPlayerName = uPlayerUnit:GetName() --.."@"..unitPlayer:GetRealm()
		self:UpdatePreloadUnitNameplates()
	end
end

function SteadyPlates:OnConfigure()
	self:ShowSettingsWindow()
end

function SteadyPlates:OnRestore(eLevel, tData)
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end
	if tData ~= nil then
		local tSettings = 
		{
			"bShowThreatMarker",
			"bShowInterruptArmorMarker",
			"bEnsureTargetNameplateOnScreen",
			"strInterruptArmorMarkerSprite",
			"bBubblesNeverHideNameplates",
			"strThreatMarkerSprite",
			"bDisableHiddenCheckIfHostile",
			"bEnsureCombatNameplatesOnScreen",
			"bThreatMarkerLeft",
			"nHealthBarTextFriendly",
			"nHealthBarTextHostile",
			"nHealthBarTextNeutral",
			"bShowCastBarTargetOnlyFriendly",
			"bShowCastBarTargetOnlyNeutral",
			"bShowCastBarTargetOnlyHostile",
			"bRemoveHealthBarBackground",
			"bHideShieldAbsorbLoss",
			"bReplaceHealthText",
			"bReplaceHealthTextPercentages",
			"bReplaceHealthTextHideShield",
			"bHideShieldProgressEdgeGlow",
			"bHideMyNameWhenNotTargeted",
			"bReplaceHealthTextFriendlyDeficit",
			"bShowClusterMarker",
			"bHidePetsFriendly",
			"bHidePetsHostile",
			"bHidePetsNeutral",
			"bHideNonPlayerFriendly",
			"bHideNonPlayerHostile",
			"bHideNonPlayerNeutral",
			"bShowHealthBarTargetOnlyFriendly",
			"bShowHealthBarTargetOnlyNeutral",
			"bShowHealthBarTargetOnlyHostile",
			"bWotWThrottle",
			"bShowCleanseMarker",
			"strCleanseMarkerSprite",
			"bCleanseMarkerGroupOnly",
			"bInterruptArmorMarkerUsePlayerClass",
			"bShowOnMouseOver",
			"bNameplateClickTargetsUnit",
			"bExtendCertainDeath",
			"bBubblesAlwaysHideNameplates",
			"bShowCastBarNameFriendly",
			"bShowCastBarNameNeutral",
			"bShowCastBarNameHostile",
			"bHideMyCastBar",
			"bAdjustCertainDeath",
			"bHideHostilePlayersOutOfPvP",
			"nTargetMarkerArrowModifier",
			"bShowTargetMarkerArrow",
			"bThreatMarkerTankMode",
			"bUseCustomHealthTextFormat",
			"strCustomHealthTextFormat",
			"bHideTitle",
			"nEnsureMode",
			"nEnsureResetDelay",
			"bHealthBarShieldAbove",
			"bShowThreatMarkerGroupOnly",
			"bShowThreatMarkerInstanceOnly",
			"bDimmHealthInCombatFriendly",
			"bOverwriteHealthBarColor",
			"bOverwriteNameColor",
			"bOverwriteColorPlayerPvPSpecific",
			"bOverwriteColorPlayerClassSpecific",
			"bOverwriteShieldBarColor",
			"bOverwriteAbsorbBarColor",
			"bAnchorMyNameplateToFeet",
			"bDisableOcclusionForTarget",
			"bDisableOcclusionForGroupMembers",
			"bShowSmallClassIconLeftToName",
			"bShowPlayerWeaponNameplate",
	    "bShowPlayerWeaponIcon",
		}
		
		for _, strSetting in pairs(tSettings) do
			local tSetting = tData[strSetting]
			if tSetting ~= nil then
				self[strSetting] = tSetting
			end
		end
		
		if tData["tColorSets"] then
			self.tColorSets = self:ImportColorSets(tData["tColorSets"])
		end
		
		self.bHealthShieldBarOverride = self:ShouldOverrideHealthShieldBar()
	
		self:UpdatePreloadUnitNameplates()
	end
	
	self.settingStatus = "character"
end

function SteadyPlates:UpdatePreloadUnitNameplates(nSource)
	if self.bAddonRestoredOrLoaded then
		SteadyPlatesInst.bThreatMarkerPositionChanged = true
		self:UpdateExistingNameplatesToSettings(true)
		SteadyPlatesInst.bThreatMarkerPositionChanged = false
		self:OnKeyBindingUpdated()
	end
	self.bAddonRestoredOrLoaded = true
end

function SteadyPlates:OnKeyBindingUpdated()
	-- force to RewardIcons to update keybinds, because it may got the default ones due to loading order issues
	local aRewardIcons = Apollo.GetAddon("RewardIcons")
	if aRewardIcons then
		aRewardIcons:OnKeyBindingUpdated("Path Action")
		aRewardIcons:OnKeyBindingUpdated("Cast Objective Ability")
	end
end

function SteadyPlates:OnSave(eLevel)
    if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
      return nil
    end
		local tSettings =
		{
			bShowThreatMarker = self.bShowThreatMarker,
			bShowInterruptArmorMarker = self.bShowInterruptArmorMarker,
			bEnsureTargetNameplateOnScreen = self.bEnsureTargetNameplateOnScreen,
			strInterruptArmorMarkerSprite = self.strInterruptArmorMarkerSprite,
			bBubblesNeverHideNameplates = self.bBubblesNeverHideNameplates,
			strThreatMarkerSprite = self.strThreatMarkerSprite,
			bDisableHiddenCheckIfHostile = self.bDisableHiddenCheckIfHostile,
			bEnsureCombatNameplatesOnScreen = self.bEnsureCombatNameplatesOnScreen,
			bThreatMarkerLeft = self.bThreatMarkerLeft,
			nHealthBarTextFriendly = self.nHealthBarTextFriendly,
			nHealthBarTextHostile = self.nHealthBarTextHostile,
			nHealthBarTextNeutral = self.nHealthBarTextNeutral,
			bShowCastBarTargetOnlyFriendly = self.bShowCastBarTargetOnlyFriendly,
			bShowCastBarTargetOnlyNeutral = self.bShowCastBarTargetOnlyNeutral,
			bShowCastBarTargetOnlyHostile = self.bShowCastBarTargetOnlyHostile,
			bRemoveHealthBarBackground = self.bRemoveHealthBarBackground,
			bHideShieldAbsorbLoss = self.bHideShieldAbsorbLoss,
			bReplaceHealthText = self.bReplaceHealthText,
			bReplaceHealthTextPercentages = self.bReplaceHealthTextPercentages,
			bReplaceHealthTextHideShield = self.bReplaceHealthTextHideShield,
			bHideShieldProgressEdgeGlow = self.bHideShieldProgressEdgeGlow,
			bHideMyNameWhenNotTargeted = self.bHideMyNameWhenNotTargeted,
			bReplaceHealthTextFriendlyDeficit = self.bReplaceHealthTextFriendlyDeficit,
			bShowClusterMarker = self.bShowClusterMarker,
			bHidePetsFriendly = self.bHidePetsFriendly,
			bHidePetsHostile = self.bHidePetsHostile,
			bHidePetsNeutral = self.bHidePetsNeutral,
			bHideNonPlayerFriendly = self.bHideNonPlayerFriendly,
			bHideNonPlayerHostile = self.bHideNonPlayerHostile,
			bHideNonPlayerNeutral = self.bHideNonPlayerNeutral,
			bShowHealthBarTargetOnlyFriendly = self.bShowHealthBarTargetOnlyFriendly,
			bShowHealthBarTargetOnlyNeutral = self.bShowHealthBarTargetOnlyNeutral,
			bShowHealthBarTargetOnlyHostile = self.bShowHealthBarTargetOnlyHostile,
			bWotWThrottle = self.bWotWThrottle,
			bShowCleanseMarker = self.bShowCleanseMarker,
			strCleanseMarkerSprite = self.strCleanseMarkerSprite,
			bCleanseMarkerGroupOnly = self.bCleanseMarkerGroupOnly,
			bInterruptArmorMarkerUsePlayerClass = self.bInterruptArmorMarkerUsePlayerClass,
			bShowOnMouseOver = self.bShowOnMouseOver,
			bNameplateClickTargetsUnit = self.bNameplateClickTargetsUnit,
			bExtendCertainDeath = self.bExtendCertainDeath,
			bBubblesAlwaysHideNameplates = self.bBubblesAlwaysHideNameplates,
			bShowCastBarNameFriendly = self.bShowCastBarNameFriendly,
			bShowCastBarNameNeutral = self.bShowCastBarNameNeutral,
			bShowCastBarNameHostile = self.bShowCastBarNameHostile,
			bHideMyCastBar = self.bHideMyCastBar,
			bAdjustCertainDeath = self.bAdjustCertainDeath,
			bHideHostilePlayersOutOfPvP = self.bHideHostilePlayersOutOfPvP,
			nTargetMarkerArrowModifier = self.nTargetMarkerArrowModifier,
			bShowTargetMarkerArrow = self.bShowTargetMarkerArrow,
			bThreatMarkerTankMode = self.bThreatMarkerTankMode,
			bUseCustomHealthTextFormat = self.bUseCustomHealthTextFormat,
			strCustomHealthTextFormat = self.strCustomHealthTextFormat,
			bHideTitle = self.bHideTitle,
			nEnsureMode = self.nEnsureMode,
			nEnsureResetDelay = self.nEnsureResetDelay,
			bHealthBarShieldAbove = self.bHealthBarShieldAbove,
			bShowThreatMarkerGroupOnly = self.bShowThreatMarkerGroupOnly,
			bShowThreatMarkerInstanceOnly = self.bShowThreatMarkerInstanceOnly,
			bDimmHealthInCombatFriendly = self.bDimmHealthInCombatFriendly,
			tColorSets = self.tColorSets,
			bOverwriteHealthBarColor = self.bOverwriteHealthBarColor,
			bOverwriteNameColor = self.bOverwriteNameColor,
			bOverwriteColorPlayerPvPSpecific = self.bOverwriteColorPlayerPvPSpecific,
			bOverwriteColorPlayerClassSpecific = self.bOverwriteColorPlayerClassSpecific,
			bOverwriteShieldBarColor = self.bOverwriteShieldBarColor,
			bOverwriteAbsorbBarColor = self.bOverwriteAbsorbBarColor,
			bAnchorMyNameplateToFeet = self.bAnchorMyNameplateToFeet,
			bDisableOcclusionForTarget = self.bDisableOcclusionForTarget,
			bDisableOcclusionForGroupMembers = self.bDisableOcclusionForGroupMembers,
			bShowSmallClassIconLeftToName = self.bShowSmallClassIconLeftToName,
			bShowPlayerWeaponNameplate = self.bShowPlayerWeaponNameplate,
			bShowPlayerWeaponIcon = self.bShowPlayerWeaponIcon,
		}
		return tSettings
end

-- on SlashCommand "/spl"
function SteadyPlates:OnSlashCommand(cmd, argument)

	--[[local unit = GameLib.GetTargetUnit()
	if unit then
		--local tNameplate = aNameplates.arUnit2Nameplate[unit:GetId()]
		--local n = tonumber(argument) or 1
		--tNameplate.wndNameplate:SetUnit(unit, n)
		--if tNameplate and tNameplate.wnd.questRewards:GetData() then
		--	PrintObj(tNameplate.wnd.questRewards:GetData().tRewardInfo, 3)
		--end
		--PrintObj(unit)	
		Print(unit:GetFaction())
	end
	
	--PrintObj(GameLib.GetInputActionCategories(), 3)
	--PrintObj(debug.getinfo(GameLib.PauseGameActionInput))

	if true then return end--]]

	--PrintObj(ApolloColor.new("ConTrivial"):ToTable(), 3)
	--PrintObj(GameLib.CodeEnumClass.Esper)
	
	if not argument or argument == "" then
		self:ShowSettingsWindow()
		return
	end
	
	local s = string.lower(argument)
	if s == "hidden" or s == "h" then
		if not self.bDisableHiddenCheckIfHostile then
			CPrint("[SteadyPlates]: Hidden check is now disabled for hostile units!")
			self.bDisableHiddenCheckIfHostile = true
		else
			CPrint("[SteadyPlates]: Hidden check is now enabled for hostile units!")
			self.bDisableHiddenCheckIfHostile = false
		end
	elseif s == "ti" then
		local unit = GameLib.GetTargetUnit()
		if unit then
			local nId = unit:GetId()
			Print("TargetUnit: Id="..tostring(nId)..", Name="..tostring(unit:GetName())..", Type="..tostring(unit:GetType())..", Disposition="..tostring(unit:GetDispositionTo(GameLib.GetPlayerUnit()))..", IsValid="..tostring(unit:IsValid())..", ShouldShowNamePlate="..tostring(unit:ShouldShowNamePlate()))
			local tNameplate = aNameplates.arUnit2Nameplate[nId]
			if tNameplate then
				Print("Found nameplate for "..tostring(nId).." IsShown="..tostring(tNameplate.wndNameplate:IsShown())..", IsOccluded="..tostring(tNameplate.wndNameplate:IsOccluded())..", IsOnScreen="..tostring(tNameplate.wndNameplate:IsOnScreen())..", bEnsured="..tostring(tNameplate.bEnsured)..", bEnsureHide="..tostring(tNameplate.bEnsureHide)..", bShow="..tostring(tNameplate.bShow))
				Print("Forced visibility check result: "..tostring(SteadyPlates.HelperVerifyVisibilityOptionsHook(aNameplates, tNameplate)))
				local tAnchor = tNameplate.unitOwner:GetOverheadAnchor()
				local nLeft, nTop, nRight, nBottom = tNameplate.wndNameplate:GetAnchorOffsets()
				Print("Anchor: x="..tostring(tAnchor.x).." y="..tostring(tAnchor.y)..", Offsets: Left="..tostring(nLeft)..", Top="..tostring(nTop)..", Right="..tostring(nRight)..", Bottom="..tostring(nBottom))
			else
				Print("No nampelate created for "..tostring(nId))
			end
		end
	elseif s == "dev" then
		--[[local unit = GameLib.GetPlayerUnit()
		--PrintObj(unit, 2)
		Print("GameLib "..unit:GetId().." "..unit:GetName())
		unit = uPlayerUnit
		if unit then
			Print("OnCreated "..unit:GetId().." "..unit:GetName())
		end]]
		local unit = GameLib.GetTargetUnit()
		if unit then
			local tNameplate = aNameplates.arUnit2Nameplate[unit:GetId()]
			if tNameplate then
				PrintObj(tNameplate.wnd.healthBar)
			end
			--PrintObj(unit:GetOverheadAnchor(0), 2)
			--PrintObj(unit:GetOverheadAnchor(1), 2)
			--PrintObj(unit:GetOverheadAnchor(2), 2)
			--PrintObj(unit:GetEliteness())
			--PrintObj(unit:GetRank())
			--PrintObj(unit:GetDifficulty())
			--PrintObj(Unit.CodeEnumRank, 2)
		end
		--PrintObj(GameLib.GetTargetUnit():GetClassId())
	else
		CPrint("[SteadyPlates]: Unknown argument: "..argument..". To open settings type /spl")
	end
end

function SteadyPlates:OnTargetUnitChanged(unitTarget)
	if not unitTarget or not uPlayerUnit then return end
	
	if self.wndTargetTest then
		--self.wndTargetTest:SetUnit(unitTarget, 1) -- 1 top, 0 bottom, 2 hand left?, 3 hand right?, 4 body center?
		self.wndTargetTest:Show(true)
		
		--PrintObj(self.wndTargetTest, 2)
		
		--ClientPointToWindowPoint
		--WindowPointToClientPoint
		--Print("GetClientRect "..o2s(self.wndTargetTest:GetClientRect(), 2))
		--Print(o2s(self.wndTargetTest:WindowPointToClientPoint(), 2))
		
		local tAnchor = unitTarget:GetOverheadAnchor()
		--PrintObj(tAnchor)
		local tSelf = Vector2.New(tAnchor.x- 200, tAnchor.y +50)
		--PrintObj(tSelf)
		self.wndTargetTest:SetSelfAnchor(tSelf)
		
		self.wndTargetTest:SetUnit(unitTarget, 1)
		
		--Print("GetPos "..o2s(self.wndTargetTest:GetPos(), 2))
		--Print("GetOriginalLocation "..o2s(self.wndTargetTest:GetOriginalLocation(), 2))
		--Print("GetTransLocation "..o2s(self.wndTargetTest:GetTransLocation(), 2))
	end
		
	--PrintObj(unitTarget, 2)
	--PrintObj(unitTarget:GetClusterContributionValue())
	--PrintObj(unitTarget:GetFacing(), 2)
	--Print(tostring(unitTarget:GetDispositionTo(uPlayerUnit)).." "..tostring(unitTarget:GetFaction()))
	--PrintObj(GameLib.CodeEnumClass, 2)
	--PrintObj(unitTarget)
	--Print(tostring(unitTarget:GetType()))
	--PrintObj(GameLib.GetCurrentZoneMap(),3)
	--local zone = GameLib.GetCurrentZoneMap()
	--if zone then
	--	Print("Map:"..zone.id.." "..zone.strName)
	--end
	--Print("worldId="..tostring(GameLib.GetCurrentWorldId()).." zoneId="..tostring(GameLib.GetCurrentZoneId()).." zoneName="..tostring(GetCurrentZoneName()))
end

function SteadyPlates:OnTargetThreatListUpdated(...)
	--PrintObj(select("#", ...))
	
end

function SteadyPlates:OnMouseOverUnitChanged(unitMouseOver)
	if not self.bShowOnMouseOver then
		return
	end
	
	for _, tNameplateOther in pairs(aNameplates.arUnit2Nameplate) do
		--if tNameplateOther.bIsMouseOver and not tNameplateOther.bIsTarget then
		--	tNameplateOther.wndNameplate:Show(false, true)
		--	aNameplates:UpdateNameplateVisibility(tNameplateOther)
		--end
		if tNameplateOther.bIsMouseOver and not ((tNameplateOther.bIsTarget and SteadyPlatesInst.bEnsureTargetNameplateOnScreen) or (bPlayerInCombat and SteadyPlatesInst.bEnsureCombatNameplatesOnScreen)) then
			if tNameplateOther.bEnsured then
				SteadyPlates.HelperResetEnsure(tNameplateOther, SteadyPlatesInst.nEnsureMode)
			end
		end
		tNameplateOther.bIsMouseOver = false
	end
	
	if unitMouseOver == nil then
		return
	end
	
	local tNameplate = aNameplates.arUnit2Nameplate[unitMouseOver:GetId()]
	if tNameplate ~= nil then
		tNameplate.bIsMouseOver = true
		aNameplates:UpdateNameplateVisibility(tNameplate)
		--if tNameplate.bShow then
		--	tNameplate.wndNameplate:Show(true, true)
		--	tNameplate.bShow = true
		--	self.DrawNameplate(aNameplates, tNameplate)
		--end
		tNameplate.wndNameplate:ToFront()
	end
end

function SteadyPlates:OnEnteredCombat(unitChecked, bInCombat)
	local bIsThePlayer = unitChecked:IsThePlayer()
	if bIsThePlayer and bPlayerInCombat ~= bInCombat then
		--Print("CombatChangeto: "..tostring(bInCombat))
		if not bInCombat and SteadyPlatesInst.bEnsureCombatNameplatesOnScreen then
			-- we moved nameplates in combat, reset positions when leaving combat
			for _, tNameplateOther in pairs(aNameplates.arUnit2Nameplate) do
				if tNameplateOther.bEnsured then
					SteadyPlates.HelperResetEnsure(tNameplateOther, SteadyPlatesInst.nEnsureMode)
				end
			end
		end
		bPlayerInCombat = bInCombat
	end
end

function SteadyPlates:OnVarChange_ZoneName()
	--Print("SteadyPlates:OnVarChange_ZoneName worldId="..tostring(GameLib.GetCurrentWorldId()).." zoneId="..tostring(GameLib.GetCurrentZoneId()).." zoneName="..tostring(GetCurrentZoneName()))
	--self.bChangedWorldRecently = true
	--Apollo.StartTimer("SteadyPlates_ChangedWorldTimer")
end

function SteadyPlates:OnCharacterCreated()
	uPlayerUnit = GameLib.GetPlayerUnit()
	--Print("SteadyPlates:OnCharacterCreated "..unit.GetName().." "..unit.GetId())
	--self.bChangedWorldRecently = true
	--Apollo.StartTimer("SteadyPlates_ChangedWorldTimer")
end

function SteadyPlates:OnChangeWorld()
	uPlayerUnit = GameLib.GetPlayerUnit()
	--Print("SteadyPlates:OnChangeWorld worldId="..tostring(GameLib.GetCurrentWorldId()))
	self.bChangedWorldRecently = true
	Apollo.StartTimer("SteadyPlates_ChangedWorldTimer")
	
	self:CheckWotW()
end

function SteadyPlates:OnChangedWorldTimer()
	uPlayerUnit = GameLib.GetPlayerUnit()
  if not uPlayerUnit then
		Apollo.StartTimer("SteadyPlates_ChangedWorldTimer")
		return 
	end
		
	self.bChangedWorldRecently = false
	Apollo.StopTimer("SteadyPlates_ChangedWorldTimer")
	
	aNameplates:OnTargetUnitChanged(GameLib:GetTargetUnit())
	--aNameplates:UpdateAllNameplateVisibility()
		
	if self.bHideMyNameWhenNotTargeted then
		local tNameplate = aNameplates.arUnit2Nameplate[uPlayerUnit:GetId()]
		if tNameplate ~= nil and not tNameplate.bIsTarget then
			tNameplate.wnd.wndName:Show(false)
			tNameplate.wnd.wndGuild:Show(false)
			
			tNameplate.bIsMe = true
			if SteadyPlatesInst.bAnchorMyNameplateToFeet then
				if tNameplate.bMounted then
					tNameplate.wndNameplate:SetUnit(tNameplate.unitOwner:GetUnitMount(), 0)
				else
					tNameplate.wndNameplate:SetUnit(tNameplate.unitOwner, 0)
				end
			end
		end
	end
end

function SteadyPlates:OnTargetChangeRequestTimer()
	local unitTarget = SteadyPlatesInst.unitTargetChangeRequestUnit
	if unitTarget and GameLib.GetTargetUnit() ~= unitTarget then
		GameLib.SetTargetUnit(unitTarget)
	end
	SteadyPlatesInst.unitTargetChangeRequestUnit = nil
end

function SteadyPlates:IsPlayerUnit(unit)
	if unit then
		return unit:IsThePlayer()
		--[[local name = unit:GetName() --.."@"..unitGetRealm()
		if name == self.unitPlayerName then
			return true
		end]]
	end
	return false
end

function SteadyPlates:OnUnitCreated(unitNew) 
	if unitNew then
		if unitNew:IsThePlayer() then --self:IsPlayerUnit(unitNew) then
			--Print("Created Player unit")
			uPlayerUnit = unitNew
		end
	end
end

function SteadyPlates:OnUnitDestroyed(unitDestroyed) 
	if unitDestroyed then
		--local tNameplate = aNameplates.arUnit2Nameplate[unitDestroyed:GetId()]
		--if tNameplate then
		--	tNameplate.wndNameplate:Show(false)
		--end
		if uPlayerUnit and unitDestroyed == uPlayerUnit then
			--Print("Destroyed Player unit")
			uPlayerUnit = nil
		end
	end
end

function SteadyPlates:OnUnitOcclusionChanged(wndHandler, wndControl, bOccluded)
	if self.wndPlayerWeapon and self.wndPlayerWeapon:GetId() == wndHandler:GetId() then
		if self.unitPlayerWeapon ~= nil then
			self.wndPlayerWeapon:Show(not bOccluded)
		end
	end
end

-----------------------------------------------------------------------------------------------
-- SteadyPlates Functions
-----------------------------------------------------------------------------------------------

function SteadyPlates:CheckWotW()
	if self.bWotWThrottle then
		if GameLib.GetCurrentWorldId() == 1393 and self.arThrottleUnitNames then
			CPrint("[SteadyPlates]: Throttle nameplates for Coldburrow units is now active!")
			self.bZoneIsWotW = true
		elseif self.bZoneIsWotW == true then
			CPrint("[SteadyPlates]: Throttle nameplates for Coldburrow units is now inactive!")
			self.bZoneIsWotW = false
		else
			self.bZoneIsWotW = false
		end
	elseif self.bZoneIsWotW and self.arThrottleUnitNames then
		CPrint("[SteadyPlates]: Throttle nameplates for Coldburrow units is now inactive!")
	else
		self.bZoneIsWotW = false
	end
end

function SteadyPlates.OnWorldLocationOnScreenHook(self, wndHandler, wndControl, bOnScreen)
	local tNameplate = self.arWnd2Nameplate[wndHandler:GetId()]
	if tNameplate ~= nil then
		tNameplate.bOnScreen = bOnScreen
		if bOnScreen then
			tNameplate.bEnsureHide = false
		end
		self:UpdateNameplateVisibility(tNameplate)
	end
end

function SteadyPlates.OnTargetUnitChangedHook(self, unitOwner)
	local oldTarget = nil
	if SteadyPlatesInst.bEnsureCombatNameplatesOnScreen or SteadyPlatesInst.bEnsureTargetNameplateOnScreen then
		local bResetAll = SteadyPlatesInst.bEnsureCombatNameplatesOnScreen and not bPlayerInCombat
		for _, tNameplateOther in pairs(aNameplates.arUnit2Nameplate) do
			if tNameplateOther.bEnsured and (bResetAll or tNameplateOther.bIsTarget) then
				SteadyPlates.HelperResetEnsure(tNameplateOther, SteadyPlatesInst.nEnsureMode)
			end
			if tNameplateOther.bIsTarget then
				oldTarget = tNameplateOther
			end
		end
	end
	
	fOnTargetUnitChanged(self, unitOwner)
	
	--if not SteadyPlatesInst.bRemoveHealthBarBackground then
		if oldTarget and not oldTarget.bIsTarget then
			SteadyPlates.HelperRepositionMarkers(oldTarget, false)
		end
		if unitOwner and GameLib.GetTargetUnit() == unitOwner then
			local tNameplate = self.arUnit2Nameplate[unitOwner:GetId()]
			if tNameplate then
				SteadyPlates.HelperRepositionMarkers(tNameplate, true)
			end
		end
	--end
end

function SteadyPlates.HelperRepositionMarkers(tNameplate, bIsTarget)
	local wnds = tNameplate.wnd
	local wnd = wnds.threatMarker
	if wnd and SteadyPlatesInst.bThreatMarkerLeft then
		local nLeft, nTop, nRight, nBottom
		local offsets = karThreatMarkerOffsets.Left
		if bIsTarget then
			nLeft, nTop, nRight, nBottom = offsets.nLeft - knMarkerSideOffset, offsets.nTop, offsets.nRight - knMarkerSideOffset, offsets.nBottom
		else
			nLeft, nTop, nRight, nBottom = offsets.nLeft, offsets.nTop, offsets.nRight, offsets.nBottom
		end
		wnd:SetAnchorOffsets(nLeft, nTop, nRight, nBottom)
	end
	wnd = wnds.interruptArmorMarker
	if wnd then
		if bIsTarget then
			wnd:SetAnchorOffsets(64 + knMarkerSideOffset +knMarkerSideOffsetMod, -1, 92 + knMarkerSideOffset +knMarkerSideOffsetMod, 27)
		else
			wnd:SetAnchorOffsets(64 +knMarkerSideOffsetMod, -1, 92 +knMarkerSideOffsetMod, 27)
		end
	end
	wnd = wnds.certainDeath
	if wnd and SteadyPlatesInst.bAdjustCertainDeath then
		if bIsTarget then
			wnd:SetAnchorOffsets(-18, -50, 18, -14)
		else
			wnd:SetAnchorOffsets(-18, -36, 18, 0)
		end
	end
end

function SteadyPlates.CheckDrawDistanceHook(self, tNameplate)
	local unitPlayer = self.unitPlayer
	local unitOwner = tNameplate.unitOwner
	
	if not unitOwner or not unitPlayer then
		return false
	end

	local tPosTarget = unitOwner:GetPosition()
	local tPosPlayer = unitPlayer:GetPosition()

	if tPosTarget == nil or tPosPlayer == nil then
		return false
	end

	local nDeltaX = tPosTarget.x - tPosPlayer.x
	local nDeltaY = tPosTarget.y - tPosPlayer.y
	local nDeltaZ = tPosTarget.z - tPosPlayer.z

	local nDistance = (nDeltaX * nDeltaX) + (nDeltaY * nDeltaY) + (nDeltaZ * nDeltaZ)
	
	if tNameplate.bIsTarget or tNameplate.bIsCluster or tNameplate.bIsMouseOver then
		return (nDistance < knTargetRange)
	else
		nMaxRange = self.nMaxRange
		return (nDistance < (nMaxRange * nMaxRange)) -- squaring for quick maths
	end
end

function SteadyPlates.OnUnitOcclusionChangedHook(self, wndHandler, wndControl, bOccluded)
	if self.bUseOcclusion then
		fNameplatesOnUnitOcclusionChanged(self, wndHandler, wndControl, bOccluded)
	end
end

function SteadyPlates.OnUnitCreatedHook(self, unitNew) 
	if unitNew == nil then
		return
	end
	
	local unitType = unitNew:GetType()
	if not unitNew:IsValid()
		--or not unitNew:ShouldShowNamePlate()
		or unitType == "Collectible"
		or unitType == "PinataLoot"
	then
		-- Never have nameplates
		return
	end
	
	local bIsMe = unitNew:IsThePlayer() 
	if bIsMe then --self:IsPlayerUnit(unitNew) then
		--Print("Created Player unit")
		uPlayerUnit = unitNew
	end

	if uPlayerUnit and unitType == "Pickup" and self.strWeaponUnitNamePattern then
		local playerName = uPlayerUnit:GetName()
		if string.match(unitNew:GetName(), self.strWeaponUnitNamePattern) == playerName then
			self.unitPlayerWeapon = unitNew
			if self.bShowPlayerWeaponIcon then
				if self.wndPlayerWeapon == nil then
					self.wndPlayerWeapon = Apollo.LoadForm(self.xmlDoc, "WeaponIcon", "InWorldHudStratum", self)
				end
				self.wndPlayerWeapon:Show(true, true)
				self.wndPlayerWeapon:SetUnit(unitNew)
			end
			if not self.bShowPlayerWeaponNameplate then
				return
			end
		end
	end
	
	local idUnit = unitNew:GetId()
	if self.arUnit2Nameplate[idUnit] ~= nil and self.arUnit2Nameplate[idUnit].wndNameplate:IsValid() then
		return
	end
	
	local wnd = nil
	local wndReferences = nil
	if next(self.arWindowPool) ~= nil then
		local poolEntry = table.remove(self.arWindowPool)
		wnd = poolEntry[1]
		wndReferences = poolEntry[2]
		wnd:SetAnchorOffsets(knDefaultLeft, knDefaultTop, knDefaultRight, knDefaultBottom)
	end

	if wnd == nil or not wnd:IsValid() then
		wnd = Apollo.LoadForm(self.xmlDoc, "NameplateNew", "InWorldHudStratum", self)
		wndReferences = nil
	end

	wnd:Show(false, true)
	if bIsMe and SteadyPlatesInst.bAnchorMyNameplateToFeet then
		wnd:SetUnit(unitNew, 0)
	else
		wnd:SetUnit(unitNew, 1)
	end

	local strNewUnitType = unitNew:GetType()
	local uMatchMakingEntry = MatchingGameLib.GetQueueEntry()
	local tNameplate =
	{
		unitOwner 		= unitNew,
		idUnit 			= idUnit,
		wndNameplate	= wnd,
		bOnScreen 		= wnd:IsOnScreen(),
		bOccluded 		= wnd:IsOccluded(),
		bSpeechBubble 	= false,
		bIsTarget 		= GameLib.GetTargetUnit() == unitNew,
		bIsCluster 		= false,
		bIsCasting 		= false,
		bGibbed			= false,
		bIsPlayer		= strNewUnitType == "Player",
		bShowPvpMatch	= uMatchMakingEntry and uMatchMakingEntry:IsPvpGame() and (strNewUnitType == "Player" or strNewUnitType == "Esper Pet" or strNewUnitType == "Pet"),
		bIsGuildMember 	= self.guildDisplayed and self.guildDisplayed:IsUnitMember(unitNew) or false,
		bIsWarPartyMember = self.guildWarParty and self.guildWarParty:IsUnitMember(unitNew) or false,
		nVulnerableTime = 0,
		nCon			= self:HelperCalculateConValue(unitOwner),
		eDisposition	= unitNew:GetDispositionTo(self.unitPlayer),
		strUnitType		= strNewUnitType,
		tActivation		= unitNew:GetActivationState(),
		tRewardInfo		= unitNew:GetRewardInfo(),
		bShow			= false,
		wnd				= wndReferences,
		-- Window visibility
		bShowHealth		= true,
	}

	if wndReferences == nil then
		tNameplate.wnd =
		{
			health = wnd:FindChild("Container:Health"),
			castBar = wnd:FindChild("Container:CastBar"),
			vulnerable = wnd:FindChild("Container:Vulnerable"),
			level = wnd:FindChild("Container:Health:Level"),
			wndGuild = wnd:FindChild("Guild"),
			wndName = wnd:FindChild("NameRewardContainer:Name"),
			certainDeath = wnd:FindChild("TargetAndDeathContainer:CertainDeath"),
			targetScalingMark = wnd:FindChild("TargetScalingMark"),
			nameRewardContainer = wnd:FindChild("NameRewardContainer:RewardContainer"),
			--healthMaxShield = wnd:FindChild("Container:Health:HealthBars:MaxShield"),
			--healthShieldFill = wnd:FindChild("Container:Health:HealthBars:MaxShield:ShieldFill"),
			--healthMaxAbsorb = wnd:FindChild("Container:Health:HealthBars:MaxAbsorb"),
			--healthAbsorbFill = wnd:FindChild("Container:Health:HealthBars:MaxAbsorb:AbsorbFill"),
			--healthMaxHealth = wnd:FindChild("Container:Health:HealthBars:MaxHealth"),
			healthHealthLabel = wnd:FindChild("Container:Health:HealthLabel"),
			castBarLabel = wnd:FindChild("Container:CastBar:Label"),
			castBarCastFill = wnd:FindChild("Container:CastBar:CastFill"),
			vulnerableVulnFill = wnd:FindChild("Container:Vulnerable:VulnFill"),
			questRewards = wnd:FindChild("NameRewardContainer:RewardContainer:QuestRewards"),
			targetMarkerArrow = wnd:FindChild("TargetAndDeathContainer:TargetMarkerArrow"),
			targetMarker = wnd:FindChild("Container:TargetMarker"),
			backgroundPlate = wnd:FindChild("Container:Health:BackgroundPlate"),
			
			healthBar = wnd:FindChild("Container:Health:HealthBars:Health"),
			healingAbsorbBar = wnd:FindChild("Container:Health:HealthBars:HealingAbsorb"),
			absorbBar = wnd:FindChild("Container:Health:HealthBars:Absorb"),
			shieldBar = wnd:FindChild("Container:Health:HealthBars:Shield"),
			healthClampMin = wnd:FindChild("Container:Health:HealthBars:HealthClampMin"),
			healthClampMax = wnd:FindChild("Container:Health:HealthBars:HealthClampMax"),
		}
	end
	
	--[[
	local tNameplate =
	{
		unitOwner 		= unitNew,
		idUnit 			= idUnit,
		wndNameplate	= wnd,
		bOnScreen 		= wnd:IsOnScreen(),
		bOccluded 		= wnd:IsOccluded(),
		bSpeechBubble 	= false,
		bIsTarget 		= false,
		bIsCluster 		= false,
		bIsCasting 		= false,
		bGibbed			= false,
		bIsGuildMember 	= self.guildDisplayed and self.guildDisplayed:IsUnitMember(unitNew) or false,
		bIsWarPartyMember = self.guildWarParty and self.guildWarParty:IsUnitMember(unitNew) or false,
		nVulnerableTime = 0,
		eDisposition	= unitNew:GetDispositionTo(self.unitPlayer),
		bShow			= false,
		wnd				= wndReferences,
	}

	if wndReferences == nil then
		tNameplate.wnd =
		{
			health = wnd:FindChild("Container:Health"),
			castBar = wnd:FindChild("Container:CastBar"),
			vulnerable = wnd:FindChild("Container:Vulnerable"),
			level = wnd:FindChild("Container:Health:Level"),
			wndGuild = wnd:FindChild("Guild"),
			wndName = wnd:FindChild("NameRewardContainer:Name"),
			premiumIcon = wnd:FindChild("NameRewardContainer:PremiumIcon"),
			certainDeath = wnd:FindChild("TargetAndDeathContainer:CertainDeath"),
			targetScalingMark = wnd:FindChild("TargetScalingMark"),
			nameRewardContainer = wnd:FindChild("NameRewardContainer:RewardContainer"),
			healthMaxShield = wnd:FindChild("Container:Health:HealthBars:MaxShield"),
			healthShieldFill = wnd:FindChild("Container:Health:HealthBars:MaxShield:ShieldFill"),
			healthMaxAbsorb = wnd:FindChild("Container:Health:HealthBars:MaxAbsorb"),
			healthAbsorbFill = wnd:FindChild("Container:Health:HealthBars:MaxAbsorb:AbsorbFill"),
			healthMaxHealth = wnd:FindChild("Container:Health:HealthBars:MaxHealth"),
			healthHealthLabel = wnd:FindChild("Container:Health:HealthLabel"),
			castBarLabel = wnd:FindChild("Container:CastBar:Label"),
			castBarCastFill = wnd:FindChild("Container:CastBar:CastFill"),
			vulnerableVulnFill = wnd:FindChild("Container:Vulnerable:VulnFill"),
			questRewards = wnd:FindChild("NameRewardContainer:RewardContainer:QuestRewards"),
			targetMarkerArrow = wnd:FindChild("TargetAndDeathContainer:TargetMarkerArrow"),
			targetMarker = wnd:FindChild("Container:TargetMarker"),
		}
	end
	]]

	if tNameplate.bShowPvpMatch then
		tNameplate.wnd.certainDeath:SetSprite(ksprPvpTarget)
	else
		tNameplate.wnd.certainDeath:SetSprite(ksprHighLevel)
	end

	if not tNameplate.bIsPlayer and tNameplate.wnd.premiumIcon then
		tNameplate.wnd.premiumIcon:Show(false)
	end
	
	SteadyPlates.UpdateNameplateToSettings(tNameplate, tNameplate.wnd, true)
	
	self.arUnit2Nameplate[idUnit] = tNameplate
	self.arWnd2Nameplate[wnd:GetId()] = tNameplate

	self:DrawName(tNameplate)
	self:DrawGuild(tNameplate)
	self:DrawLevel(tNameplate)
	self:UpdateNameplateRewardInfo(tNameplate)
	self:DrawRewards(tNameplate)

	--[[if unitNew then
		fNameplatesOnUnitCreated(self, unitNew)
		local tNameplate = self.arUnit2Nameplate[unitNew:GetId()]
		if tNameplate then
			SteadyPlates.UpdateNameplateToSettings(tNameplate, tNameplate.wnd, true)
		end
	end]]
end

function SteadyPlates.OnUnitDestroyedHook(self, unitOwner)
	if uPlayerUnit and unitOwner == uPlayerUnit then
		--Print("Destroyed Player unit")
		uPlayerUnit = nil
	end
	
	if self.unitPlayerWeapon == unitOwner then
		self.unitPlayerWeapon = nil
		if self.wndPlayerWeapon ~=nil then
			self.wndPlayerWeapon:Show(false, true)
		end
	end
	
	local idUnit = unitOwner:GetId()
	local tNameplate = self.arUnit2Nameplate[idUnit]
	if tNameplate == nil then
		return
	end
	fNameplatesOnUnitDestroyed(self, unitOwner)
end

function SteadyPlates.UpdateNameplateRewardInfoHook(self, tNameplate)
	fNameplatesUpdateNameplateRewardInfo(self, tNameplate)
	
	tNameplate.bIsObjective = false
	
	local data = tNameplate.wnd.questRewards:GetData()
	if not data then return end
	
	local tRewardInfo = data.tRewardInfo
	if not tRewardInfo then return end
	
	if tRewardInfo then
		for _,v in ipairs(tRewardInfo) do
		--for k=1, #tRewardInfo do 
			--local v = tRewardInfo[k]
			if v.strType and v.strType == "Quest" and v.nCompleted and v.nCompleted < v.nNeeded then
				tNameplate.bIsObjective = true
				break
			end
		end
	end
end

function SteadyPlates.UpdateNameplateVisibilityHook(self, tNameplate)
	local unitOwner = tNameplate.unitOwner
	local wndNameplate = tNameplate.wndNameplate
	local bIsMounted = unitOwner:IsMounted()
	local unitWindow = wndNameplate:GetUnit()
	
	local nTo = 1
	if SteadyPlatesInst.bAnchorMyNameplateToFeet and tNameplate.bIsMe then
		nTo = 0
	end
	
	if bIsMounted and unitWindow == unitOwner then
		if not tNameplate.bMounted then
			wndNameplate:SetUnit(unitOwner:GetUnitMount(), nTo)
			tNameplate.bMounted = true
		end
	elseif not bIsMounted and unitWindow ~= unitOwner then
		if tNameplate.bMounted then
			wndNameplate:SetUnit(unitOwner, nTo)
			tNameplate.bMounted = false
		end
	end
	
	local eDisposition = unitOwner:GetDispositionTo(self.unitPlayer)
	local nCon = fnHelperCalculateConValue(self, unitOwner)
	
	tNameplate.bOnScreen = wndNameplate:IsOnScreen()
	tNameplate.bOccluded = wndNameplate:IsOccluded()
	
	tNameplate.eDisposition = unitOwner:GetDispositionTo(self.unitPlayer)
	local bNewShow = self:HelperVerifyVisibilityOptions(tNameplate) and self:CheckDrawDistance(tNameplate)
	
	if bNewShow then
		if tNameplate.bIsMouseOver or (SteadyPlatesInst.bEnsureCombatNameplatesOnScreen and bPlayerInCombat) or (tNameplate.bIsTarget and SteadyPlatesInst.bEnsureTargetNameplateOnScreen) then
			fnHelperEnsureNameplateOnScreen(self, tNameplate, SteadyPlatesInst.nEnsureMode)
			if tNameplate.bEnsureHide then
				bNewShow = false
			end
		end
	end
	
	if bNewShow ~= tNameplate.bShow then
		wndNameplate:Show(bNewShow)
		tNameplate.bShow = bNewShow
	end
	
	if bNewShow or self.bRedrawRewardIcons or tNameplate.eDisposition ~= eDisposition then
		local wnd = tNameplate.wnd
		
		-- Disposition based update
		--[[if eDisposition ~= tNameplate.eDisposition then
			wnd.targetMarkerArrow:SetSprite(karDisposition.tTargetSecondary[tNameplate.eDisposition])
			wnd.targetMarker:SetSprite(karDisposition.tTargetPrimary[tNameplate.eDisposition])
		end]]

		-- Handle Certain Death icon
		--[[local bShowCertainDeath = self.bShowCertainDeathMain -- Display if user preference is to display
			and tNameplate.eDisposition ~= Unit.CodeEnumDisposition.Friendly -- Display for hostile/neutral units
			and (tNameplate.bShowPvpMatch or unitOwner:IsCertainDeath()) -- Always show in PVP, or if Certain Death
			and unitOwner:ShouldShowNamePlate() -- Display if unit flagged to display
			and unitOwner:GetHealth() -- Display if unit has health value
			and not unitOwner:IsDead() -- Display if unit is not dead

		-- Only show/hide certain death icon if the state has changed
		if bShowCertainDeath ~= wnd.certainDeath:IsShown() then
			wnd.certainDeath:Show(bShowCertainDeath)
		end]]
		
		fnDrawCertainDeath(self, tNameplate, nCon)

		-- Does not need to update every frame
		local bShowScaled = unitOwner:IsScaled()
		if bShowScaled ~= wnd.targetScalingMark:IsShown() then
			wnd.targetScalingMark:Show(bShowScaled)
		end

		fnColorNameplate(self, tNameplate)

		fnDrawHealth(self, tNameplate)
		fnDrawTargeting(self, tNameplate)

		fnDrawRewards(self, tNameplate)
		
		wnd.targetScalingMark:Show(unitOwner:IsScaled())

		fnDrawRewards(self, tNameplate)
		fnDrawCastBar(self, tNameplate)
		fnDrawVulnerable(self, tNameplate)
		fnColorNameplate(self, tNameplate, nCon)
		
		fnDrawTargeting(self, tNameplate)
		fnDrawThreat(self, tNameplate)
		fnDrawInterruptArmor(self, tNameplate)
		fnDrawCleanse(self, tNameplate)
		fnDrawCluster(self, tNameplate)
		fnHelperChangeHealthText(self, tNameplate)
		--SteadyPlates.HelperScaleNameplate(self, tNameplate)
	end

	tNameplate.eDisposition = eDisposition
	tNameplate.nCon = nCon
	
	--fUpdateNameplateVisibility(self, tNameplate)
end

function SteadyPlates.DrawNameHook(self, tNameplate)
	local bIsTarget = tNameplate.bIsTarget
	if not bIsTarget and SteadyPlatesInst.bHideMyNameWhenNotTargeted and tNameplate.bIsMe then
		tNameplate.wnd.wndName:Show(false)
	else
		if not bIsTarget and SteadyPlatesInst.bHideTitle and self.bShowTitle then
			self.bShowTitle = false
			fDrawName(self, tNameplate)	
			self.bShowTitle = true
		else
			fDrawName(self, tNameplate)	
		end
	end
	
	if SteadyPlatesInst.bShowSmallClassIconLeftToName and tNameplate.bIsPlayer and not tNameplate.bIsMe then
		local classIcon = tNameplate.wnd.classIcon
		if not classIcon then
			classIcon = Apollo.LoadForm(SteadyPlatesInst.xmlDoc, "ThreatMarker", tNameplate.wnd.wndName, self)
			tNameplate.wnd.classIcon = classIcon
			classIcon:SetAnchorPoints(0.5, 0, 0.5, 0)
		end		
		local nHalfNameWidth = math.ceil(Apollo.GetTextWidth("Nameplates", tNameplate.wnd.wndName:GetText()) / 2)
		classIcon:SetAnchorOffsets(-18 -nHalfNameWidth, -1, 0 -nHalfNameWidth, 17)
		classIcon:SetSprite(karClassIcons[tNameplate.unitOwner:GetClassId() or 0])
		classIcon:Show(true)
	elseif tNameplate.wnd.classIcon then
		tNameplate.wnd.classIcon:Show(false)
	end
end

function SteadyPlates.DrawGuildHook(self, tNameplate)
	if SteadyPlatesInst.bHideMyNameWhenNotTargeted and not tNameplate.bIsTarget and tNameplate.bIsMe then
		tNameplate.wnd.wndGuild:Show(false)
	else
		fDrawGuild(self, tNameplate)		
	end
end

function SteadyPlates.DrawCastBarHook(self, tNameplate)
	local bShowCastBarTargetOnly, bShowCastBarName = false, true
	local unitOwner = tNameplate.unitOwner
	local bShow = unitOwner:ShouldShowCastBar()
	if bShow then
		if tNameplate.bIsFriendly then
			bShowCastBarTargetOnly, bShowCastBarName = SteadyPlatesInst.bShowCastBarTargetOnlyFriendly, SteadyPlatesInst.bShowCastBarNameFriendly
		elseif tNameplate.bIsNeutral then
			bShowCastBarTargetOnly, bShowCastBarName = SteadyPlatesInst.bShowCastBarTargetOnlyNeutral, SteadyPlatesInst.bShowCastBarNameNeutral
		elseif tNameplate.bIsHostile then
			bShowCastBarTargetOnly, bShowCastBarName = SteadyPlatesInst.bShowCastBarTargetOnlyHostile, SteadyPlatesInst.bShowCastBarNameHostile
		end
	
		if tNameplate.bIsMe then
			bShow = not SteadyPlatesInst.bHideMyCastBar
		elseif bShowCastBarTargetOnly and not tNameplate.bIsTarget then
			bShow = false
		end
	end
	if bShow then
		fDrawCastBar(self, tNameplate)
		tNameplate.wnd.castBarLabel:Show(bShowCastBarName)
		--[[if bShowCastBarName then
			local wndCastLabel = tNameplate.wnd.castBarLabel
			wndCastLabel:Show(true)
			local nLeft, nTop, nRight, nBottom = wndCastLabel:GetAnchorOffsets()
			local nHalfNameWidth = math.ceil(Apollo.GetTextWidth("Nameplates", wndCastLabel:GetText()) / 2)
			if nHalfNameWidth > nRight then
				wndCastLabel:SetAnchorOffsets(-nHalfNameWidth, nTop, nHalfNameWidth, nBottom)
			end
		else
			tNameplate.wnd.castBarLabel:Show(false)
		end]]
	else
		tNameplate.bIsCasting = false
		tNameplate.wnd.castBar:Show(false)
	end
end

function SteadyPlates.DrawHealthHook(self, tNameplate)
	local wndHealth = tNameplate.wnd.health
	if tNameplate.bIsTarget or tNameplate.bIsMouseOver
		or (not SteadyPlatesInst.bShowHealthBarTargetOnlyFriendly and tNameplate.bIsFriendly)
		or (not SteadyPlatesInst.bShowHealthBarTargetOnlyNeutral and tNameplate.bIsNeutral)
		or (not SteadyPlatesInst.bShowHealthBarTargetOnlyHostile and tNameplate.bIsHostile)
	then
		local wndNameplate = tNameplate.wndNameplate
		local unitOwner = tNameplate.unitOwner
		
		local nHealth, nMaxHealth = unitOwner:GetHealth(), unitOwner:GetMaxHealth()
		if nHealth == nil or (nHealth == 0 and nMaxHealth == 0) then
			wndHealth:Show(false)
			return
		end
		
		local bUseTarget = tNameplate.bIsTarget or tNameplate.bIsMouseOver
		local bShow
		if bUseTarget then
			bShow = self.bShowHealthTarget
		else
			bShow = self.bShowHealthMain or (self.bShowHealthMainDamaged and nHealth ~= nMaxHealth)
		end
		
		if bShow then
			if SteadyPlatesInst.bDimmHealthInCombatFriendly then
				if bPlayerInCombat and tNameplate.bIsFriendly and not bUseTarget and not tNameplate.bIsMe then
					local nOpacity = 1 - nHealth/nMaxHealth
					if nOpacity < 0.05 then 
						wndHealth:Show(false)
						return
					else
						--if nOpacity > 0.8 then nOpacity = 1 end
						wndHealth:SetOpacity(nOpacity)
					end
				else
					wndHealth:SetOpacity(1)
				end
			end
			self:DrawHealthShieldBar(wndHealth, unitOwner, tNameplate.eDisposition, tNameplate)
		end
		
		wndHealth:Show(bShow)
	else
		wndHealth:Show(false)
	end
	tNameplate.bShowHealth = bShow
end

function SteadyPlates.OnPreloadUnitCreateTimerHook(self)
	if self.arPreloadUnits == nil then
		if self.timerPreloadUnitCreateDelay then
			self.timerPreloadUnitCreateDelay:Stop()
			self.timerPreloadUnitCreateDelay = nil
		end
	else
		fOnPreloadUnitCreateTimer(self)
	end
end

function SteadyPlates.ColorNameplateHook(self, tNameplate, nCon)
	local unitPlayer = self.unitPlayer
	local unitOwner = tNameplate.unitOwner
	local wndNameplate = tNameplate.wndNameplate

	local eDisposition = tNameplate.eDisposition
	if nCon == nil then
		nCon = SteadyPlates.HelperCalculateConValueHook(self, unitOwner)
	end

	local crLevelColorToUse = karConColors[nCon][2]
	if tNameplate.wnd.targetScalingMark:IsShown() then
		crLevelColorToUse = kcrScalingCColor
	elseif unitOwner:GetLevel() == nil then
		crLevelColorToUse = karConColors[1][2]
	end

	local crColorToUse
	
	if unitOwner:IsDead() then
			crColorToUse = kcrDeadColor
			crLevelColorToUse = kcrDeadColor
	else
		if SteadyPlatesInst.bOverwriteNameColor then
			crColorToUse = SteadyPlates.HelperGetColorFromColorSets("Name", unitOwner, eDisposition, tNameplate)
		end
		if not crColorToUse then
			crColorToUse = karDisposition.tTextColors[eDisposition]
			if tNameplate.bIsPlayer or tNameplate.bIsPet then
				if tNameplate.bIsFriendly or tNameplate.bIsMe then
					crColorToUse = kcrUnflaggedAlly
					if tNameplate.bIsPvpFlagged then
						crColorToUse = kcrFlaggedAlly
					elseif tNameplate.bIsGroupMember then
						crColorToUse = kcrUnflaggedGroupmate
					elseif tNameplate.bIsGuildMember then
						crColorToUse = kcrUnflaggedGuildmate
					end
				else
					local bIsUnitFlagged = tNameplate.bIsPvpFlagged
					local bAmIFlagged = GameLib.IsPvpFlagged()

					if not bAmIFlagged and not bIsUnitFlagged then
						crColorToUse = kcrUnflaggedEnemyWhenUnflagged
					elseif bAmIFlagged and not bIsUnitFlagged then
						crColorToUse = kcrUnflaggedEnemyWhenFlagged
					elseif not bAmIFlagged and bIsUnitFlagged then
						crColorToUse = kcrFlaggedEnemyWhenUnflagged
					elseif bAmIFlagged and bIsUnitFlagged then
						crColorToUse = kcrFlaggedEnemyWhenFlagged
					end
				end
			end
		end
		if not tNameplate.bIsPlayer and not tNameplate.bIsFriendly and unitOwner:IsTagged() and not unitOwner:IsTaggedByMe() and not unitOwner:IsSoftKill() then
			crColorToUse = kcrDefaultTaggedColor
		end
	end

	tNameplate.wnd.level:SetTextColor(crLevelColorToUse)
	tNameplate.wnd.wndName:SetTextColor(crColorToUse)
	tNameplate.wnd.wndGuild:SetTextColor(crColorToUse)
end

function SteadyPlates.OnFrameHook(self)
	if uPlayerUnit then
		self.unitPlayer = uPlayerUnit
	else
		self.unitPlayer = GameLib.GetPlayerUnit()
	end

	local unitOwner
	local arThrottle = SteadyPlatesInst.arThrottleUnitNames
	local bThrottle = SteadyPlatesInst.bWotWThrottle and SteadyPlatesInst.bZoneIsWotW and arThrottle
	--local bEnsureAll = SteadyPlatesInst.bEnsureCombatNameplatesOnScreen and bPlayerInCombat
	--local bEnsureTarget = SteadyPlatesInst.bEnsureTargetNameplateOnScreen
	--local nEnsureMode = SteadyPlatesInst.nEnsureMode
	
	for idx, tNameplate in pairs(self.arUnit2Nameplate) do
	--local arNameplates = self.arUnit2Nameplate
	--for idx=1, #arNameplates do 
		--local tNameplate = arNameplates[idx]
		if tNameplate.bShow then
			
			if bThrottle then
				unitOwner = tNameplate.unitOwner
				if unitOwner ~= nil and not (tNameplate.bIsTarget or tNameplate.bIsMouseOver) then
					local name = unitOwner:GetName()
					for _,v in pairs(arThrottle) do
					--for k=1, #arThrottle do 
						--local v = arThrottle[k]
						if name == v then
							return
						end
					end
				end
			end
			
			--if bEnsureAll or (bEnsureTarget and tNameplate.bIsTarget) or tNameplate.bIsMouseOver then 
			--	fnHelperEnsureNameplateOnScreen(self, tNameplate, nEnsureMode)
			--else
			if bPerspectivePlates and not tNameplate.bIsMe then
				Event_FireGenericEvent("GenericEvent_PerspectivePlates_PerspectiveResize", tNameplate, 1, {	left = knDefaultLeft,	top = knDefaultTop, right = knDefaultRight,	bottom = knDefaultBottom})
			end
	
			if tNameplate.bShowHealth then
				fnDrawHealth(self, tNameplate)
			end
		end
	end
end

function SteadyPlates.DrawNameplate(self, tNameplate) -- deprecated for normal updating (OnFrame), but still used when mouse over!
	if tNameplate.bShow then
		local unitOwner = tNameplate.unitOwner
		local wnds = tNameplate.wnd

		if SteadyPlatesInst.bWotWThrottle and SteadyPlatesInst.bZoneIsWotW and SteadyPlatesInst.arThrottleUnitNames then
			if unitOwner ~= nil and not (tNameplate.bIsTarget or tNameplate.bIsMouseOver) then
				local name = unitOwner:GetName()
				local arThrottle = SteadyPlatesInst.arThrottleUnitNames
				for _,v in pairs(arThrottle) do
				--for k=1, #arThrottle do 
					--local v = arThrottle[k]
					if name == v then
						return
					end
				end
			end
		end
				
		if tNameplate.bIsMouseOver or (SteadyPlatesInst.bEnsureCombatNameplatesOnScreen and bPlayerInCombat) or (tNameplate.bIsTarget and SteadyPlatesInst.bEnsureTargetNameplateOnScreen) then
			fnHelperEnsureNameplateOnScreen(self, tNameplate, SteadyPlatesInst.nEnsureMode)
		elseif bPerspectivePlates and not tNameplate.bIsMe then
			Event_FireGenericEvent("GenericEvent_PerspectivePlates_PerspectiveResize", tNameplate, 1, {	left = knDefaultLeft,	top = knDefaultTop, right = knDefaultRight,	bottom = knDefaultBottom})
		end

		SteadyPlates.DrawHealthHook(self, tNameplate)
		
		local nCon = SteadyPlates.HelperCalculateConValueHook(self, unitOwner)
		
		SteadyPlates.DrawCertainDeath(self, tNameplate, nCon)
		wnds.targetScalingMark:Show(unitOwner:IsScaled())

		self.DrawRewards(self, tNameplate)
		SteadyPlates.DrawCastBarHook(self, tNameplate)
		self.DrawVulnerable(self, tNameplate)
		SteadyPlates.ColorNameplateHook(self, tNameplate, nCon)

		SteadyPlates.DrawTargeting(self, tNameplate)
		SteadyPlates.DrawThreat(self, tNameplate)
		SteadyPlates.DrawInterruptArmor(self, tNameplate)
		SteadyPlates.DrawCleanse(self, tNameplate)
		SteadyPlates.DrawCluster(self, tNameplate)
		SteadyPlates.HelperChangeHealthText(self, tNameplate)
		--SteadyPlates.HelperScaleNameplate(self, tNameplate)
	end
end

function SteadyPlates.DrawInterruptArmor(self, tNameplate)
	local interruptArmorMarker = tNameplate.wnd.interruptArmorMarker
	
	if not SteadyPlatesInst.bShowInterruptArmorMarker or not tNameplate.wnd.health:IsShown() then
		if interruptArmorMarker then interruptArmorMarker:Show(false) end
		return
	end
	
	local unitOwner = tNameplate.unitOwner
	local nArmor = unitOwner:GetInterruptArmorValue() or 0
	local nArmorMax = unitOwner:GetInterruptArmorMax()
	local bIsInfinite = nArmorMax == -1
	--local bIsBroken = nArmor == 0 and nArmorMax > 0
	local bHasArmor = nArmorMax > 0
	local bAlwaysShowClass = SteadyPlatesInst.bInterruptArmorMarkerUsePlayerClass and tNameplate.bIsPlayer and not tNameplate.bIsMe
	
	if not interruptArmorMarker and (bHasArmor or bIsInfinite or bAlwaysShowClass) then
		local wndParent = tNameplate.wndNameplate:FindChild("Container")
		
		interruptArmorMarker = Apollo.LoadForm(SteadyPlatesInst.xmlDoc, "InterruptArmorMarker", wndParent, self)
		if interruptArmorMarker then
			local strSprite = SteadyPlatesInst.strInterruptArmorMarkerSprite
			if bAlwaysShowClass then
				local nClass = unitOwner:GetClassId() or 0
				strSprite = karClassIcons[nClass]
			end
			interruptArmorMarker:SetSprite(strSprite)
			tNameplate.wnd.interruptArmorMarker = interruptArmorMarker
			--if tNameplate.bIsTarget and not SteadyPlatesInst.bRemoveHealthBarBackground then
			if tNameplate.bIsTarget then
				interruptArmorMarker:SetAnchorOffsets(64 + knMarkerSideOffset +knMarkerSideOffsetMod, -1, 92 + knMarkerSideOffset +knMarkerSideOffsetMod, 27)
			else
				interruptArmorMarker:SetAnchorOffsets(64 +knMarkerSideOffsetMod, -1, 92 +knMarkerSideOffsetMod, 27)
			end
			interruptArmorMarker:Show(true)
		end
	end
	
	if interruptArmorMarker then
		if bIsInfinite then
			interruptArmorMarker:SetText("X")
			interruptArmorMarker:SetTextColor("xkcdNeonBlue")
			interruptArmorMarker:Show(true)
		--elseif bIsBroken then
			--interruptArmorMarker:SetText("")
			--interruptArmorMarker:Show(true)
		elseif bHasArmor then
			interruptArmorMarker:SetText(nArmor)
			if tNameplate.bIsCasting then
				interruptArmorMarker:SetTextColor("yellow")
			else
				interruptArmorMarker:SetTextColor("xkcdPastelOrange")
			end
			interruptArmorMarker:Show(true)
		elseif bAlwaysShowClass then
			interruptArmorMarker:SetText("")
			interruptArmorMarker:Show(true)
		else
			interruptArmorMarker:Show(false)
		end
	end
end

function SteadyPlates.DrawThreat(self, tNameplate)
	local threatMarker = tNameplate.wnd.threatMarker
	local unitOwner = tNameplate.unitOwner
	local bIsInGroup = GroupLib.InGroup() or GroupLib.InRaid()
	
	if not SteadyPlatesInst.bShowThreatMarker or not tNameplate.bIsHostile or tNameplate.bIsMe or (SteadyPlatesInst.bShowThreatMarkerGroupOnly and not bIsInGroup) or (SteadyPlatesInst.bShowThreatMarkerInstanceOnly and not GroupLib.InInstance()) then
		if threatMarker then threatMarker:Show(false) end
		return
	end
	
	local unitTarget = unitOwner:GetTarget()
	local bHasThreat = false
	if unitTarget then
		if SteadyPlatesInst.bThreatMarkerTankMode and bIsInGroup then
			bHasThreat = unitTarget:IsInYourGroup() and unitOwner:IsInCombat() and not unitTarget:IsThePlayer()
		else
			bHasThreat = unitTarget:IsThePlayer()
		end
	end
	
	if not threatMarker and bHasThreat then
		local wndParent
		if SteadyPlatesInst.bThreatMarkerLeft then
			wndParent = tNameplate.wndNameplate:FindChild("Container")
		else
			wndParent = tNameplate.wndNameplate:FindChild("TargetAndDeathContainer") --tNameplate.wndNameplate:FindChild("NameRewardContainer")
		end
		
		threatMarker = Apollo.LoadForm(SteadyPlatesInst.xmlDoc, "ThreatMarker", wndParent, self)
		if threatMarker then
			threatMarker:SetSprite(SteadyPlatesInst.strThreatMarkerSprite)
			tNameplate.wnd.threatMarker = threatMarker
			local nLeft, nTop, nRight, nBottom
			if SteadyPlatesInst.bThreatMarkerLeft then
				local offsets = karThreatMarkerOffsets.Left
				--if tNameplate.bIsTarget and not SteadyPlatesInst.bRemoveHealthBarBackground then
				if tNameplate.bIsTarget then
					nLeft, nTop, nRight, nBottom = offsets.nLeft - knMarkerSideOffset, offsets.nTop, offsets.nRight - knMarkerSideOffset, offsets.nBottom
				else
					nLeft, nTop, nRight, nBottom = offsets.nLeft, offsets.nTop, offsets.nRight, offsets.nBottom
				end
			else
				local offsets = karThreatMarkerOffsets.Top
				nLeft, nTop, nRight, nBottom = offsets.nLeft, offsets.nTop, offsets.nRight, offsets.nBottom
			end
			threatMarker:SetAnchorOffsets(nLeft, nTop, nRight, nBottom)
		end
	end
	
	if threatMarker then
		if bHasThreat then
			threatMarker:Show(true)
		else
			threatMarker:Show(false)
		end
	end
end

function SteadyPlates.DrawCertainDeath(self, tNameplate, nCon)
	local certainDeathMarker = tNameplate.wnd.certainDeath
	
	if not self.bShowCertainDeathMain or not self.unitPlayer then
		if certainDeathMarker then certainDeathMarker:Show(false) end
		return
	end
	
	local bShow = false
	local unitOwner = tNameplate.unitOwner
	
	if not tNameplate.bIsMe and tNameplate.eDisposition ~= Unit.CodeEnumDisposition.Friendly and unitOwner:GetHealth() and unitOwner:ShouldShowNamePlate() and not unitOwner:IsDead() then
		--local bIsElite = unitOwner:IsElite()
		if nCon == nil then
			nCon = SteadyPlates.HelperCalculateConValueHook(self, unitOwner)
		end

		-- level based check
		bShow = nCon == knConColors
		
		-- elite/rank based check
		if not bShow and SteadyPlatesInst.bExtendCertainDeath then
			local nRank = unitOwner:GetRank()
			local nUnitCon = self.unitPlayer:GetLevelDifferential(unitOwner) or 0
			local bIsDangerous = nUnitCon > -10 and (nRank == Unit.CodeEnumRank.Elite or unitOwner:GetEliteness() > 0)
			bShow =  bIsDangerous and not (GroupLib.InGroup() or GroupLib.InRaid()) and not GroupLib.InInstance() -- or (bPlayerInCombat and unitOwner:IsInCombat()))
		end
	end
		
	if bShow then
		certainDeathMarker:Show(true)
	else
		certainDeathMarker:Show(false)
	end
end

function SteadyPlates.DrawCleanse(self, tNameplate)
	local cleanseMarker = tNameplate.wnd.cleanseMarker
	
	if not SteadyPlatesInst.bShowCleanseMarker or tNameplate.eDisposition ~= Unit.CodeEnumDisposition.Friendly then
		if cleanseMarker then cleanseMarker:Show(false) end
		return
	end
	
	local unitOwner = tNameplate.unitOwner
	
	if not tNameplate.bIsPlayer or (SteadyPlatesInst.bCleaseMarkerGroupOnly and not unitOwner:IsInYourGroup()) then
		if cleanseMarker then cleanseMarker:Show(false) end
		return
	end

	local bHasRemoveableDebuff = false
	
	local unitBuffs = unitOwner:GetBuffs() 
	local unitDebuffs = unitBuffs["arHarmful"]
	--local unitBuffs   = unitBuffs["arBeneficial"]
	
	if unitDebuffs then
		for _, buff in pairs(unitDebuffs) do
		--for k=1, #unitDebuffs do 
			--local buff = unitDebuffs[k]
			if buff["splEffect"]:GetClass() == 38 then -- 38 = debuff dispellable
				bHasRemoveableDebuff = true
				break
			end
		end
	end
		
	if not cleanseMarker and bHasRemoveableDebuff then
		local wndParent = tNameplate.wndNameplate:FindChild("TargetAndDeathContainer")
		
		cleanseMarker = Apollo.LoadForm(SteadyPlatesInst.xmlDoc, "CleanseMarker", wndParent, self)
		if cleanseMarker then
			cleanseMarker:SetSprite(SteadyPlatesInst.strCleanseMarkerSprite)
			tNameplate.wnd.cleanseMarker = cleanseMarker
			cleanseMarker:SetAnchorOffsets(24, 24, 52, 52)
		end
	end
	
	if cleanseMarker then
		if bHasRemoveableDebuff then
			cleanseMarker:Show(true)
		else
			cleanseMarker:Show(false)
		end
	end
end

function SteadyPlates.DrawTargeting(self, tNameplate)
	local wnd = tNameplate.wnd
	local bIsTarget = tNameplate.bIsTarget
	local bUseTarget = bIsTarget and wnd.health:IsShown()
	local bShowTargetMarkerArrow = bIsTarget and SteadyPlatesInst.bShowTargetMarkerArrow
	local targetMarker = wnd.targetMarker
	local targetMarkerArrow = wnd.targetMarkerArrow
	
	--if bUseTarget and self.bShowMarkerTarget and not SteadyPlatesInst.bRemoveHealthBarBackground then
	if bUseTarget and self.bShowMarkerTarget then
		if tNameplate.nVulnerableTime > 0 then
			targetMarker:SetSprite("sprNP_BaseSelectedPurple")
		else
			targetMarker:SetSprite(karDisposition.tTargetPrimary[ tNameplate.eDisposition])
		end
		targetMarker:Show(true)
	else
		targetMarker:Show(false)
	end
	
	if bShowTargetMarkerArrow then
		targetMarkerArrow:SetSprite(karDisposition.tTargetSecondary[SteadyPlatesInst.nTargetMarkerArrowModifier + tNameplate.eDisposition])
	end
	targetMarkerArrow:Show(bShowTargetMarkerArrow)
end

function SteadyPlates.DrawCluster(self, tNameplate)
	local clusterMarker = tNameplate.wnd.clusterMarker
	
	if not SteadyPlatesInst.bShowClusterMarker or tNameplate.nClusterUnits == 0 or bPlayerInCombat then
		if clusterMarker then clusterMarker:Show(false) end
		return
	end
	
	if not clusterMarker and tNameplate.nClusterUnits ~= 0 then
		local wndParent = tNameplate.wndNameplate:FindChild("Container")
		clusterMarker = Apollo.LoadForm(SteadyPlatesInst.xmlDoc, "ClusterMarker", wndParent, self)
		if clusterMarker then
			tNameplate.wnd.clusterMarker = clusterMarker
		end
	end
	
	if clusterMarker then
		--if (not tNameplate.bIsTarget or SteadyPlatesInst.bRemoveHealthBarBackground) and tNameplate.wnd.health:IsShown() then
		if (not tNameplate.bIsTarget) and tNameplate.wnd.health:IsShown() then
			clusterMarker:SetText(tNameplate.nClusterUnits)
			clusterMarker:Show(true)
		else
			clusterMarker:Show(false)
		end
	end
end

function SteadyPlates.OnNameplateClick(self, wndHandler, wndControl, eMouseButton) --, nX, nY, bDoubleClick)
	if not wndHandler then return end
	if eMouseButton == GameLib.CodeEnumInputMouse.Left and not SteadyPlatesInst.unitTargetChangeRequestUnit then
		local name = wndHandler:GetName()
		local wnd = wndHandler:GetParent()
		if not wnd then return end
		if name ~= "Guild" then
			wnd = wnd:GetParent()
			if not wnd then return end
		end
		local tNameplate = aNameplates.arWnd2Nameplate[wnd:GetId()]
		if tNameplate == nil then
			return
		end
		local unitOwner = tNameplate.unitOwner
		if GameLib.GetTargetUnit() ~= unitOwner then
			GameLib.SetTargetUnit(unitOwner)
			--SteadyPlatesInst.unitTargetChangeRequestUnit = unitOwner
			--Apollo.StartTimer("SteadyPlates_TargetChangeRequestTimer")
			return true -- stop propogation (don't want to accidentally click through it)
		end
	end
end

function SteadyPlates.HelperCalculateConValueHook(self, unitTarget)
	if unitTarget == nil or self.unitPlayer == nil then --or not self.unitPlayer:IsValid() then
		return 1
	end

	local nUnitCon = self.unitPlayer:GetLevelDifferential(unitTarget) or 0
	
	local nCon = 1 --default setting

	if nUnitCon <= karConColors[1][1] then -- lower bound
		nCon = 1
	elseif nUnitCon >= karConColors[#karConColors][1] then -- upper bound
		nCon = #karConColors
	else
		for idx = 2, (#karConColors - 1) do -- everything in between
			if nUnitCon == karConColors[idx][1] then
				nCon = idx
			end
		end
	end

	return nCon
end
 
function SteadyPlates.DrawHealthShieldBarHook(self, wndHealth, unitOwner, eDisposition, tNameplate)
	if SteadyPlatesInst.bHealthShieldBarOverride then
		local nHealthCurr	= unitOwner:GetHealth()
		local wndHealthHealthLabel = tNameplate.wnd.healthHealthLabel
		--local wndHealthMaxHealth = tNameplate.wnd.healthMaxHealth
		--local wndHealthShieldFill = tNameplate.wnd.healthShieldFill
		--local wndHealthAbsorbFill = tNameplate.wnd.healthAbsorbFill
		local wndHealthBar = tNameplate.wnd.healthBar
		local wndHealingAbsorbBar = tNameplate.wnd.healingAbsorbBar
		local wndShieldBar = tNameplate.wnd.shieldBar
		local wndAbsorbBar = tNameplate.wnd.absorbBar
		local wndHealthClampMin = tNameplate.wnd.healthClampMin
		local wndHealthClampMax = tNameplate.wnd.healthClampMax
		
		
		local bReplaceHealthBar = SteadyPlatesInst.bOverwriteHealthBarColor
		local bReplaceShieldBar = SteadyPlatesInst.bOverwriteShieldBarColor
		local bReplaceAbsorbBar = SteadyPlatesInst.bOverwriteAbsorbBarColor
		local nVulnerabilityTime = unitOwner:GetCCStateTimeRemaining(Unit.CodeEnumCCState.Vulnerability)
		
		if tNameplate.bIsSimple or nHealthCurr == nil then
			--wndHealthBar:SetAnchorOffsets(self.nFrameLeft, self.nFrameTop, self.nFrameRight, self.nFrameBottom)
			wndHealthBar:SetAnchorPoints(0, 0, 1, 1)
			wndHealthHealthLabel:SetText("")
		else
			local nHealthMax = unitOwner:GetMaxHealth()
			local nShieldCurr = unitOwner:GetShieldCapacity()
			local nShieldMax = unitOwner:GetShieldCapacityMax()
			local nAbsorbCurr = 0
			local nAbsorbMax = unitOwner:GetAbsorptionMax()
			local nShieldMax = unitOwner:GetShieldCapacityMax()
			local nAbsorbCurr = 0
			if nAbsorbMax > 0 then
				nAbsorbCurr = unitOwner:GetAbsorptionValue() -- Since it doesn't clear when the buff drops off
			end
			local bHealthBarShieldAbove = SteadyPlatesInst.bHealthBarShieldAbove
			local nHealingAbsorb = unitOwner:GetHealingAbsorptionValue()
			local nHealthClampMin = unitOwner:GetHealthFloor()
			local nHealthClampMax = unitOwner:GetHealthCeiling()
			
			local nShieldMaxOrg, nAbsorbMaxOrg = nShieldMax, nAbsorbMax
			if SteadyPlatesInst.bHideShieldAbsorbLoss then
				nShieldMax = nShieldCurr
				nAbsorbMax = nAbsorbCurr
			end
			
			local nTotalMax = nHealthMax + nShieldMax + nAbsorbMax
			
			if unitOwner:IsDead() then
				nHealthCurr = 0
			end
				
			local nLastAbsorbShieldCurr = nAbsorbCurr + nShieldCurr
			if tNameplate.nLastHealthCurr ~= nHealthCurr or 
				 tNameplate.nLastAbsorbShieldCurr ~= nLastAbsorbShieldCurr or 
				 tNameplate.nLastHealthShieldTotalMax ~= nTotalMax or
				 tNameplate.nLastHealingAbsorb ~= nHealingAbsorb
			then
				tNameplate.nLastHealthCurr = nHealthCurr
				tNameplate.nLastAbsorbShieldCurr = nLastAbsorbShieldCurr
				tNameplate.nLastHealthShieldTotalMax = nTotalMax
				tNameplate.nLastHealingAbsorb = nHealingAbsorb
					
				-- Scaling
				--[[local nPointHealthRight = self.nFrameR * (nHealthCurr / nTotalMax) -
				local nPointShieldRight = self.nFrameR * ((nHealthCurr + nShieldMax) / nTotalMax)
				local nPointAbsorbRight = self.nFrameR * ((nHealthCurr + nShieldMax + nAbsorbMax) / nTotalMax)--]]
				
				local nPointHealthRight, nPointShieldRight, nPointAbsorbRight
				local nPointHealthLeft, nPointShieldLeft, nPointAbsorbLeft
				local nPointHealthTop, nPointAbsorbTop
				local nPointShieldBottom
				local nFrameLeft = self.nFrameLeft
				local nFrameTop, nFrameBottom = self.nFrameTop, self.nFrameBottom
				--local nHealthWidth = self.nHealthWidth
							
				--[[if bHealthBarShieldAbove and nShieldMaxOrg > 0 then
					nPointHealthTop = 4
					nPointAbsorbTop = 4
					nPointShieldBottom = -9
					
					local nHealthAbsorbMax = nHealthMax + nAbsorbMax
					nPointHealthRight = nFrameLeft + (nHealthWidth * (nHealthCurr / nHealthAbsorbMax)) -- applied to the difference between L and R
					nPointShieldRight = nFrameLeft + nHealthWidth
					nPointAbsorbRight = nFrameLeft + (nHealthWidth * ((nHealthCurr + nAbsorbMax) / nHealthAbsorbMax))
					
					nPointHealthLeft, nPointShieldLeft, nPointAbsorbLeft = self.nFrameLeft, self.nFrameLeft, nPointHealthRight - 1
					
					self:SetBarValue(wndHealthShieldFill, 0, nShieldCurr, nShieldMaxOrg)
					self:SetBarValue(wndHealthAbsorbFill, 0, nAbsorbCurr, nAbsorbMax)
					
					wndHealthShieldFill:EnableGlow(false)
				else--]]
					nPointHealthRight = nHealthMax / nTotalMax
					nPointShieldRight = (nHealthMax + nShieldMax) / nTotalMax
				
					if nShieldMax > 0 and nPointShieldRight - nPointHealthRight < 0.2 then
						nPointHealthRight = nPointHealthRight - (0.2 - (nPointShieldRight - nPointHealthRight))
					end
					
					wndHealingAbsorbBar:Show(nHealingAbsorb ~= 0)
					wndShieldBar:Show(nShieldMax ~= 0)
					wndAbsorbBar:Show(nAbsorbMax ~= 0)
					
					wndHealthBar:SetAnchorPoints(0, 0, nPointHealthRight, 1)
					wndHealingAbsorbBar:SetAnchorPoints(0, 0, nPointHealthRight, 1)
					wndHealthClampMin:SetAnchorPoints(0, 0, nPointHealthRight, 1)
					wndHealthClampMax:SetAnchorPoints(0, 0, nPointHealthRight, 1)
					wndShieldBar:SetAnchorPoints(nPointHealthRight, 0, nPointShieldRight, 1)
					wndAbsorbBar:SetAnchorPoints(nPointShieldRight, 0, 1, 1)
					
					wndShieldBar:SetMax(nShieldMax)
					wndAbsorbBar:SetMax(nAbsorbMax)
					
					nPointHealthTop = nFrameTop
					nPointAbsorbTop = nFrameTop
					nPointShieldBottom = nFrameBottom
										
					if not SteadyPlatesInst.bHideShieldProgressEdgeGlow then
						wndShieldBar:EnableGlow(nShieldCurr > 0 and nShieldCurr ~= nShieldMax)
					else
						wndShieldBar:EnableGlow(false)
					end
				--end
				
				wndHealthBar:SetMax(nHealthMax + nHealingAbsorb)
				wndHealthBar:SetProgress(nHealthCurr, 0) --(nHealthMax + nHealingAbsorb) * 4)
				wndHealthBar:Show(nHealthCurr > 0)
				
				wndHealingAbsorbBar:SetMax(nHealthMax + nHealingAbsorb)
				wndHealingAbsorbBar:SetProgress(nHealthCurr + nHealingAbsorb, 0) --, (nHealthMax + nHealingAbsorb) * 4)
				wndHealingAbsorbBar:Show(nHealthCurr > 0 and nHealingAbsorb > 0)
				
				wndHealthClampMin:SetMax(nHealthMax + nHealingAbsorb)
				wndHealthClampMin:SetProgress(nHealthClampMin)
				wndHealthClampMin:Show(nHealthClampMin > 0)
				
				wndHealthClampMax:SetMax(nHealthMax + nHealingAbsorb)
				wndHealthClampMax:SetProgress(nHealthClampMax)
				wndHealthClampMax:Show(nHealthClampMax ~= nHealthMax)
				
				--wndShieldBar:EnableGlow(nShieldCurr > 0 and nShieldCurr ~= nShieldMax)
				wndShieldBar:SetProgress(nShieldCurr, nShieldMax * 4)
				wndShieldBar:Show(nHealthCurr > 0 and nShieldCurr > 0)
				
				wndAbsorbBar:SetProgress(nAbsorbCurr, nAbsorbMax * 4)
				wndAbsorbBar:Show(nHealthCurr > 0 and nAbsorbCurr > 0)
				
				-- Resize
				--wndHealthBar:SetAnchorOffsets(nPointHealthLeft, nPointHealthTop, nPointHealthRight, nFrameBottom)
				--wndShieldBar:SetAnchorOffsets(nPointShieldLeft, nFrameTop, nPointShieldRight, nPointShieldBottom)
				--wndAbsorbBar:SetAnchorOffsets(nPointAbsorbLeft, nPointAbsorbTop, nPointAbsorbRight, nFrameBottom)
				
				-- Bars
				--wndHealingAbsorbBar:Show(nHealingAbsorb > 0)
				--wndHealthBar:Show(nHealthCurr > 0)
				--wndShieldBar:Show(nHealthCurr > 0 and nShieldMaxOrg > 0)
				--wndAbsorbBar:Show(nHealthCurr > 0 and nAbsorbMaxOrg > 0)
				
				-- Text
				local strText
				if SteadyPlatesInst.bReplaceHealthText then
					local fnHelperFormatBigNumber = SteadyPlates.HelperFormatBigNumber
					
					if SteadyPlatesInst.bUseCustomHealthTextFormat then
						strText = SteadyPlatesInst.strCustomHealthTextFormat or "error"
						local strHealthCurr = fnHelperFormatBigNumber(nHealthCurr)
						local strShieldCurr = fnHelperFormatBigNumber(nShieldCurr)
						local strHealthMax = fnHelperFormatBigNumber(nHealthMax)
						local strShieldMax = fnHelperFormatBigNumber(nShieldMax)
						local nHealthDeficit = nHealthCurr - nHealthMax
						local nShieldDeficit = nShieldCurr - nShieldMax
						local strHealthDeficit = fnHelperFormatBigNumber(nHealthDeficit)
						local strShieldDeficit = fnHelperFormatBigNumber(nShieldDeficit)
						local replacements = { 
							["h"] = strHealthCurr,
							["h%"] = nHealthMax and "" or tostring(math.floor((nHealthCurr/nHealthMax)*100)),
							["H"] = strHealthMax,
							["h-"] = nHealthDeficit == 0 and "" or strHealthDeficit,
							["h+"] = ((tNameplate.bIsFriendly and nHealthDeficit ~= 0)) and strHealthDeficit or strHealthCurr,
							["s"] = strShieldCurr,
							["s%"] = nShieldMaxOrg and "" or tostring(math.floor((nShieldCurr/nShieldMaxOrg)*100)),
							["s-"] = (nShieldMaxOrg and nShieldDeficit == 0) and "" or strShieldDeficit,
							["s+"] = ((tNameplate.bIsFriendly and strShieldDeficit ~= 0)) and strShieldDeficit or nShieldCurr,
							["S"] = strShieldMax,
						}
						strText = string.gsub(SteadyPlatesInst.strCustomHealthTextFormat, "%a+[%%%-%+]?", function (str) return replacements[str] end)
					else
						local bIsWounded = nHealthMax ~= nHealthCurr
						strText = fnHelperFormatBigNumber(nHealthMax)
						
						if bIsWounded then
							if SteadyPlatesInst.bReplaceHealthTextFriendlyDeficit and tNameplate.bIsFriendly then
								strText = fnHelperFormatBigNumber(nHealthCurr - nHealthMax)
							else
								if SteadyPlatesInst.bReplaceHealthTextPercentages then
									strText = tostring(math.floor((nHealthCurr/nHealthMax)*100)).."%"
								else
									strText = fnHelperFormatBigNumber(nHealthCurr)
								end
							end
						end
						
						if not SteadyPlatesInst.bReplaceHealthTextHideShield and nShieldMaxOrg > 0 and nShieldCurr > 0 then
							if not (bIsWounded and SteadyPlatesInst.bReplaceHealthTextFriendlyDeficit and tNameplate.bIsFriendly) then
								if SteadyPlatesInst.bReplaceHealthTextPercentages and nShieldMaxOrg ~= nShieldCurr then
									strText = strText.." / "..tostring(math.floor((nShieldCurr/nShieldMaxOrg)*100)).."%"
								else
									strText = strText.." / "..fnHelperFormatBigNumber(nShieldCurr)
								end
							end
						end
					end
				else
					local strHealthMax = self:HelperFormatBigNumber(nHealthMax)
					local strHealthCurr = self:HelperFormatBigNumber(nHealthCurr)
					local strShieldCurr = self:HelperFormatBigNumber(nShieldCurr)
					strText = nHealthMax == nHealthCurr and strHealthMax or String_GetWeaselString(Apollo.GetString("TargetFrame_HealthText"), strHealthCurr, strHealthMax)
					if nShieldMax > 0 and nShieldCurr > 0 then
						strText = String_GetWeaselString(Apollo.GetString("TargetFrame_HealthShieldText"), strText, strShieldCurr)
					end
				end
				wndHealthHealthLabel:SetText(strText)
			end 
		
			-- Shield & Absorb Sprites
			if bReplaceShieldBar then
				if tNameplate.strShieldBarColor == nil then
					local strColor = SteadyPlatesInst.tColorSets.Common.Shield.HealthBar
					tNameplate.strShieldBarColor = strColor
					wndShieldBar:SetBarColor(strColor)
					wndShieldBar:SetFullSprite("SP_Nameplates:sprNP_WhiteProgNew")
				end
			else
				if tNameplate.strShieldBarColor == nil then
					tNameplate.strShieldBarColor = "white"
					wndShieldBar:SetBarColor("white")
					wndShieldBar:SetFullSprite("SP_Nameplates:sprNP_BlueProg")
				end
			end
			if bReplaceAbsorbBar then
				if tNameplate.strAbsorbBarColor == nil then
					local strColor = SteadyPlatesInst.tColorSets.Common.Absorb.HealthBar
					tNameplate.strAbsorbBarColor = strColor
					wndAbsorbBar:SetBarColor(strColor)
					wndAbsorbBar:SetFullSprite("SP_Nameplates:sprNP_WhiteProgNew")
				end
			else
				if tNameplate.strAbsorbBarColor == nil then
					tNameplate.strAbsorbBarColor = "white"
					wndAbsorbBar:SetBarColor("white")
					wndAbsorbBar:SetFullSprite("SP_Nameplates:sprNP_YellowProg")
				end
			end
		end
		
		-- HealthBar Sprite
		if bReplaceHealthBar then
			if nVulnerabilityTime and nVulnerabilityTime > 0 then
				tNameplate.strHealthBarColor = nil
				wndHealthBar:SetFullSprite("SP_Nameplates:sprNP_WhiteProgNew")
				wndHealthBar:SetBarColor(SteadyPlatesInst.tColorSets.Common.Vulnerable.HealthBar)
			elseif tNameplate.strHealthBarColor == nil then
				local strColor = SteadyPlates.HelperGetColorFromColorSets("HealthBar", unitOwner, eDisposition, tNameplate)
				if not strColor then
					strColor = karDisposition.tHealthBarColor[eDisposition]
				end
				wndHealthBar:SetFullSprite("SP_Nameplates:sprNP_WhiteProgNew")
				wndHealthBar:SetBarColor(strColor)
				tNameplate.strHealthBarColor = strColor
			end
		else
			if nVulnerabilityTime and nVulnerabilityTime > 0 then
				tNameplate.strHealthBarColor = nil
				wndHealthBar:SetFullSprite("CRB_Nameplates:sprNP_PurpleProg")
			elseif tNameplate.strHealthBarColor == nil then
				tNameplate.strHealthBarColor = "white"
				wndHealthBar:SetBarColor("white")
				wndHealthBar:SetFullSprite(karDisposition.tHealthBar[eDisposition])
			end
		end

	else
		fDrawHealthShieldBar(self, wndHealth, unitOwner, eDisposition, tNameplate)
	end
end

function SteadyPlates.HelperGetColorFromColorSets(strSource, unitOwner, eDisposition, tNameplate)
	if SteadyPlatesInst.settingStatus == nil or eDisposition == Unit.CodeEnumDisposition.Unknown then
		return nil
	end
	local color = nil
	if tNameplate.bIsNPC then
		color = SteadyPlatesInst.tColorSets.NonPlayer[karDispositionNames[eDisposition]][strSource]
	elseif tNameplate.bIsSimple then
		color = SteadyPlatesInst.tColorSets.Simple[karDispositionNames[eDisposition]][strSource]
	elseif tNameplate.bIsPet or tNameplate.bIsScanner then
		color = SteadyPlatesInst.tColorSets.Pet[karDispositionNames[eDisposition]][strSource]
	elseif tNameplate.bIsPlayer then
		if tNameplate.bIsMe then
			if tNameplate.bIsPvpFlagged then
				color = SteadyPlatesInst.tColorSets.Self.PvP[strSource]
			else
				color = SteadyPlatesInst.tColorSets.Self.NoPvP[strSource]
			end
		else
			local tColorSet = nil
			local PvPMe = GameLib.IsPvpFlagged()
			local PvPHim = tNameplate.bIsPvpFlagged
			local strAppend
			if eDisposition == Unit.CodeEnumDisposition.Hostile or SteadyPlatesInst.bOverwriteColorPlayerPvPSpecific then
				if PvPMe and PvPHim then
					strAppend = "PvPBoth"
				elseif PvPMe then
					strAppend = "PvPMe"
				elseif PvPHim then
					strAppend = "PvPHim"
				else
					strAppend = ""
				end
			else
				if PvPHim then
					strAppend = "PvPHim"
				else
					strAppend = ""
				end
			end
			if SteadyPlatesInst.bOverwriteColorPlayerClassSpecific then
				tColorSet = SteadyPlatesInst.tColorSets.PlayerClass[karClassNames[unitOwner:GetClassId()]]
			end
			if tColorSet == nil then
				tColorSet = SteadyPlatesInst.tColorSets.Player
			end
			if eDisposition == Unit.CodeEnumDisposition.Hostile then
				color = tColorSet["Hostile"..strAppend][strSource]
			elseif eDisposition == Unit.CodeEnumDisposition.Friendly then
				local strFriendly
				if tNameplate.bIsGroupMember then
					strFriendly = "Group"
				elseif tNameplate.bIsGuildMember then
					strFriendly = "Guild"
				else
					strFriendly = "Friendly"
				end
				color = tColorSet[strFriendly..strAppend][strSource]
			end
		end
	else -- fallback to NonPlayer for all other types
		color = SteadyPlatesInst.tColorSets.NonPlayer[karDispositionNames[eDisposition]][strSource]
	end
	return color
end

function SteadyPlates.HelperVerifyVisibilityOptionsHook(self, tNameplate)
	if not tNameplate.bOnScreen then
		return false
  end
	
	local unitOwner = tNameplate.unitOwner
	local bIsTarget = tNameplate.bIsTarget
	local unitController = unitOwner:GetUnitOwner() or unitOwner
	local bPlayerInCombat = bPlayerInCombat
	local bIsMe = unitOwner:IsThePlayer()
	tNameplate.bIsMe = bIsMe
	
	local bIsGroupMember = unitController:IsInYourGroup()
	if bIsGroupMember ~= tNameplate.bIsGroupMember then
		tNameplate.bIsGroupMember = bIsGroupMember
		tNameplate.strHealthBarColor = nil
	end
	
	if (self.bUseOcclusion and tNameplate.bOccluded) then
		if not (bPlayerInCombat and ((bIsTarget and SteadyPlatesInst.bDisableOcclusionForTarget) or (bIsGroupMember and SteadyPlatesInst.bDisableOcclusionForGroupMembers))) then
			return false
		end
	end
	
	local eDisposition = tNameplate.eDisposition
	local bIsFriendly = eDisposition == Unit.CodeEnumDisposition.Friendly
	local bIsNeutral = eDisposition == Unit.CodeEnumDisposition.Neutral
	local bIsHostile = eDisposition == Unit.CodeEnumDisposition.Hostile
	if tNameplate.bIsFriendly ~= bIsFriendly or	tNameplate.bIsNeutral ~= bIsNeutral or tNameplate.bIsHostile ~= bIsHostile then
		tNameplate.bIsFriendly = bIsFriendly
		tNameplate.bIsNeutral = bIsNeutral
		tNameplate.bIsHostile = bIsHostile
		tNameplate.strHealthBarColor = nil
	end
	
	local bIsPvpFlagged = unitController:IsPvpFlagged()
	if bIsPvpFlagged ~= tNameplate.bIsPvpFlagged then
		tNameplate.bIsPvpFlagged = bIsPvpFlagged
		tNameplate.strHealthBarColor = nil
	end
	
	local bIsMouseOver = tNameplate.bIsMouseOver
	local bHiddenUnit = not unitOwner:ShouldShowNamePlate()
	if bHiddenUnit and not (bIsTarget) then -- or bIsMouseOver) then
    return (SteadyPlatesInst.bDisableHiddenCheckIfHostile and bIsHostile)
  end
	
	local bSpeechBubble = tNameplate.bSpeechBubble
	
	if bSpeechBubble and SteadyPlatesInst.bBubblesAlwaysHideNameplates then
		return false
	end
	
	if bIsMouseOver then
		--tNameplate.wndNameplate:ToFront()
		return true
	end
	
  if bIsTarget then
		if bPlayerInCombat or bIsHostile or (not bSpeechBubble or SteadyPlatesInst.bBubblesNeverHideNameplates) then
			return true
		end
  end
	
  if tNameplate.bGibbed then
    return false
  end
	
  local bIsDead = unitOwner:IsDead()  
  
	if not bIsDead and self.bShowMainObjectiveOnly and tNameplate.bIsObjective and (not bSpeechBubble or SteadyPlatesInst.bBubblesNeverHideNameplates) then
		return true
	end
  
	local bIsPet = tNameplate.bIsPet or tNameplate.bIsScanner
  
	if bIsPet then
		if SteadyPlatesInst.bHidePetsHostile and bIsHostile then
			return false
		elseif SteadyPlatesInst.bHidePetsFriendly and bIsFriendly then
			return false
		elseif SteadyPlatesInst.bHidePetsNeutral and bIsNeutral then
			return false
		end
	end
	
	local bIsNonPlayerCharacter = tNameplate.bIsNPC
	
	if bIsNonPlayerCharacter then
		if SteadyPlatesInst.bHideNonPlayerHostile and bIsHostile then
			return false
		elseif SteadyPlatesInst.bHideNonPlayerFriendly and bIsFriendly then
			return false
		elseif SteadyPlatesInst.bHideNonPlayerNeutral and bIsNeutral then
			return false
		end
	end
	
	local bIsPlayerCharacter = tNameplate.bIsPlayer
	
	if (bIsPlayerCharacter or bIsPet) then
		if bIsHostile and not bIsPvpFlagged and SteadyPlatesInst.bHideHostilePlayersOutOfPvP and not GameLib.IsPvpFlagged() then
			return false
		end
	end
	
  if bIsDead and (bPlayerInCombat or bIsHostile or not bIsPlayerCharacter) then --(bIsGroupMember or tNameplate.bIsGuildMember) then
    return false
  end
	
  if bSpeechBubble and not (bPlayerInCombat or bIsHostile or SteadyPlatesInst.bBubblesNeverHideNameplates) then
		return false
  end
	
	if bIsMe then
		if self.bShowMyNameplate and not bIsDead then
			return not (bPlayerInCombat and self.bHideInCombat)
		else
			return false
		end
	end
	
	local bShowNameplate = false
	
	if self.bShowMainObjectiveOnly and tNameplate.bIsObjective then
		bShowNameplate = true
	end

	if self.bShowMainGroupOnly and bIsGroupMember then
		bShowNameplate = true
	end

	if self.bShowDispositionHostile and bIsHostile then
		bShowNameplate = true
	end
	
	if self.bShowDispositionNeutral and bIsNeutral then
		bShowNameplate = true
	end
	
	if self.bShowDispositionFriendly and bIsFriendly then
		bShowNameplate = true
	end
	
	if self.bShowDispositionFriendlyPlayer and bIsFriendly and bIsPlayerCharacter then
		bShowNameplate = true
	end

	local tActivation = unitOwner:GetActivationState()
	
	if self.bShowVendor and tActivation.Vendor ~= nil then
		bShowNameplate = true
	end
	
	if self.bShowTaxi and (tActivation.FlightPathSettler ~= nil or tActivation.FlightPath ~= nil or tActivation.FlightPathNew) then
		bShowNameplate = true
	end
	
	if self.bShowOrganization and tNameplate.bIsGuildMember then
		bShowNameplate = true
	end

	if self.bShowMainObjectiveOnly then
		-- QuestGivers too
		if tActivation.QuestReward ~= nil then
			bShowNameplate = true
		end
		
		if tActivation.QuestNew ~= nil or tActivation.QuestNewMain ~= nil then
			bShowNameplate = true
		end
		
		if tActivation.QuestReceiving ~= nil then
			bShowNameplate = true
		end
		
		if tActivation.TalkTo ~= nil then
			bShowNameplate = true
		end
	end

	if bShowNameplate then
		bShowNameplate = not (bPlayerInCombat and self.bHideInCombat)
	end

	return bShowNameplate or bIsTarget
	
	--return fNameplatesHelperVerifyVisibilityOptions(self, tNameplate)
end

function SteadyPlates.HelperFormatBigNumber(nArg)
	local strResult
	if nArg < -10000 then
		if math.floor(nArg%1000/100) == 0 then
			strResult = String_GetWeaselString(Apollo.GetString("TargetFrame_ShortNumberWhole"), math.floor(nArg / 1000))
		else
			strResult = String_GetWeaselString(Apollo.GetString("TargetFrame_ShortNumberFloat"), nArg / 1000)
		end
	elseif nArg < 1000 then
		strResult = tostring(nArg)
	elseif nArg < 1000000 then
		if math.floor(nArg%1000/100) == 0 then
			strResult = String_GetWeaselString(Apollo.GetString("TargetFrame_ShortNumberWhole"), math.floor(nArg / 1000))
		else
			strResult = String_GetWeaselString(Apollo.GetString("TargetFrame_ShortNumberFloat"), nArg / 1000)
		end
	elseif nArg < 1000000000 then
		if math.floor(nArg%1000000/100000) == 0 then
			strResult = String_GetWeaselString(Apollo.GetString("TargetFrame_MillionsNumberWhole"), math.floor(nArg / 1000000))
		else
			strResult = String_GetWeaselString(Apollo.GetString("TargetFrame_MillionsNumberFloat"), nArg / 1000000)
		end
	elseif nArg < 1000000000000 then
		if math.floor(nArg%1000000/100000) == 0 then
			strResult = String_GetWeaselString(Apollo.GetString("TargetFrame_BillionsNumberWhole"), math.floor(nArg / 1000000))
		else
			strResult = String_GetWeaselString(Apollo.GetString("TargetFrame_BillionsNumberFloat"), nArg / 1000000)
		end
	else
		strResult = tostring(nArg)
	end
	return strResult
end

function SteadyPlates.HelperChangeHealthText(self, tNameplate)
	local eDisposition = tNameplate.eDisposition
	local wndHealthText = tNameplate.wnd.healthHealthLabel
	
	if not wndHealthText then
		return
	end
	
	if tNameplate.bIsHostile then
		SteadyPlates.HelperChangeHealthTextVisibility(wndHealthText, tNameplate, SteadyPlatesInst.nHealthBarTextHostile, bPlayerInCombat)
	elseif tNameplate.bIsNeutral then
		SteadyPlates.HelperChangeHealthTextVisibility(wndHealthText, tNameplate, SteadyPlatesInst.nHealthBarTextNeutral, bPlayerInCombat)
	elseif tNameplate.bIsFriendly then
		SteadyPlates.HelperChangeHealthTextVisibility(wndHealthText, tNameplate, SteadyPlatesInst.nHealthBarTextFriendly, bPlayerInCombat)
	end
end

function SteadyPlates.HelperChangeHealthTextVisibility(wndHealthText, tNameplate, nValue, bPlayerInCombat)
	--SPrint("%d %s %s", nValue , tostring(tNameplate.bIsTarget), tostring(bPlayerInCombat))
	if nValue == 1 then -- Never
		wndHealthText:Show(false)
	elseif nValue == 2 then -- In Combat Only
		wndHealthText:Show(bPlayerInCombat or tNameplate.bIsMouseOver)
	elseif nValue == 3 then -- Out Of Combat Only
		wndHealthText:Show(not bPlayerInCombat or tNameplate.bIsMouseOver)
	elseif nValue == 4 then -- Targeted Only
		wndHealthText:Show(tNameplate.bIsTarget or tNameplate.bIsMouseOver)
	else
		wndHealthText:Show(true)
	end
end

function SteadyPlates.HelperResetEnsure(tNameplate, nMode)
	tNameplate.bEnsureHide = false
	tNameplate.bEnsured = false
	tNameplate.nEnsureResetDelay = nil
	if not tNameplate.bIsMe then
		if nMode == 1 then
			if tNameplate.bMounted then
					tNameplate.wndNameplate:SetUnit(tNameplate.unitOwner:GetUnitMount(), 1)
				else
					tNameplate.wndNameplate:SetUnit(tNameplate.unitOwner, 1)
				end
		else
			if bPerspectivePlates then
				Event_FireGenericEvent("GenericEvent_PerspectivePlates_PerspectiveResize", tNameplate, 1, {	left = knDefaultLeft,	top = knDefaultTop, right = knDefaultRight,	bottom = knDefaultBottom})
			else
				tNameplate.wndNameplate:SetAnchorOffsets(knDefaultLeft, knDefaultTop, knDefaultRight, knDefaultBottom)
			end
		end
	end
end

function SteadyPlates.HelperEnsureNameplateOnScreen(self, tNameplate, nMode)
	if not tNameplate.bOnScreen or tNameplate.bIsMe then
		tNameplate.bEnsureHide = false
		if tNameplate.bEnsured then
			SteadyPlates.HelperResetEnsure(tNameplate, nMode)
		end
    return
  end
	
	local unitOwner = tNameplate.unitOwner
	if tNameplate.bMounted then
		unitOwner = unitOwner:GetUnitMount()
	end
	local tAnchor = unitOwner:GetOverheadAnchor() 
	local nEnsureResetDelay = tNameplate.nEnsureResetDelay
	
	if tAnchor then
		local y
		if nMode == 1 then
			y = tAnchor.y - 40
		else
			y = tAnchor.y - 72
		end
		local wndNameplate = tNameplate.wndNameplate
		
		if y < 0 then
			tNameplate.bEnsured = true
			if nMode == 1 then
				tNameplate.nEnsureResetDelay = 0
				wndNameplate:SetUnit(unitOwner, 0)
			else
				if bPerspectivePlates then
					Event_FireGenericEvent("GenericEvent_PerspectivePlates_PerspectiveResize", tNameplate, 1, {	left = knDefaultLeft,	top = (knDefaultTop - y), right = knDefaultRight,	bottom = (knDefaultBottom - y)})
				else
					wndNameplate:SetAnchorOffsets(knDefaultLeft, knDefaultTop-y, knDefaultRight, knDefaultBottom-y)
				end
				local px, py = wndNameplate:GetPos()
				--Print(py)
				if py > 0 then
					--if not tNameplate.bEnsureHide then
					--	Print("Hide "..px..", "..py.." / "..tAnchor.x..", "..tAnchor.y.." "..tostring(tNameplate.bIsTarget).." "..tostring(tNameplate.bIsMouseOver).." "..unitOwner:GetName())
					--end
					tNameplate.bEnsureHide = true
				else
					--if tNameplate.bEnsureHide then
					--	Print("UnHide "..px..", "..py.." / "..tAnchor.x..", "..tAnchor.y.." "..tostring(tNameplate.bIsTarget).." "..tostring(tNameplate.bIsMouseOver).." "..unitOwner:GetName())
					--end
					tNameplate.bEnsureHide = false
				end
			end
		elseif tNameplate.bEnsured then
			if nMode == 1 then
				if nEnsureResetDelay then
					if nEnsureResetDelay > (SteadyPlatesInst.nEnsureResetDelay / 50) then
						tNameplate.bEnsured = false
						wndNameplate:SetUnit(unitOwner, 1)
					else
						tNameplate.nEnsureResetDelay = nEnsureResetDelay + 1
					end
				else
					tNameplate.bEnsured = false
					wndNameplate:SetUnit(unitOwner, 1)
				end
			else
				--if tNameplate.bEnsureHide then
				--	Print("UnHide2 "..tAnchor.x..", "..tAnchor.y.." "..tostring(tNameplate.bIsTarget).." "..tostring(tNameplate.bIsMouseOver).." "..unitOwner:GetName())
				--end
				tNameplate.bEnsured = false
				tNameplate.bEnsureHide = false
				if bPerspectivePlates then
					Event_FireGenericEvent("GenericEvent_PerspectivePlates_PerspectiveResize", tNameplate, 1, {	left = knDefaultLeft,	top = knDefaultTop, right = knDefaultRight,	bottom = knDefaultBottom})
				else
					wndNameplate:SetAnchorOffsets(knDefaultLeft, knDefaultTop, knDefaultRight, knDefaultBottom)
				end
			end
		else
			--if tNameplate.bEnsureHide then
			--	Print("UnHide3 "..tAnchor.x..", "..tAnchor.y.." "..tostring(tNameplate.bIsTarget).." "..tostring(tNameplate.bIsMouseOver).." "..unitOwner:GetName())
			--end
			tNameplate.bEnsureHide = false
		end
	end
	
	if nMode == 1 and bPerspectivePlates then
		Event_FireGenericEvent("GenericEvent_PerspectivePlates_PerspectiveResize", tNameplate, 1, {	left = knDefaultLeft,	top = knDefaultTop, right = knDefaultRight,	bottom = knDefaultBottom})
	end
end

function SteadyPlates.HelperScaleNameplate(self, tNameplate)
	local newScale = 1.0
	if tNameplate.bIsTarget then
		newScale = 1.1
	end
	
	if tNameplate.nActualScale ~= newScale then
		local wnd = tNameplate.wndNameplate
		--wnd:Show(false, true)

		local nLeft, nTop, nRight, nBottom = wnd:GetAnchorOffsets()
		local nWidth = wnd:GetWidth()
		local nHeight = wnd:GetHeight()
		local nWidthDiff = (nWidth * newScale - nWidth)/2
		local nHeightDiff = (nHeight * newScale - nHeight)/2
		--Print(string.format("l=%d t=%d r=%d b=%d / h=%d w=%d / %d %d", nLeft, nTop, nRight, nBottom, nHeight, nWidth, nWidthDiff, nHeightDiff))
		
		wnd:SetAnchorOffsets(nLeft - nWidthDiff, nTop - nHeightDiff, nRight - nWidthDiff, nBottom - nHeightDiff)
		wnd:SetScale(newScale)
		--wnd:Show(true)
		tNameplate.nActualScale = newScale
	end
end

function SteadyPlates.CalculateDistance(vector3)
	return math.sqrt(math.pow(vector3.x, 2) + math.pow(vector3.y, 2) + math.pow(vector3.z, 2))
end

function SteadyPlates.GetDistanceToPlayer(target)
	local playerV3 = GameLib.GetPlayerUnit():GetPosition()
	if playerV3 then
		local targetV3
		if type(target) == 'table' then
			--assume vector3
			targetV3 = target
		else
			targetV3 = target:GetPosition()
		end
		
		--local distance = ( playerV3 - targetV3 ):Length()
		local distance = SteadyPlates.CalculateDistance(Vector3.New(playerV3.x - targetV3.x, playerV3.y - targetV3.y, playerV3.z - targetV3.z))
		return distance
	else
		return -1
	end
end

---------------------------------------------------------------------------------------------------
-- SteadyPlates Settings
---------------------------------------------------------------------------------------------------

function SteadyPlates:LoadSettingsWindow()
	self.wndSettings = Apollo.LoadForm(self.xmlSettingsDoc, "SteadyPlatesSettingsForm", nil, self)
	local wndContentStandard = self.wndSettings:FindChild("ContentStandard")
	local wndContentMarkers = self.wndSettings:FindChild("ContentMarkers")
	local wndContentColors = self.wndSettings:FindChild("ContentColors")
	
	if not aNameplates then
		self.wndSettingsStandard = Apollo.LoadForm(self.xmlSettingsDoc, "DependencyError", wndContentStandard, self)
		self.wndSettingsMarkers = Apollo.LoadForm(self.xmlSettingsDoc, "DependencyError", wndContentMarkers, self)
		self.wndSettingsColors = Apollo.LoadForm(self.xmlSettingsDoc, "DependencyError", wndContentColors, self)
		return
	end
	
	self.wndSettingsStandard = Apollo.LoadForm(self.xmlSettingsDoc, "StandardModule", wndContentStandard, self)
	self.wndSettingsMarkers = Apollo.LoadForm(self.xmlSettingsDoc, "MarkersModule", wndContentMarkers, self)
	self.wndSettingsColors = Apollo.LoadForm(self.xmlSettingsDoc, "ColorsModule", wndContentColors, self)
	
	self.wndSettings:FindChild("ContentToggleContainer:StandardViewCheck"):SetCheck(true)
	
	local wndFrame = self.wndSettingsStandard:FindChild("HealthBarFraming")
	local wndHealthBarOptions = wndFrame:FindChild("HealthBarOptions")
	
	local wndParent = wndHealthBarOptions:FindChild("HealthBarFriendlyOptions")
	local wndOptions = Apollo.LoadForm(self.xmlSettingsDoc, "HealthBarOptionsForm", wndParent, self)
	local nHeight = wndOptions:GetHeight()
	local nLeft, nTop, nRight, nBottom = wndParent:GetAnchorOffsets()
	wndParent:SetAnchorOffsets(nLeft, nTop, nRight, nTop + nHeight)
		
	wndParent = wndHealthBarOptions:FindChild("HealthBarNeutralOptions")
	wndOptions = Apollo.LoadForm(self.xmlSettingsDoc, "HealthBarOptionsForm", wndParent, self)
	nLeft, nTop, nRight, nBottom = wndParent:GetAnchorOffsets()
	wndParent:SetAnchorOffsets(nLeft, nTop, nRight, nTop + nHeight)
	
	wndParent = wndHealthBarOptions:FindChild("HealthBarHostileOptions")
	wndOptions = Apollo.LoadForm(self.xmlSettingsDoc, "HealthBarOptionsForm", wndParent, self)
	nLeft, nTop, nRight, nBottom = wndParent:GetAnchorOffsets()
	wndParent:SetAnchorOffsets(nLeft, nTop, nRight, nTop + nHeight)
	
	wndHealthBarOptions:ArrangeChildrenVert()
	local nHeightAdd = (nHeight-20) * 3

	nLeft, nTop, nRight, nBottom = wndHealthBarOptions:GetAnchorOffsets()
	wndHealthBarOptions:SetAnchorOffsets(nLeft, nTop, nRight, nBottom + nHeightAdd)
	
	nLeft, nTop, nRight, nBottom = wndFrame:GetAnchorOffsets()
	wndFrame:SetAnchorOffsets(nLeft, nTop, nRight, nBottom + nHeightAdd)
	
	nLeft, nTop, nRight, nBottom = self.wndSettingsStandard:GetAnchorOffsets()
	self.wndSettingsStandard:SetAnchorOffsets(nLeft, nTop, nRight, nBottom + nHeightAdd)
	
	wndContentStandard:RecalculateContentExtents()
	
	if not aGeminiColor then
		aGeminiColor = Apollo.GetPackage("GeminiColor").tPackage
	end
	if not aLibJSON then
		aLibJSON = Apollo.GetPackage("Lib:dkJSON-2.5").tPackage
	end
	
	self:LoadColorSets()
	
	--self.wndSettings:FindChild("SteadyPlatesSettingsForm:PerspectivePlatesInfo"):Show(bPerspectivePlates)
	
	self.wndSettings:Show(false, true)
end

function SteadyPlates:LoadColorSets()
	local wndFrame = self.wndSettingsColors:FindChild("ColorDefinitionsFraming")
	local wndColorsDefinitionSets = wndFrame:FindChild("ColorsDefinitionSets")
	
	wndColorsDefinitionSets:DestroyChildren()
	
	if SteadyPlatesInst.bOverwriteHealthBarColor or SteadyPlatesInst.bOverwriteShieldBarColor or SteadyPlatesInst.bOverwriteAbsorbBarColor then
		local wndSetCommon = self:AddColorSetCommon(wndColorsDefinitionSets)
		self:BuildColorSetContextMenu(wndSetCommon, {}, { Name = "Defaults", ColorSet = karColorSets.Common, Type = 3 })
	end
	
	if SteadyPlatesInst.bOverwriteHealthBarColor then 
		local wndSetSelf = self:AddColorSetSelf(wndColorsDefinitionSets)
		self:BuildColorSetContextMenu(wndSetSelf, {}, { Name = "Defaults", ColorSet = karColorSets.Self, Type = 0 })
		
		local wndSetNPC = self:AddColorSetDispositionType(wndColorsDefinitionSets, "NPC", self.tColorSets.NonPlayer)
		local wndSetPet = self:AddColorSetDispositionType(wndColorsDefinitionSets, "Pet", self.tColorSets.Pet)
		local wndSetSimple = self:AddColorSetDispositionType(wndColorsDefinitionSets, "Simple", self.tColorSets.Simple)
		
		self:BuildColorSetContextMenu(wndSetNPC, { wndSetPet, wndSetSimple }, { Name = "Defaults", ColorSet = karColorSets.NonPlayer, Type = 1 })
		self:BuildColorSetContextMenu(wndSetPet, { wndSetNPC, wndSetSimple }, { Name = "Defaults", ColorSet = karColorSets.Pet, Type = 1 })
		self:BuildColorSetContextMenu(wndSetSimple, { wndSetNPC, wndSetPet }, { Name = "Defaults", ColorSet = karColorSets.Simple, Type = 1 })
		
		if self.bOverwriteColorPlayerClassSpecific then
			local wndSetEsper = self:AddColorSetPlayerType(wndColorsDefinitionSets, "Player - Esper", self.tColorSets.PlayerClass, GameLib.CodeEnumClass.Esper)
			local wndSetWarrior = self:AddColorSetPlayerType(wndColorsDefinitionSets, "Player - Warrior", self.tColorSets.PlayerClass, GameLib.CodeEnumClass.Warrior)
			local wndSetStalker = self:AddColorSetPlayerType(wndColorsDefinitionSets, "Player - Stalker", self.tColorSets.PlayerClass, GameLib.CodeEnumClass.Stalker)
			local wndSetMedic = self:AddColorSetPlayerType(wndColorsDefinitionSets, "Player - Medic", self.tColorSets.PlayerClass, GameLib.CodeEnumClass.Medic)
			local wndSetSpellslinger = self:AddColorSetPlayerType(wndColorsDefinitionSets, "Player - Spellslinger", self.tColorSets.PlayerClass, GameLib.CodeEnumClass.Spellslinger)
			local wndSetEngineer = self:AddColorSetPlayerType(wndColorsDefinitionSets, "Player - Engineer", self.tColorSets.PlayerClass, GameLib.CodeEnumClass.Engineer)
			local wndSetPlayer = self:AddColorSetPlayerType(wndColorsDefinitionSets, "Player - Class Unknown", self.tColorSets.Player)
			
			self:BuildColorSetContextMenu(wndSetEsper, { wndSetPlayer, wndSetWarrior, wndSetStalker, wndSetMedic, wndSetSpellslinger, wndSetEngineer }, { Name = "Defaults", ColorSet = karColorSets.PlayerClass[karClassNames[GameLib.CodeEnumClass.Esper]], Type = 2 })
			self:BuildColorSetContextMenu(wndSetWarrior, { wndSetPlayer, wndSetEsper, wndSetSpellslinger, wndSetStalker, wndSetMedic, wndSetEngineer }, { Name = "Defaults", ColorSet = karColorSets.PlayerClass[karClassNames[GameLib.CodeEnumClass.Warrior]], Type = 2 })
			self:BuildColorSetContextMenu(wndSetStalker, { wndSetPlayer, wndSetEsper, wndSetWarrior, wndSetSpellslinger, wndSetMedic, wndSetEngineer }, { Name = "Defaults", ColorSet = karColorSets.PlayerClass[karClassNames[GameLib.CodeEnumClass.Stalker]], Type = 2 })
			self:BuildColorSetContextMenu(wndSetMedic, { wndSetPlayer, wndSetEsper, wndSetWarrior, wndSetStalker, wndSetSpellslinger, wndSetEngineer }, { Name = "Defaults", ColorSet = karColorSets.PlayerClass[karClassNames[GameLib.CodeEnumClass.Medic]], Type = 2 })
			self:BuildColorSetContextMenu(wndSetSpellslinger, { wndSetPlayer, wndSetEsper, wndSetWarrior, wndSetStalker, wndSetMedic, wndSetEngineer }, { Name = "Defaults", ColorSet = karColorSets.PlayerClass[karClassNames[GameLib.CodeEnumClass.Spellslinger]], Type = 2 })
			self:BuildColorSetContextMenu(wndSetEngineer, { wndSetPlayer, wndSetEsper, wndSetWarrior, wndSetStalker, wndSetMedic, wndSetSpellslinger }, { Name = "Defaults", ColorSet = karColorSets.PlayerClass[karClassNames[GameLib.CodeEnumClass.Engineer]], Type = 2 })
			self:BuildColorSetContextMenu(wndSetPlayer, { wndSetEsper, wndSetWarrior, wndSetStalker, wndSetMedic, wndSetSpellslinger, wndSetEngineer }, { Name = "Defaults", ColorSet = karColorSets.Player, Type = 2 })
		else
			self:AddColorSetPlayerType(wndColorsDefinitionSets, "Player", self.tColorSets.Player)
		end
	end
	
	self:RecalculateContentColorExtents()
end

function SteadyPlates:RecalculateContentColorExtents()
	local wndFrame = self.wndSettingsColors:FindChild("ColorDefinitionsFraming")
	local wndColorsDefinitionSets = wndFrame:FindChild("ColorsDefinitionSets")
	
	nHeightAdd = wndColorsDefinitionSets:ArrangeChildrenVert() + 50
	
	nLeft, nTop, nRight, nBottom = wndColorsDefinitionSets:GetAnchorOffsets()
	wndColorsDefinitionSets:SetAnchorOffsets(nLeft, nTop, nRight, nHeightAdd)
	
	nLeft, nTop, nRight, nBottom = wndFrame:GetAnchorOffsets()
	wndFrame:SetAnchorOffsets(nLeft, nTop, nRight, nHeightAdd + 20)

	wndFrame = self.wndSettingsColors:FindChild("ColorOptionssFraming")
	nHeight = wndFrame:GetHeight()
	
	nLeft, nTop, nRight, nBottom = self.wndSettingsColors:GetAnchorOffsets()
	self.wndSettingsColors:SetAnchorOffsets(nLeft, nTop, nRight, nHeight + nHeightAdd + 20)
	
	local wndContentColors = self.wndSettings:FindChild("ContentColors")
	wndContentColors:RecalculateContentExtents()
end

function SteadyPlates:AddContextMenuEntry(wndParent, strText, tData)
	local wndButton = Apollo.LoadForm(self.xmlSettingsDoc, "ColorSetContextMenuEntryButton", wndParent, self)
	wndButton:SetText(strText)
	wndButton:SetData(tData)
end

function SteadyPlates:BuildColorSetContextMenu(wndSet, tSourceSets, tDefaults)
	local wndContextMenuButton = wndSet:FindChild("ContextMenuButton")
	local tData = wndSet:GetData()
	local n = #tSourceSets
	
	if n < 1 and not tDefaults then 
		wndContextMenuButton:Show(false)
		return
	end
	
	local wndContextMenu = wndSet:FindChild("ContextMenuFlyOut")
	local wndCopySources = wndContextMenu:FindChild("CopySources")
	
	if n > 0 then
		for i=1, n, 1 do
			local wndSource = tSourceSets[i]
			self:AddContextMenuEntry(wndCopySources, wndSource:GetData().Name, { Target = wndSet, Source = wndSource})
		end
	end
	if tDefaults then
		self:AddContextMenuEntry(wndCopySources, tDefaults.Name, { Target = wndSet, Source = tDefaults})
	end
	
	local nHeight = wndCopySources:ArrangeChildrenVert()
	local nLeft, nTop, nRight, nBottom = wndCopySources:GetAnchorOffsets()
	wndCopySources:SetAnchorOffsets(nLeft, nTop, nRight, nTop + nHeight)
	nLeft, nTop, nRight, nBottom = wndContextMenu:GetAnchorOffsets()
	wndContextMenu:SetAnchorOffsets(nLeft, nTop, nRight, nHeight + 55)
end

function SteadyPlates:AddColorSetCommon(wndSets)
	local strSetName = "Common"
	local wndSet = Apollo.LoadForm(self.xmlSettingsDoc, "ColorDefinitionSet", wndSets, self)
	wndSet:FindChild("SetName:Label"):SetText(strSetName)
	local wndElements = wndSet:FindChild("SetElements")
	local tColorSet = self.tColorSets.Common
	wndSet:SetData({ Name = strSetName, ColorSet = tColorSet, Type = 3 })
	
	if SteadyPlatesInst.bOverwriteHealthBarColor then
		self:AddColorElement(wndElements, "Vulnerable", tColorSet.Vulnerable)
	end
	if SteadyPlatesInst.bOverwriteShieldBarColor then
		self:AddColorElement(wndElements, "Shield", tColorSet.Shield)
	end
	if SteadyPlatesInst.bOverwriteAbsorbBarColor then
		self:AddColorElement(wndElements, "Absorb", tColorSet.Absorb)
	end
	
	self:ToggleColorSet(wndSet, self.strLastOpenColorSet == strSetName)
	return wndSet
end

function SteadyPlates:AddColorSetSelf(wndSets)
	local strSetName = "Self"
	local wndSet = Apollo.LoadForm(self.xmlSettingsDoc, "ColorDefinitionSet", wndSets, self)
	wndSet:FindChild("SetName:Label"):SetText(strSetName)
	local wndElements = wndSet:FindChild("SetElements")
	local tColorSet = self.tColorSets.Self
	wndSet:SetData({ Name = strSetName, ColorSet = tColorSet, Type = 0 })
	
	self:AddColorElement(wndElements, strSetName, tColorSet.NoPvP)
	self:AddColorElement(wndElements, strSetName.." PvP", tColorSet.PvP)
	
	self:ToggleColorSet(wndSet, self.strLastOpenColorSet == strSetName)
	return wndSet
end

function SteadyPlates:AddColorSetDispositionType(wndSets, strSetName, tColorSet)
	local wndSet = Apollo.LoadForm(self.xmlSettingsDoc, "ColorDefinitionSet", wndSets, self)
	wndSet:FindChild("SetName:Label"):SetText(strSetName)
	local wndElements = wndSet:FindChild("SetElements")
	wndSet:SetData({ Name = strSetName, ColorSet = tColorSet, Type = 1 })
	
	self:AddColorElement(wndElements, "Hostile", tColorSet[karDispositionNames[Unit.CodeEnumDisposition.Hostile]])
	self:AddColorElement(wndElements, "Neutral", tColorSet[karDispositionNames[Unit.CodeEnumDisposition.Neutral]])
	self:AddColorElement(wndElements, "Friendly", tColorSet[karDispositionNames[Unit.CodeEnumDisposition.Friendly]])
	
	self:ToggleColorSet(wndSet, self.strLastOpenColorSet == strSetName)
	return wndSet
end

function SteadyPlates:AddColorSetPlayerType(wndSets, strSetName, tColorSet, eClass)
	local wndSet = Apollo.LoadForm(self.xmlSettingsDoc, "ColorDefinitionSet", wndSets, self)
	wndSet:FindChild("SetName:Label"):SetText(strSetName)
	local wndElements = wndSet:FindChild("SetElements")
	
	local tColorSetClass
	if eClass then
		local strClass = karClassNames[eClass]
		tColorSetClass = tColorSet[strClass]
	else
		tColorSetClass = tColorSet
	end
	wndSet:SetData({ Name = strSetName, ColorSet = tColorSetClass, Type = 2 })
	
	self:AddColorElements(wndElements, "Hostile", tColorSetClass)
	self:AddColorElements(wndElements, "Friendly", tColorSetClass)
	self:AddColorElements(wndElements, "Group", tColorSetClass)
	self:AddColorElements(wndElements, "Guild", tColorSetClass)
	
	self:ToggleColorSet(wndSet, self.strLastOpenColorSet == strSetName)
	return wndSet
end

function SteadyPlates:AddColorElements(wndElements, strBaseName, tColorSetClass)
	self:AddColorElement(wndElements, strBaseName, tColorSetClass[strBaseName])
	if strBaseName == "Hostile" or self.bOverwriteColorPlayerPvPSpecific then
		self:AddColorElement(wndElements, strBaseName.." PvP Him", tColorSetClass[strBaseName.."PvPHim"])
		self:AddColorElement(wndElements, strBaseName.." PvP Me", tColorSetClass[strBaseName.."PvPMe"])
		self:AddColorElement(wndElements, strBaseName.." PvP Both", tColorSetClass[strBaseName.."PvPBoth"])
	else
		self:AddColorElement(wndElements, strBaseName.." PvP", tColorSetClass[strBaseName.."PvPHim"])
	end
end

function SteadyPlates:AddColorElement(wndElements, strName, tColorElement)
	local wndElement = Apollo.LoadForm(self.xmlSettingsDoc, "ColorDefinitionElement", wndElements, self)
	local wndNamePreview = wndElement:FindChild("NamePreview")
	local wndHealthBarPreview = wndElement:FindChild("HealthBarPreview")
	local wndNameColorPickerButton = wndElement:FindChild("NameColorPickerButton")
	local wndNameColorValueEdit = wndElement:FindChild("NameColorValueEdit")
	local wndHealthBarColorPickerButton = wndElement:FindChild("HealthBarColorPickerButton")
	local wndHealthBarColorValueEdit = wndElement:FindChild("HealthBarColorValueEdit")
	local tData = {
		Name = strName,
		Values = {},
		OldValues = {},
		ColorElement = tColorElement
	}
	
	wndNameColorPickerButton:SetData("Name")
	wndNameColorValueEdit:SetData("Name")
	wndHealthBarColorPickerButton:SetData("HealthBar")
	wndHealthBarColorValueEdit:SetData("HealthBar")
	
	if tColorElement.Name then
		local strColorName = tColorElement.Name or "white"
		wndNamePreview:SetTextColor(strColorName)
		wndNamePreview:SetText(strName)
		wndNameColorPickerButton:SetBGColor(strColorName)
		local tColor = wndNameColorPickerButton:GetBGColor():ToTable()
		local strColor = aGeminiColor:RGBAPercToHex(tColor.r, tColor.g, tColor.b, tColor.a)
		wndNameColorValueEdit:SetText("#"..strColor)
		tData.Values["Name"] = strColor
		tData.OldValues["Name"] = strColor
	else
		wndNamePreview:SetTextColor("white")
		wndNamePreview:SetText(strName)
		wndNameColorPickerButton:Show(false)
		wndNameColorValueEdit:Show(false)
	end
	
	if tColorElement.HealthBar then
		local strColorHealthBar = tColorElement.HealthBar or "white"
		local strColorHex = self:HelperColorToHex(strColorHealthBar)
		wndHealthBarPreview:SetSprite("SP_Nameplates:sprNP_WhiteProgNew")
		wndHealthBarPreview:SetBGColor(strColorHealthBar)
		wndHealthBarColorPickerButton:SetBGColor(strColorHealthBar)
		local tColor = wndHealthBarColorPickerButton:GetBGColor():ToTable()
		local strColor = aGeminiColor:RGBAPercToHex(tColor.r, tColor.g, tColor.b, tColor.a)
		wndHealthBarColorValueEdit:SetText("#"..strColor)
		tData.Values["HealthBar"] = strColor
		tData.OldValues["HealthBar"] = strColor
	else
		wndHealthBarPreview:Show(false)
		wndHealthBarColorPickerButton:Show(false)
		wndHealthBarColorValueEdit:Show(false)
	end
	
	wndElement:SetData(tData)
end

function SteadyPlates:ApplyColorChange(tColorElement, tValues)
	tColorElement.Name = tValues.Name
	if tColorElement.HealthBar ~= tValues.HealthBar then
		tColorElement.HealthBar = tValues.HealthBar
		for _, tNameplateOther in pairs(aNameplates.arUnit2Nameplate) do
			tNameplateOther.strHealthBarColor = nil
			tNameplateOther.strShieldBarColor = nil
			tNameplateOther.strAbsorbBarColor = nil
		end
	end
end

function SteadyPlates:OnSettingsPickColor(wndHandler, wndControl, eMouseButton)
	local tColor = wndHandler:GetBGColor():ToTable()
	local strColor = aGeminiColor:RGBAPercToHex(tColor.r, tColor.g, tColor.b, tColor.a)
	
	if not self.wndColorPickerPopup then
		local tPickerOptions =
		{
			strInitialColor = strColor,
			callback = self.OnSettingsColorPicked,
			bCustomColor = true
		}
		self.wndColorPickerPopup = aGeminiColor:CreateColorPicker(self, tPickerOptions, wndHandler)
	else
		aGeminiColor:SetRGB(self.wndColorPickerPopup, tColor.r, tColor.g, tColor.b, tColor.a)
		aGeminiColor:SetHSV(self.wndColorPickerPopup, strColor)
		aGeminiColor:UpdateCurrPrevColors(self.wndColorPickerPopup)
		local tData = self.wndColorPickerPopup:GetData()
		tData.args = {wndHandler}
		tData.strInitialColor = strColor
		self.wndColorPickerPopup:SetData(tData)
	end
	
	self.wndColorPickerPopup:Show(true)
	self.wndColorPickerPopup:ToFront()	
end

function SteadyPlates:OnSettingsColorPicked(strColor, wndHandler)
	local strName = wndHandler:GetData()
	local wndElement = wndHandler:GetParent()
	local tData = wndElement:GetData()
	
	if strName == "Name" then
		local wndNamePreview = wndElement:FindChild("NamePreview")
		local wndNameColorPickerButton = wndElement:FindChild("NameColorPickerButton")
		local wndNameColorValueEdit = wndElement:FindChild("NameColorValueEdit")
		wndNamePreview:SetTextColor(strColor)
		wndNameColorPickerButton:SetBGColor(strColor)
		wndNameColorValueEdit:SetText("#"..strColor)
	else
		local wndHealthBarPreview = wndElement:FindChild("HealthBarPreview")
		local wndHealthBarColorPickerButton = wndElement:FindChild("HealthBarColorPickerButton")
		local wndHealthBarColorValueEdit = wndElement:FindChild("HealthBarColorValueEdit")
		wndHealthBarPreview:SetBGColor(strColor)
		wndHealthBarColorPickerButton:SetBGColor(strColor)
		wndHealthBarColorValueEdit:SetText("#"..strColor)
	end
	
	tData.Values[strName] = strColor
	tData.OldValues[strName] = strColor
	self:ApplyColorChange(tData.ColorElement, tData.Values)
	--Print("Picked "..tData.Name.."-"..strName..": "..tostring(strText).." = "..tostring(strColor))
end

function SteadyPlates:OnSettingsColorEditChanged(wndHandler, wndControl, strText)
	local strName = wndHandler:GetData()
	local wndElement = wndHandler:GetParent()
	local tData = wndElement:GetData()
	local strColor = self:HelperColorToHex(strText)
	
	if strColor then
		if strName == "Name" then
			local wndNamePreview = wndElement:FindChild("NamePreview")
			local wndNameColorPickerButton = wndElement:FindChild("NameColorPickerButton")
			wndNamePreview:SetTextColor(strColor)
			wndNameColorPickerButton:SetBGColor(strColor)
		else
			local wndHealthBarPreview = wndElement:FindChild("HealthBarPreview")
			local wndHealthBarColorPickerButton = wndElement:FindChild("HealthBarColorPickerButton")
			wndHealthBarPreview:SetBGColor(strColor)
			wndHealthBarColorPickerButton:SetBGColor(strColor)
		end
		
		tData.Values[strName] = strColor
		self:ApplyColorChange(tData.ColorElement, tData.Values)
		wndHandler:SetTextColor("xkcdBabyBlue")
	else
		wndHandler:SetTextColor("red")
	end
	--Print("Changed "..tData.Name.."-"..strName..": "..tostring(strText).." = "..tostring(strColor))
end

function SteadyPlates:OnSettingsColorEditReturn(wndHandler, wndControl, strText)
	local strName = wndHandler:GetData()
	local wndElement = wndHandler:GetParent()
	local tData = wndElement:GetData()
	local strColor = self:HelperColorToHex(strText)
	
	if strColor then
		if strName == "Name" then
			local wndNamePreview = wndElement:FindChild("NamePreview")
			local wndNameColorPickerButton = wndElement:FindChild("NameColorPickerButton")
			wndNamePreview:SetTextColor(strColor)
			wndNameColorPickerButton:SetBGColor(strColor)
			tData.Values["Name"] = strColor
		else
			local wndHealthBarPreview = wndElement:FindChild("HealthBarPreview")
			local wndHealthBarColorPickerButton = wndElement:FindChild("HealthBarColorPickerButton")
			wndHealthBarPreview:SetBGColor(strColor)
			wndHealthBarColorPickerButton:SetBGColor(strColor)
			tData.Values["HealthBar"] = strColor
		end
		
		tData.Values[strName] = strColor
		tData.OldValues[strName] = strColor
		wndHandler:SetText("#"..strColor)
		self:ApplyColorChange(tData.ColorElement, tData.Values)
	else
		wndHandler:SetTextColor("BubbleTextShout")
		wndHandler:SetText("#"..tData.OldValues[strName])
	end
	--Print("Return "..tData.Name.."-"..strName..": "..tostring(strText).." = "..tostring(strColor))
end

function SteadyPlates:OnSettingsColorEditEscape(wndHandler, wndControl)
	local strName = wndHandler:GetData()
	local wndElement = wndHandler:GetParent()
	local tData = wndElement:GetData()
	local strColor = tData.OldValues[strName]
	
	if strName == "Name" then
		local wndNamePreview = wndElement:FindChild("NamePreview")
		local wndNameColorPickerButton = wndElement:FindChild("NameColorPickerButton")
		wndNamePreview:SetTextColor(strColor)
		wndNameColorPickerButton:SetBGColor(strColor)
	else
		local wndHealthBarPreview = wndElement:FindChild("HealthBarPreview")
		local wndHealthBarColorPickerButton = wndElement:FindChild("HealthBarColorPickerButton")
		wndHealthBarPreview:SetBGColor(strColor)
		wndHealthBarColorPickerButton:SetBGColor(strColor)
	end
	
	self:ApplyColorChange(tData.ColorElement, tData.Values)
	wndHandler:SetText("#"..strColor)
	wndHandler:SetTextColor("xkcdBabyBlue")
	--Print("Escape "..tData.Name.."-"..strName..": Reset to "..tostring(strColor))
end

function SteadyPlates:HelperColorToHex(color, strFallback)
	local colorHex = strFallback
	if type(color) == "table" then
		colorHex = aGeminiColor:RGBAPercToHex(color.r, color.g, color.b, color.a)
	elseif type(color) == "string"  then
		local bFound, _, strColor = color:find("^#?(%x%x%x%x%x%x%x%x)$")
		if not bFound then
			bFound, _, strColor = color:find("^#?(%x%x%x%x%x%x)$")
		end
		
		if bFound then
			colorHex = strColor
		else
			local strColor = nil
			strColor = aGeminiColor:GetColorStringByName(color)
			if strColor then
				colorHex = strColor
			else
				local cr = ApolloColor.new(color)
				if cr then
					local tColor = cr:ToTable()
					if (tColor.r + tColor.g + tColor.b) > 0 then
						colorHex = aGeminiColor:RGBAPercToHex(tColor.r, tColor.g, tColor.b, tColor.a)
					end
					cr = nil
				end
			end
		end
	elseif type(color) == "userdata" then
		color = color:ToTable()
		colorHex = aGeminiColor:RGBAPercToHex(color.r, color.g, color.b, color.a)		
	end
	return colorHex	or strFallback
end

function SteadyPlates:OnSettingsDocumentReady()
	self:LoadSettingsWindow()
	self.loadingSettings = nil
	self:ShowSettingsWindow()
end

function SteadyPlates:OnStandardViewCheck(wndHandler, wndCtrl)
	self.wndSettings:FindChild("ContentStandard"):Show(true)
	self.wndSettings:FindChild("ContentMarkers"):Show(false)
	self.wndSettings:FindChild("ContentColors"):Show(false)
	
	if self.wndColorPickerPopup then
		self.wndColorPickerPopup:Show(false)
	end
	if self.wndColorSetShare then
		self.wndColorSetShare:Show(false)
	end
end

function SteadyPlates:OnMarkersViewCheck(wndHandler, wndCtrl)
	self.wndSettings:FindChild("ContentStandard"):Show(false)
	self.wndSettings:FindChild("ContentMarkers"):Show(true)
	self.wndSettings:FindChild("ContentColors"):Show(false)
	
	if self.wndColorPickerPopup then
		self.wndColorPickerPopup:Show(false)
	end
	if self.wndColorSetShare then
		self.wndColorSetShare:Show(false)
	end
end

function SteadyPlates:OnColorsViewCheck(wndHandler, wndCtrl)
	self.wndSettings:FindChild("ContentStandard"):Show(false)
	self.wndSettings:FindChild("ContentMarkers"):Show(false)
	self.wndSettings:FindChild("ContentColors"):Show(true)
end

function SteadyPlates:HideSettingsWindow()
	self.wndSettings:Close() --Show(false)
	if self.wndColorPickerPopup then
		self.wndColorPickerPopup:Show(false)
	end
end

function SteadyPlates:ShowSettingsWindow()
	if self.loadingSettings then return end
	
	if not self.wndSettings then
		self.loadingSettings = true
		self.xmlSettingsDoc = XmlDoc.CreateFromFile("SteadyPlatesSettings.xml")
		self.xmlSettingsDoc:RegisterCallback("OnSettingsDocumentReady", self) 
		return
	end
	
	if self.wndColorSetShare then
		self.wndColorSetShare:Show(false)
	end
	
	if not aNameplates then
		return
	end

	local wndControl = self.wndSettingsStandard:FindChild("EnhancementEnsureTarget")
	wndControl:SetCheck(self.bEnsureTargetNameplateOnScreen)
	wndControl:SetData("bEnsureTargetNameplateOnScreen")
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementEnsureInCombat")
	wndControl:SetCheck(self.bEnsureCombatNameplatesOnScreen)
	wndControl:SetData("bEnsureCombatNameplatesOnScreen")
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementEnsureModeBottom")
	wndControl:SetCheck(self.nEnsureMode == 1)
	wndControl:SetData("nEnsureMode")
	wndControl:Enable(self.bEnsureCombatNameplatesOnScreen or self.bEnsureTargetNameplateOnScreen)
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementEnsureModeBottomResetDelay")
	wndControl:SetData("nEnsureResetDelay")
	wndControl:SetText(self.nEnsureResetDelay)
	wndControl:Enable(self.nEnsureMode == 1)
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementShowOnMousOver")
	wndControl:SetCheck(self.bShowOnMouseOver)
	wndControl:SetData("bShowOnMouseOver")
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementShowClusterMarker")
	wndControl:SetCheck(self.bShowClusterMarker)
	wndControl:SetData("bShowClusterMarker")
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementShowMyName")
	wndControl:SetCheck(self.bHideMyNameWhenNotTargeted)
	wndControl:SetData("bHideMyNameWhenNotTargeted")
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementAnchorMyNameplateToFeet")
	wndControl:SetCheck(self.bAnchorMyNameplateToFeet)
	wndControl:SetData("bAnchorMyNameplateToFeet")
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementWotWThrottle")
	wndControl:SetCheck(self.bWotWThrottle)
	wndControl:SetData("bWotWThrottle")
	if not self.arThrottleUnitNames then
		wndControl:Enable(false)
	end
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementNameplateClickTargetsUnit")
	wndControl:SetCheck(self.bNameplateClickTargetsUnit)
	wndControl:SetData("bNameplateClickTargetsUnit")
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementExtendCertainDeath")
	wndControl:SetCheck(self.bExtendCertainDeath)
	wndControl:SetData("bExtendCertainDeath")
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementHideMyCastBar")
	wndControl:SetCheck(self.bHideMyCastBar)
	wndControl:SetData("bHideMyCastBar")
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementAdjustCertainDeath")
	wndControl:SetCheck(self.bAdjustCertainDeath)
	wndControl:SetData("bAdjustCertainDeath")
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementHideHostilePlayersOutOfPvp")
	wndControl:SetCheck(self.bHideHostilePlayersOutOfPvP)
	wndControl:SetData("bHideHostilePlayersOutOfPvP")
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementHideTitle")
	wndControl:SetCheck(self.bHideTitle)
	wndControl:SetData("bHideTitle")
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementDisableOcclusionForGroup")
	wndControl:SetCheck(self.bDisableOcclusionForGroupMembers)
	wndControl:SetData("bDisableOcclusionForGroupMembers")
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementDisableOcclusionForTarget")
	wndControl:SetCheck(self.bDisableOcclusionForTarget)
	wndControl:SetData("bDisableOcclusionForTarget")
	
	wndControl = self.wndSettingsStandard:FindChild("EnhancementShowClassIcon")
	wndControl:SetCheck(self.bShowSmallClassIconLeftToName)
	wndControl:SetData("bShowSmallClassIconLeftToName")
	
		
	wndControl = self.wndSettingsStandard:FindChild("BubbleHideNonHostile")
	wndControl:SetCheck(not self.bBubblesNeverHideNameplates and not self.bBubblesAlwaysHideNameplates)
	wndControl:SetData("bBubblesNeverHideNameplates")
	
	wndControl = self.wndSettingsStandard:FindChild("BubbleHideNever")
	wndControl:SetCheck(self.bBubblesNeverHideNameplates)
	wndControl:SetData("bBubblesNeverHideNameplates")
	
	wndControl = self.wndSettingsStandard:FindChild("BubbleHideAlways")
	wndControl:SetCheck(self.bBubblesAlwaysHideNameplates)
	wndControl:SetData("bBubblesAlwaysHideNameplates")
	
	wndControl = self.wndSettingsStandard:FindChild("HealthBarHideBackground")
	wndControl:SetCheck(self.bRemoveHealthBarBackground)
	wndControl:SetData("bRemoveHealthBarBackground")
	
	wndControl = self.wndSettingsStandard:FindChild("HealthBarHideShieldAbsorbLoss")
	wndControl:SetCheck(self.bHideShieldAbsorbLoss)
	wndControl:SetData("bHideShieldAbsorbLoss")
	
	wndControl = self.wndSettingsStandard:FindChild("HealthBarHideShieldGlow")
	wndControl:SetCheck(self.bHideShieldProgressEdgeGlow)
	wndControl:SetData("bHideShieldProgressEdgeGlow")
	
	wndControl = self.wndSettingsStandard:FindChild("HealthBarShieldAbove")
	wndControl:SetCheck(self.bHealthBarShieldAbove)
	wndControl:SetData("bHealthBarShieldAbove")
	
	wndControl = self.wndSettingsStandard:FindChild("HealthBarDimmFriendlyInCombat")
	wndControl:SetCheck(self.bDimmHealthInCombatFriendly)
	wndControl:SetData("bDimmHealthInCombatFriendly")
	
	wndControl = self.wndSettingsStandard:FindChild("HealthBarTextCustom")
	wndControl:SetCheck(self.bReplaceHealthText)
	wndControl:SetData("bReplaceHealthText")
	
	wndControl = self.wndSettingsStandard:FindChild("HealthBarTextCustomFriendlyDeficit")
	wndControl:SetCheck(self.bReplaceHealthTextFriendlyDeficit)
	wndControl:SetData("bReplaceHealthTextFriendlyDeficit")
	wndControl:Enable(self.bReplaceHealthText)
	
	wndControl = self.wndSettingsStandard:FindChild("HealthBarTextCustomHideShield")
	wndControl:SetCheck(self.bReplaceHealthTextHideShield)
	wndControl:SetData("bReplaceHealthTextHideShield")
	wndControl:Enable(self.bReplaceHealthText)
	
	wndControl = self.wndSettingsStandard:FindChild("HealthBarTextCustomPercentages")
	wndControl:SetCheck(self.bReplaceHealthTextPercentages)
	wndControl:SetData("bReplaceHealthTextPercentages")
	wndControl:Enable(self.bReplaceHealthText)
	
	wndControl = self.wndSettingsStandard:FindChild("HealthBarTextCustomDefault")
	wndControl:SetCheck(not self.bUseCustomHealthTextFormat)
	wndControl:SetData("bUseCustomHealthTextFormat2")
	wndControl:Enable(self.bReplaceHealthText)
	
	wndControl = self.wndSettingsStandard:FindChild("HealthBarTextCustomFormat")
	wndControl:SetCheck(self.bUseCustomHealthTextFormat)
	wndControl:SetData("bUseCustomHealthTextFormat1")
	wndControl:Enable(self.bReplaceHealthText)
	
	wndControl = self.wndSettingsStandard:FindChild("HealthBarTextCustomFormatEdit")
	wndControl:SetText(self.strCustomHealthTextFormat)
	wndControl:SetData("strCustomHealthTextFormat")
	wndControl:Enable(self.bReplaceHealthText)
	
	wndControl = self.wndSettingsStandard:FindChild("HealthBarTextCustomFormatHelp")
	wndControl:Enable(self.bReplaceHealthText)
	wndControl = self.wndSettingsStandard:FindChild("HealthBarTextCustomFormatHelpFlyout")
	wndControl:Show(false)
	
	
	local wndOptions = self.wndSettingsStandard:FindChild("HealthBarFriendlyOptions")
	wndControl = wndOptions:FindChild("TextShowNever")
	wndControl:SetCheck(self.nHealthBarTextFriendly == 1)
	wndControl:SetData("nHealthBarTextFriendly1")
	wndControl = wndOptions:FindChild("TextShowInCombat")
	wndControl:SetCheck(self.nHealthBarTextFriendly == 2)
	wndControl:SetData("nHealthBarTextFriendly2")
	wndControl = wndOptions:FindChild("TextShowOutOfCombat")
	wndControl:SetCheck(self.nHealthBarTextFriendly == 3)
	wndControl:SetData("nHealthBarTextFriendly3")
	wndControl = wndOptions:FindChild("TextShowTargetedOnly")
	wndControl:SetCheck(self.nHealthBarTextFriendly == 4)
	wndControl:SetData("nHealthBarTextFriendly4")
	wndControl = wndOptions:FindChild("TextShowAlways")
	wndControl:SetCheck(self.nHealthBarTextFriendly == 0 or self.nHealthBarTextFriendly == nil)
	wndControl:SetData("nHealthBarTextFriendly0")
	wndControl = wndOptions:FindChild("ShowCastBarTargetOnly")
	wndControl:SetCheck(self.bShowCastBarTargetOnlyFriendly)
	wndControl:SetData("bShowCastBarTargetOnlyFriendly")
	wndControl = wndOptions:FindChild("ShowCastName")
	wndControl:SetCheck(self.bShowCastBarNameFriendly)
	wndControl:SetData("bShowCastBarNameFriendly")
	wndControl = wndOptions:FindChild("HidePets")
	wndControl:SetCheck(self.bHidePetsFriendly)
	wndControl:SetData("bHidePetsFriendly")
	wndControl = wndOptions:FindChild("HideNonPlayer")
	wndControl:SetCheck(self.bHideNonPlayerFriendly)
	wndControl:SetData("bHideNonPlayerFriendly")
	wndControl = wndOptions:FindChild("ShowHealthBarTargetOnly")
	wndControl:SetCheck(self.bShowHealthBarTargetOnlyFriendly)
	wndControl:SetData("bShowHealthBarTargetOnlyFriendly")
	
	wndOptions = self.wndSettingsStandard:FindChild("HealthBarNeutralOptions")
	wndControl = wndOptions:FindChild("TextShowNever")
	wndControl:SetCheck(self.nHealthBarTextNeutral == 1)
	wndControl:SetData("nHealthBarTextNeutral1")
	wndControl = wndOptions:FindChild("TextShowInCombat")
	wndControl:SetCheck(self.nHealthBarTextNeutral == 2)
	wndControl:SetData("nHealthBarTextNeutral2")
	wndControl = wndOptions:FindChild("TextShowOutOfCombat")
	wndControl:SetCheck(self.nHealthBarTextNeutral == 3)
	wndControl:SetData("nHealthBarTextNeutral3")
	wndControl = wndOptions:FindChild("TextShowTargetedOnly")
	wndControl:SetCheck(self.nHealthBarTextNeutral == 4)
	wndControl:SetData("nHealthBarTextNeutral4")
	wndControl = wndOptions:FindChild("TextShowAlways")
	wndControl:SetCheck(self.nHealthBarTextNeutral == 0 or self.nHealthBarTextNeutral == nil)
	wndControl:SetData("nHealthBarTextNeutral0")
	wndControl = wndOptions:FindChild("ShowCastBarTargetOnly")
	wndControl:SetCheck(self.bShowCastBarTargetOnlyNeutral)
	wndControl:SetData("bShowCastBarTargetOnlyNeutral")
	wndControl = wndOptions:FindChild("ShowCastName")
	wndControl:SetCheck(self.bShowCastBarNameNeutral)
	wndControl:SetData("bShowCastBarNameNeutral")
	wndControl = wndOptions:FindChild("HidePets")
	wndControl:SetCheck(self.bHidePetsNeutral)
	wndControl:SetData("bHidePetsNeutral")
	wndControl = wndOptions:FindChild("HideNonPlayer")
	wndControl:SetCheck(self.bHideNonPlayerNeutral)
	wndControl:SetData("bHideNonPlayerNeutral")
	wndControl = wndOptions:FindChild("ShowHealthBarTargetOnly")
	wndControl:SetCheck(self.bShowHealthBarTargetOnlyNeutral)
	wndControl:SetData("bShowHealthBarTargetOnlyNeutral")
	
	wndOptions = self.wndSettingsStandard:FindChild("HealthBarHostileOptions")
	wndControl = wndOptions:FindChild("TextShowNever")
	wndControl:SetCheck(self.nHealthBarTextHostile == 1)
	wndControl:SetData("nHealthBarTextHostile1")
	wndControl = wndOptions:FindChild("TextShowInCombat")
	wndControl:SetCheck(self.nHealthBarTextHostile == 2)
	wndControl:SetData("nHealthBarTextHostile2")
	wndControl = wndOptions:FindChild("TextShowOutOfCombat")
	wndControl:SetCheck(self.nHealthBarTextHostile == 3)
	wndControl:SetData("nHealthBarTextHostile3")
	wndControl = wndOptions:FindChild("TextShowTargetedOnly")
	wndControl:SetCheck(self.nHealthBarTextHostile == 4)
	wndControl:SetData("nHealthBarTextHostile4")
	wndControl = wndOptions:FindChild("TextShowAlways")
	wndControl:SetCheck(self.nHealthBarTextHostile == 0 or self.nHealthBarTextHostile == nil)
	wndControl:SetData("nHealthBarTextHostile0")
	wndControl = wndOptions:FindChild("ShowCastBarTargetOnly")
	wndControl:SetCheck(self.bShowCastBarTargetOnlyHostile)
	wndControl:SetData("bShowCastBarTargetOnlyHostile")
	wndControl = wndOptions:FindChild("ShowCastName")
	wndControl:SetCheck(self.bShowCastBarNameHostile)
	wndControl:SetData("bShowCastBarNameHostile")
	wndControl = wndOptions:FindChild("HidePets")
	wndControl:SetCheck(self.bHidePetsHostile)
	wndControl:SetData("bHidePetsHostile")
	wndControl = wndOptions:FindChild("HideNonPlayer")
	wndControl:SetCheck(self.bHideNonPlayerHostile)
	wndControl:SetData("bHideNonPlayerHostile")
	wndControl = wndOptions:FindChild("ShowHealthBarTargetOnly")
	wndControl:SetCheck(self.bShowHealthBarTargetOnlyHostile)
	wndControl:SetData("bShowHealthBarTargetOnlyHostile")
	
	
	wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerDisabled")
	wndControl:SetCheck(not self.bShowThreatMarker)
	
	wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerIcons")
	for idx, wnd in pairs(wndControl:GetChildren()) do
		if wnd.IsChecked ~= nil then
			if self.bShowThreatMarker and wnd:FindChild("Icon"):GetSprite() == self.strThreatMarkerSprite then
				wnd:SetCheck(true)
			else
				wnd:SetCheck(false)
			end
			wnd:SetData("strThreatMarkerSprite")
		end
	end
	
	wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerTop")
	wndControl:SetCheck(not self.bThreatMarkerLeft)
	wndControl:SetData("bThreatMarkerLeft")
	wndControl:Enable(self.bShowThreatMarker)
	
	wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerLeft")
	wndControl:SetCheck(self.bThreatMarkerLeft)
	wndControl:SetData("bThreatMarkerLeft")
	wndControl:Enable(self.bShowThreatMarker)
	
	wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerTankMode")
	wndControl:SetCheck(self.bThreatMarkerTankMode)
	wndControl:SetData("bThreatMarkerTankMode")
	wndControl:Enable(self.bShowThreatMarker)
	
	wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerGroupOnly")
	wndControl:SetCheck(self.bShowThreatMarkerGroupOnly)
	wndControl:SetData("bShowThreatMarkerGroupOnly")
	wndControl:Enable(self.bShowThreatMarker)
	
	wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerInstanceOnly")
	wndControl:SetCheck(self.bShowThreatMarkerInstanceOnly)
	wndControl:SetData("bShowThreatMarkerInstanceOnly")
	wndControl:Enable(self.bShowThreatMarker)
	
	
	wndControl = self.wndSettingsMarkers:FindChild("InterruptArmorMarkerDisabled")
	wndControl:SetCheck(not self.bShowInterruptArmorMarker)
	
	wndControl = self.wndSettingsMarkers:FindChild("InterruptMarkerIcons")
	for idx, wnd in pairs(wndControl:GetChildren()) do
		if wnd.IsChecked ~= nil then
			if self.bShowInterruptArmorMarker and wnd:FindChild("Icon"):GetSprite() == self.strInterruptArmorMarkerSprite then
				wnd:SetCheck(true)
			else
				wnd:SetCheck(false)
			end
			wnd:SetData("strInterruptArmorMarkerSprite")
		end
	end
	
	wndControl = self.wndSettingsMarkers:FindChild("InterruptArmorMarkerUsePlayerClass")
	wndControl:SetCheck(self.bInterruptArmorMarkerUsePlayerClass)
	wndControl:SetData("bInterruptArmorMarkerUsePlayerClass")
	wndControl:Enable(self.bShowInterruptArmorMarker)
	
	
	wndControl = self.wndSettingsMarkers:FindChild("CleanseMarkerDisabled")
	wndControl:SetCheck(not self.bShowCleanseMarker)
	
	wndControl = self.wndSettingsMarkers:FindChild("CleanseMarkerIcons")
	for idx, wnd in pairs(wndControl:GetChildren()) do
		if wnd.IsChecked ~= nil then
			if self.bShowCleanseMarker and wnd:FindChild("Icon"):GetSprite() == self.strCleanseMarkerSprite then
				wnd:SetCheck(true)
			else
				wnd:SetCheck(false)
			end
			wnd:SetData("strCleanseMarkerSprite")
		end
	end
	
	wndControl = self.wndSettingsMarkers:FindChild("CleanseMarkerGroupOnly")
	wndControl:SetCheck(self.bCleanseMarkerGroupOnly)
	wndControl:SetData("bCleanseMarkerGroupOnly")
	wndControl:Enable(self.bShowCleanseMarker)
	
	wndControl = self.wndSettingsMarkers:FindChild("TargetMarkerDisabled")
	wndControl:SetCheck(not self.bShowTargetMarkerArrow)
	wndControl:SetData("nTargetMarkerArrowModifier")
	
	wndContainer = self.wndSettingsMarkers:FindChild("TargetMarkerIcons")
	wndControl = wndContainer:FindChild("TargetMarker1")
	wndControl:SetCheck(self.bShowTargetMarkerArrow and self.nTargetMarkerArrowModifier == 0)
	wndControl:SetData("nTargetMarkerArrowModifier")
	wndControl = wndContainer:FindChild("TargetMarker2")
	wndControl:SetCheck(self.bShowTargetMarkerArrow and self.nTargetMarkerArrowModifier == 10)
	wndControl:SetData("nTargetMarkerArrowModifier")
	wndControl = wndContainer:FindChild("TargetMarker3")
	wndControl:SetCheck(self.bShowTargetMarkerArrow and self.nTargetMarkerArrowModifier == 20)
	wndControl:SetData("nTargetMarkerArrowModifier")
	wndControl = wndContainer:FindChild("TargetMarker4")
	wndControl:SetCheck(self.bShowTargetMarkerArrow and self.nTargetMarkerArrowModifier == 30)
	wndControl:SetData("nTargetMarkerArrowModifier")
	
	
	wndControl = self.wndSettingsColors:FindChild("ColorsNames")
	wndControl:SetCheck(self.bOverwriteNameColor)
	wndControl:SetData("bOverwriteNameColor")
	
	wndControl = self.wndSettingsColors:FindChild("ColorsHealthBars")
	wndControl:SetCheck(self.bOverwriteHealthBarColor)
	wndControl:SetData("bOverwriteHealthBarColor")
	
	wndControl = self.wndSettingsColors:FindChild("ColorsShieldBar")
	wndControl:SetCheck(self.bOverwriteShieldBarColor)
	wndControl:SetData("bOverwriteShieldBarColor")
	
	wndControl = self.wndSettingsColors:FindChild("ColorsAbsorbBar")
	wndControl:SetCheck(self.bOverwriteAbsorbBarColor)
	wndControl:SetData("bOverwriteAbsorbBarColor")
	
	wndControl = self.wndSettingsColors:FindChild("ColorsHealthBarsPlayerClassSpecific")
	wndControl:SetCheck(self.bOverwriteColorPlayerClassSpecific)
	wndControl:SetData("bOverwriteColorPlayerClassSpecific")
	
	wndControl = self.wndSettingsColors:FindChild("ColorsHealthBarsPlayerPvPSpecific")
	wndControl:SetCheck(self.bOverwriteColorPlayerPvPSpecific)
	wndControl:SetData("bOverwriteColorPlayerPvPSpecific")
	
	self.wndSettings:Invoke()
end

function SteadyPlates:ToggleColorSet(wndSet, bOpen)
	if bOpen == nil then
		bOpen = not (wndSet:GetHeight() > 36)
	end
	local wndElements = wndSet:FindChild("SetElements")
	wndSet:FindChild("SetToggleButton"):SetCheck(bOpen)
	wndSet:FindChild("SetName"):SetCheck(bOpen)
	if bOpen then
		local nHeight = wndElements:ArrangeChildrenVert()
		wndElements:SetAnchorOffsets(0, 40, 0, 40 + nHeight)
		wndSet:SetAnchorOffsets(0, 0, 0, 50 + nHeight)
	else
		wndSet:SetAnchorOffsets(0, 0, 0, 36)
	end
	wndSet:FindChild("ContextMenuFlyOut"):Show(false)
	wndSet:FindChild("ContextMenuButton"):Show(bOpen and wndSet:GetData().Type)
end

function SteadyPlates:ScrollToColorSet(wndSet)
	local wndContentColors = self.wndSettings:FindChild("ContentColors")	
	local wndFrame = self.wndSettingsColors:FindChild("ColorOptionssFraming")
	local nHeight = wndFrame:GetHeight()
	local nTop = ({wndSet:GetAnchorOffsets()})[2]
	local nNewPos = nTop + nHeight + 35
	local nCurrPos = wndContentColors:GetVScrollPos()
	--Print(nNewPos.." "..nCurrPos)
	if nNewPos > (nCurrPos + nHeight/2) then
		wndContentColors:SetVScrollPos(nTop + nHeight + 35)
	end
end

function SteadyPlates:OnSettingsColorSetToggle(wndHandler, wndControl, eMouseButton)
	local wndSet = wndHandler:GetParent()
	self:ToggleColorSet(wndSet)
	self:RecalculateContentColorExtents()
	self:ScrollToColorSet(wndSet)
end

function SteadyPlates:OnSettingsGenericSingleUncheck(wndHandler, wndControl, eMouseButton)
	local strSettingName = wndControl:GetData()
	if strSettingName ~= nil then
		if strSettingName == "strInterruptArmorMarkerSprite" then
			local bAny = false
			for idx, wnd in pairs(wndControl:GetParent():GetChildren()) do
				if wnd.IsChecked ~= nil and wnd:IsChecked() then
					bAny = true
					break
				end
			end
			if not bAny then
				wndControl = self.wndSettingsMarkers:FindChild("InterruptArmorMarkerDisabled")
				wndControl:SetCheck(true)
				self.bShowInterruptArmorMarker = false
			end
			wndControl = self.wndSettingsMarkers:FindChild("InterruptArmorMarkerUsePlayerClass")
			wndControl:Enable(bAny)
		elseif strSettingName == "strThreatMarkerSprite" then
			local bAny = false
			for idx, wnd in pairs(wndControl:GetParent():GetChildren()) do
				if wnd.IsChecked ~= nil and wnd:IsChecked() then
					bAny = true
					break
				end
			end
			if not bAny then
				wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerDisabled")
				wndControl:SetCheck(true)
				self.bShowThreatMarker = false
			end
			wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerTop")
			wndControl:Enable(bAny)
			wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerLeft")
			wndControl:Enable(bAny)
			wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerTankMode")
			wndControl:Enable(bAny)
		elseif strSettingName == "strCleanseMarkerSprite" then
			local bAny = false
			for idx, wnd in pairs(wndControl:GetParent():GetChildren()) do
				if wnd.IsChecked ~= nil and wnd:IsChecked() then
					bAny = true
					break
				end
			end
			if not bAny then
				wndControl = self.wndSettingsMarkers:FindChild("CleanseMarkerDisabled")
				wndControl:SetCheck(true)
				self.bShowCleanseMarker = false
			end
			wndControl = self.wndSettingsMarkers:FindChild("CleanseMarkerGroupOnly")
			wndControl:Enable(bAny)
		elseif strSettingName == "nTargetMarkerArrowModifier" then
			local bAny = false
			for idx, wnd in pairs(wndControl:GetParent():GetChildren()) do
				if wnd.IsChecked ~= nil and wnd:IsChecked() then
					bAny = true
					break
				end
			end
			if not bAny then
				wndControl = self.wndSettingsMarkers:FindChild("TargetMarkerDisabled")
				wndControl:SetCheck(true)
				self.bShowTargetMarkerArrow = false
			end
		end
	end
end

function SteadyPlates:OnSettingsMouseButtonDown(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation)
	if wndControl then
		local name = wndControl:GetName()
		if name == "HealthBarTextCustomFormatHelpFlyout" then
			wndControl:Show(false)
			return true
		end
	end
end

function SteadyPlates:OnSettingsTextCustomFormatHelpFlyout(wndHandler, wndControl, eMouseButton)
	-- Toggle the help window
	local wnd = wndHandler:FindChild("HealthBarTextCustomFormatHelpFlyout")
	wnd:Show(not wnd:IsVisible())
end

function SteadyPlates:OnSettingsTextCustomFormatEditReturn(wndHandler, wndControl, strText)
	self.strCustomHealthTextFormat = strText
	self:UpdateExistingNameplatesToSettings(false, true)
	
	local wndParent = wndHandler:GetParent()
	if wndParent then
		local wnd = wndParent:FindChild("HealthBarTextCustomFormatHelpFlyout")
		wnd:Show(false)
	end
end

function SteadyPlates:OnSettingsEnsureResetDelayReturn(wndHandler, wndControl, strText)
	local nDelay = tonumber(strText)
	self.nEnsureResetDelay = nDelay
end

local EnhancementEnsureModeBottomResetDelayTextOld
function SteadyPlates:OnSettingsEnsureResetDelayChanged(wndHandler, wndControl, strText)
	local nDelay
	if string.len(strText) == 0 then
		nDelay = 0
		strText = "0"
	else
		nDelay = tonumber(strText)
	end
	if nDelay == nil then
		local wndControl = self.wndSettingsStandard:FindChild("EnhancementEnsureModeBottomResetDelay"):SetText(EnhancementEnsureModeBottomResetDelayTextOld or tostring(self.nEnsureResetDelay or 0))
	else
		EnhancementEnsureModeBottomResetDelayTextOld = strText
	end
end

function SteadyPlates:OnSettingsGenericSingleCheck(wndHandler, wndControl, eMouseButton)
	local strSettingName = wndControl:GetData()
	if strSettingName ~= nil then
		--Print(strSettingName)
		if strSettingName == "bEnsureCombatNameplatesOnScreen" then
			self.bEnsureCombatNameplatesOnScreen = wndControl:IsChecked()
			wndControl = self.wndSettingsStandard:FindChild("EnhancementEnsureModeBottom")
			wndControl:Enable(self.bEnsureCombatNameplatesOnScreen or self.bEnsureTargetNameplateOnScreen)
			wndControl = self.wndSettingsStandard:FindChild("EnhancementEnsureModeBottomResetDelay")
			wndControl:Enable((self.bEnsureCombatNameplatesOnScreen or self.bEnsureTargetNameplateOnScreen) and self.nEnsureMode == 1)
		elseif strSettingName == "bEnsureTargetNameplateOnScreen" then
			self.bEnsureTargetNameplateOnScreen = wndControl:IsChecked()
			wndControl = self.wndSettingsStandard:FindChild("EnhancementEnsureModeBottom")
			wndControl:Enable(self.bEnsureCombatNameplatesOnScreen or self.bEnsureTargetNameplateOnScreen)
			wndControl = self.wndSettingsStandard:FindChild("EnhancementEnsureModeBottomResetDelay")
			wndControl:Enable((self.bEnsureCombatNameplatesOnScreen or self.bEnsureTargetNameplateOnScreen) and self.nEnsureMode == 1)
		elseif strSettingName == "nEnsureMode" then
			self.nEnsureMode = wndControl:IsChecked() and 1 or 0
			wndControl = self.wndSettingsStandard:FindChild("EnhancementEnsureModeBottomResetDelay")
			wndControl:Enable(self.nEnsureMode == 1)
		elseif strSettingName == "bShowOnMouseOver" then
			self.bShowOnMouseOver = wndControl:IsChecked()
		elseif strSettingName == "bShowClusterMarker" then
			self.bShowClusterMarker = wndControl:IsChecked()
		elseif strSettingName == "bHidePetsHostile" then
			self.bHidePetsHostile = wndControl:IsChecked()
		elseif strSettingName == "bHidePetsFriendly" then
			self.bHidePetsFriendly = wndControl:IsChecked()
		elseif strSettingName == "bHidePetsNeutral" then
			self.bHidePetsNeutral = wndControl:IsChecked()
		elseif strSettingName == "bHideNonPlayerHostile" then
			self.bHideNonPlayerHostile = wndControl:IsChecked()
		elseif strSettingName == "bHideNonPlayerFriendly" then
			self.bHideNonPlayerFriendly = wndControl:IsChecked()
		elseif strSettingName == "bHideNonPlayerNeutral" then
			self.bHideNonPlayerNeutral = wndControl:IsChecked()
		elseif strSettingName == "bWotWThrottle" then
			self.bWotWThrottle = wndControl:IsChecked()
			self:CheckWotW()
		elseif strSettingName == "bNameplateClickTargetsUnit" then
			self.bNameplateClickTargetsUnit = wndControl:IsChecked()
		elseif strSettingName == "bExtendCertainDeath" then
			self.bExtendCertainDeath = wndControl:IsChecked()
		elseif strSettingName == "bCleanseMarkerGroupOnly" then
			self.bCleanseMarkerGroupOnly = wndControl:IsChecked()
		elseif strSettingName == "bInterruptArmorMarkerUsePlayerClass" then
			self.bInterruptArmorMarkerUsePlayerClass = wndControl:IsChecked()
		elseif strSettingName == "bHideMyCastBar" then
			self.bHideMyCastBar = wndControl:IsChecked()
		elseif strSettingName == "bAnchorMyNameplateToFeet" then
			self.bAnchorMyNameplateToFeet = wndControl:IsChecked()	
			local tNameplate = aNameplates.arUnit2Nameplate[GameLib.GetPlayerUnit():GetId()]
			if tNameplate then
				if tNameplate.bMounted then
					tNameplate.wndNameplate:SetUnit(tNameplate.unitOwner:GetUnitMount(), (self.bAnchorMyNameplateToFeet and 0 or 1))
				else
					tNameplate.wndNameplate:SetUnit(tNameplate.unitOwner, (self.bAnchorMyNameplateToFeet and 0 or 1))
				end
			end
		elseif strSettingName == "bAdjustCertainDeath" then
			self.bAdjustCertainDeath = wndControl:IsChecked()
		elseif strSettingName == "bHideHostilePlayersOutOfPvP" then
			self.bHideHostilePlayersOutOfPvP = wndControl:IsChecked()
		elseif strSettingName == "bDisableOcclusionForTarget" then
			self.bDisableOcclusionForTarget = wndControl:IsChecked()
		elseif strSettingName == "bDisableOcclusionForGroupMembers" then
			self.bDisableOcclusionForGroupMembers = wndControl:IsChecked()
		elseif strSettingName == "bHideTitle" then
			self.bHideTitle = wndControl:IsChecked()	
			--self:UpdateExistingNameplatesToSettings(false, true)
			aNameplates:OnSettingNameChanged()
		elseif strSettingName == "bShowSmallClassIconLeftToName" then
			self.bShowSmallClassIconLeftToName = wndControl:IsChecked()	
			--self:UpdateExistingNameplatesToSettings(false, true)
			aNameplates:OnSettingNameChanged()	
		elseif strSettingName == "bHideMyNameWhenNotTargeted" then
			self.bHideMyNameWhenNotTargeted = wndControl:IsChecked()
			local tNameplate = aNameplates.arUnit2Nameplate[GameLib.GetPlayerUnit():GetId()]
			if tNameplate then
				SteadyPlates.DrawNameHook(aNameplates, tNameplate)
				SteadyPlates.DrawGuildHook(aNameplates, tNameplate)
			end
		elseif strSettingName == "bBubblesNeverHideNameplates" or strSettingName == "bBubblesAlwaysHideNameplates" then
			self.bBubblesAlwaysHideNameplates = self.wndSettingsStandard:FindChild("BubbleHideAlways"):IsChecked()
			self.bBubblesNeverHideNameplates = self.wndSettingsStandard:FindChild("BubbleHideNever"):IsChecked()
		elseif strSettingName == "bThreatMarkerLeft" then
			local bNew = self.wndSettingsMarkers:FindChild("ThreatMarkerLeft"):IsChecked()
			if self.bThreatMarkerLeft ~= bNew then
				self.bThreatMarkerLeft = bNew
				self.bThreatMarkerPositionChanged = true
				self:UpdateExistingNameplatesToSettings()
				self.bThreatMarkerPositionChanged = false
			end
		elseif strSettingName == "bThreatMarkerTankMode" then
			self.bThreatMarkerTankMode = wndControl:IsChecked()
		elseif strSettingName == "bShowThreatMarkerGroupOnly" then
			self.bShowThreatMarkerGroupOnly = wndControl:IsChecked()
		elseif strSettingName == "bShowThreatMarkerInstanceOnly" then
			self.bShowThreatMarkerInstanceOnly = wndControl:IsChecked()
		elseif strSettingName == "bReplaceHealthText" then
			self.bReplaceHealthText = wndControl:IsChecked()
			local wnd = self.wndSettingsStandard:FindChild("HealthBarTextCustomHideShield")
			wnd:Enable(self.bReplaceHealthText)
			wnd = self.wndSettingsStandard:FindChild("HealthBarTextCustomPercentages")
			wnd:Enable(self.bReplaceHealthText)
			wnd = self.wndSettingsStandard:FindChild("HealthBarTextCustomFriendlyDeficit")
			wnd:Enable(self.bReplaceHealthText)
			wnd = self.wndSettingsStandard:FindChild("HealthBarTextCustomDefault")
			wnd:Enable(self.bReplaceHealthText)
			wnd = self.wndSettingsStandard:FindChild("HealthBarTextCustomFormat")
			wnd:Enable(self.bReplaceHealthText)
			wnd = self.wndSettingsStandard:FindChild("HealthBarTextCustomFormatEdit")
			wnd:Enable(self.bReplaceHealthText)
			wnd = self.wndSettingsStandard:FindChild("HealthBarTextCustomFormatHelp")
			wnd:Enable(self.bReplaceHealthText)
			if not self.bReplaceHealthText then
				wnd = self.wndSettingsStandard:FindChild("HealthBarTextCustomFormatHelpFlyout")
				wnd:Show(false)
			end
			self.bHealthShieldBarOverride = self:ShouldOverrideHealthShieldBar()
			self:UpdateExistingNameplatesToSettings(false, true)
		elseif strSettingName == "bReplaceHealthTextFriendlyDeficit" then
			self.bReplaceHealthTextFriendlyDeficit = wndControl:IsChecked()
			self:UpdateExistingNameplatesToSettings(false, true)
		elseif strSettingName == "bReplaceHealthTextPercentages" then
			self.bReplaceHealthTextPercentages = wndControl:IsChecked()
			self:UpdateExistingNameplatesToSettings(false, true)
		elseif strSettingName == "bReplaceHealthTextHideShield" then
			self.bReplaceHealthTextHideShield = wndControl:IsChecked()
			self:UpdateExistingNameplatesToSettings(false, true)
		elseif strSettingName == "bHideShieldAbsorbLoss" then
			self.bHideShieldAbsorbLoss = wndControl:IsChecked()
			self.bHealthShieldBarOverride = self:ShouldOverrideHealthShieldBar()
		elseif strSettingName == "bHideShieldProgressEdgeGlow" then
			self.bHideShieldProgressEdgeGlow = wndControl:IsChecked()
			self:UpdateExistingNameplatesToSettings()
			self.bHealthShieldBarOverride = self:ShouldOverrideHealthShieldBar()
		elseif strSettingName == "bHealthBarShieldAbove" then
			self.bHealthBarShieldAbove = wndControl:IsChecked()
			self:UpdateExistingNameplatesToSettings()
			self.bHealthShieldBarOverride = self:ShouldOverrideHealthShieldBar()
		elseif strSettingName == "bDimmHealthInCombatFriendly" then
			self.bDimmHealthInCombatFriendly = wndControl:IsChecked()
			self:UpdateExistingNameplatesToSettings(false, true)
		elseif strSettingName == "bRemoveHealthBarBackground" then
			self.bRemoveHealthBarBackground = wndControl:IsChecked()
			self:UpdateExistingNameplatesToSettings()
		elseif strSettingName == "bUseCustomHealthTextFormat1" then
			self.bUseCustomHealthTextFormat = wndControl:IsChecked()
			self:UpdateExistingNameplatesToSettings(false, true)
		elseif strSettingName == "bUseCustomHealthTextFormat2" then
			self.bUseCustomHealthTextFormat = not wndControl:IsChecked()
			self:UpdateExistingNameplatesToSettings(false, true)
		elseif strSettingName == "nHealthBarTextFriendly1" then
			self.nHealthBarTextFriendly = 1
		elseif strSettingName == "nHealthBarTextFriendly2" then
			self.nHealthBarTextFriendly = 2
		elseif strSettingName == "nHealthBarTextFriendly3" then
			self.nHealthBarTextFriendly = 3
		elseif strSettingName == "nHealthBarTextFriendly4" then
			self.nHealthBarTextFriendly = 4
		elseif strSettingName == "nHealthBarTextFriendly0" then
			self.nHealthBarTextFriendly = 0
		elseif strSettingName == "nHealthBarTextHostile1" then
			self.nHealthBarTextHostile = 1
		elseif strSettingName == "nHealthBarTextHostile2" then
			self.nHealthBarTextHostile = 2
		elseif strSettingName == "nHealthBarTextHostile3" then
			self.nHealthBarTextHostile = 3
		elseif strSettingName == "nHealthBarTextHostile4" then
			self.nHealthBarTextHostile = 4
		elseif strSettingName == "nHealthBarTextHostile0" then
			self.nHealthBarTextHostile = 0
		elseif strSettingName == "nHealthBarTextNeutral1" then
			self.nHealthBarTextNeutral = 1
		elseif strSettingName == "nHealthBarTextNeutral2" then
			self.nHealthBarTextNeutral = 2
		elseif strSettingName == "nHealthBarTextNeutral3" then
			self.nHealthBarTextNeutral = 3
		elseif strSettingName == "nHealthBarTextNeutral4" then
			self.nHealthBarTextNeutral = 4
		elseif strSettingName == "nHealthBarTextNeutral0" then
			self.nHealthBarTextNeutral = 0
		elseif strSettingName == "bShowCastBarTargetOnlyFriendly" then
			self.bShowCastBarTargetOnlyFriendly = wndControl:IsChecked()
		elseif strSettingName == "bShowCastBarTargetOnlyNeutral" then
			self.bShowCastBarTargetOnlyNeutral = wndControl:IsChecked()
		elseif strSettingName == "bShowCastBarTargetOnlyHostile" then
			self.bShowCastBarTargetOnlyHostile = wndControl:IsChecked()
		elseif strSettingName == "bShowCastBarNameFriendly" then
			self.bShowCastBarNameFriendly = wndControl:IsChecked()
		elseif strSettingName == "bShowCastBarNameNeutral" then
			self.bShowCastBarNameNeutral = wndControl:IsChecked()
		elseif strSettingName == "bShowCastBarNameHostile" then
			self.bShowCastBarNameHostile = wndControl:IsChecked()
		elseif strSettingName == "bShowHealthBarTargetOnlyFriendly" then
			self.bShowHealthBarTargetOnlyFriendly = wndControl:IsChecked()
		elseif strSettingName == "bShowHealthBarTargetOnlyNeutral" then
			self.bShowHealthBarTargetOnlyNeutral = wndControl:IsChecked()
		elseif strSettingName == "bShowHealthBarTargetOnlyHostile" then
			self.bShowHealthBarTargetOnlyHostile = wndControl:IsChecked()
		elseif strSettingName == "strInterruptArmorMarkerSprite" then
			local strSprite = wndControl:FindChild("Icon"):GetSprite()
			if strSprite and string.len(strSprite) > 0 then
				self.strInterruptArmorMarkerSprite = strSprite
				self.bShowInterruptArmorMarker = true
			else
				self.bShowInterruptArmorMarker = false
			end
			wndControl = self.wndSettingsMarkers:FindChild("InterruptArmorMarkerUsePlayerClass")
			wndControl:Enable(self.bShowInterruptArmorMarker)
			self:UpdateExistingNameplatesToSettings()
		elseif strSettingName == "strThreatMarkerSprite" then
			local strSprite = wndControl:FindChild("Icon"):GetSprite()
			if strSprite and string.len(strSprite) > 0 then
				self.strThreatMarkerSprite = strSprite
				self.bShowThreatMarker = true
			else
				self.bShowThreatMarker = false
			end
			wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerTop")
			wndControl:Enable(self.bShowThreatMarker)
			wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerLeft")
			wndControl:Enable(self.bShowThreatMarker)
			wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerTankMode")
			wndControl:Enable(self.bShowThreatMarker)
			wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerGroupOnly")
			wndControl:Enable(self.bShowThreatMarker)
			wndControl = self.wndSettingsMarkers:FindChild("ThreatMarkerInstanceOnly")
			wndControl:Enable(self.bShowThreatMarker)
			self:UpdateExistingNameplatesToSettings()
		elseif strSettingName == "strCleanseMarkerSprite" then
			local strSprite = wndControl:FindChild("Icon"):GetSprite()
			if strSprite and string.len(strSprite) > 0 then
				self.strCleanseMarkerSprite = strSprite
				self.bShowCleanseMarker = true
			else
				self.bShowCleanseMarker = false
			end
			wndControl = self.wndSettingsMarkers:FindChild("CleanseMarkerGroupOnly")
			wndControl:Enable(self.bShowCleanseMarker)
			self:UpdateExistingNameplatesToSettings()
		elseif strSettingName == "nTargetMarkerArrowModifier" then
			local strName = wndControl:GetName()
			if strName == "TargetMarker1" then
				self.nTargetMarkerArrowModifier = 0
				self.bShowTargetMarkerArrow = true
			elseif strName == "TargetMarker2" then
				self.nTargetMarkerArrowModifier = 10
				self.bShowTargetMarkerArrow = true
			elseif strName == "TargetMarker3" then
				self.nTargetMarkerArrowModifier = 20
				self.bShowTargetMarkerArrow = true
			elseif strName == "TargetMarker4" then
				self.nTargetMarkerArrowModifier = 30
				self.bShowTargetMarkerArrow = true
			else
				self.bShowTargetMarkerArrow = false
			end
			self:UpdateExistingNameplatesToSettings()
		elseif strSettingName == "bOverwriteHealthBarColor" then
			self.bOverwriteHealthBarColor = wndControl:IsChecked()
			self:UpdateExistingNameplatesToSettings()
		elseif strSettingName == "bOverwriteNameColor" then
			self.bOverwriteNameColor = wndControl:IsChecked()
			self:UpdateExistingNameplatesToSettings()
		elseif strSettingName == "bOverwriteShieldBarColor" then
			self.bOverwriteShieldBarColor = wndControl:IsChecked()
			self:UpdateExistingNameplatesToSettings()
			self:LoadColorSets()
		elseif strSettingName == "bOverwriteAbsorbBarColor" then
			self.bOverwriteAbsorbBarColor = wndControl:IsChecked()
			self:UpdateExistingNameplatesToSettings()
			self:LoadColorSets()
		elseif strSettingName == "bOverwriteColorPlayerPvPSpecific" then
			self.bOverwriteColorPlayerPvPSpecific = wndControl:IsChecked()
			self:LoadColorSets()
		elseif strSettingName == "bOverwriteColorPlayerClassSpecific" then
			self.bOverwriteColorPlayerClassSpecific = wndControl:IsChecked()
			self:LoadColorSets()
		end
	end
end

function SteadyPlates:OnSettingsDefaultNameplateSettings()
	aNameplates:OnNameplatesOn()
	aNameplates.wndMain:ToFront()
end

function SteadyPlates:OnSettingsOk()
	self:HideSettingsWindow()
end

function SteadyPlates:OnSettingsCancel()
	self:HideSettingsWindow()
end

function SteadyPlates:OnSettingsDefaults(wndHandler, wndControl, eMouseButton)
	wndHandler:FindChild("ConfirmFlyOut"):Show(true)
end

function SteadyPlates:OnSettingsDefaultsYes(wndHandler, wndControl, eMouseButton)
	wndHandler:GetParent():Show(false)
	self:LoadDefaultSettings()
	self:ShowSettingsWindow()
end

function SteadyPlates:OnSettingsDefaultsNo(wndHandler, wndControl, eMouseButton)
	wndHandler:GetParent():Show(false)
end

function SteadyPlates:OnSettingsColorSetDefault(wndHandler, wndControl, eMouseButton)
	wndHandler:FindChild("ConfirmFlyOut"):Show(true)
end

function SteadyPlates:OnSettingsColorSetDefaultYes(wndHandler, wndControl, eMouseButton)
	wndHandler:GetParent():Show(false)
	self:LoadDefaultColorSets()
	self:LoadColorSets()
end

function SteadyPlates:OnSettingsColorSetDefaultNo(wndHandler, wndControl, eMouseButton)
	wndHandler:GetParent():Show(false)
end

function SteadyPlates:OnSettingsColorSetContextMenuButton(wndHandler, wndControl, eMouseButton)
	local wndContextMenu = wndHandler:FindChild("ContextMenuFlyOut")
	if not wndContextMenu:IsShown() then
		wndContextMenu:Show(true)
	end
end

function SteadyPlates:OnSettingsColorSetContextMenuElementButton(wndHandler, wndControl, eMouseButton)
	if wndHandler:GetName() == "CancelButton" then
		wndHandler:GetParent():Show(false)
	else
		wndHandler:GetParent():GetParent():Show(false)
		tData = wndHandler:GetData()
		self:HelperCopyColorSet(tData.Source, tData.Target)
	end
end

function SteadyPlates:HelperCopyColorSet(wndSetFrom, wndSetTo)
	local tDataFrom = nil
	if wndSetFrom.GetData then
		tDataFrom = wndSetFrom:GetData()
	else
		tDataFrom = wndSetFrom
	end
	local tDataTo = wndSetTo:GetData()
	if tDataTo.Type == 1 then
		ImportColorSetDispositions(tDataFrom.ColorSet, tDataTo.ColorSet)
	elseif tDataTo.Type == 2 then
		ImportColorSetPlayer(tDataFrom.ColorSet, tDataTo.ColorSet)
	elseif tDataTo.Type == 0 then
		ImportColorSetSelf(tDataFrom.ColorSet, tDataTo.ColorSet)
	elseif tDataTo.Type == 3 then
		ImportColorSetCommon(tDataFrom.ColorSet, tDataTo.ColorSet)
	end
	self.strLastOpenColorSet = tDataTo.Name
	self:LoadColorSets()
end

function SteadyPlates:OnSettingsColorSetShare(wndHandler, wndControl, eMouseButton)
	if not self.wndColorSetShare then
		self.wndColorSetShare = Apollo.LoadForm(self.xmlSettingsDoc, "ColorSetsExportImportForm", self.wndSettings, self)
	end
	local strExport = aLibJSON.encode(self.tColorSets)
	self.wndColorSetShare:FindChild("EditExport:EditBox"):SetText(strExport)
	self.wndColorSetShare:FindChild("ExportButton"):SetActionData(GameLib.CodeEnumConfirmButtonType.CopyToClipboard, strExport)
	self.wndColorSetShare:Show(true)
end

function SteadyPlates:OnSettingsColorSetShareImport(wndHandler, wndControl, eMouseButton)
	local wndEdit = wndHandler:GetParent():FindChild("EditImport:EditBox")
	local wndStatus = wndHandler:GetParent():FindChild("StatusLabel")
	local strImport = wndEdit:GetText()
	local tColorSets, nPos, strError = aLibJSON.decode(strImport, 1, nil, nil)
	local bSuccess = false
	
	if tColorSets ~= nil and strError == nil then
		local tColorSetsNew
		--bSuccess, tColorSetsNew = pcall(self.ImportColorSets, self, tColorSets)
		bSuccess, tColorSetsNew = true, self.ImportColorSets(self, tColorSets)
		if bSuccess then
			self.tColorSets = tColorSetsNew
		end
	end
	
	if bSuccess then
		self:LoadColorSets()
		wndStatus:SetText("Success!")
		wndStatus:SetTextColor("green")
		wndStatus:SetTooltip("")
	else
		wndStatus:SetText("Failed!")
		wndStatus:SetTextColor("red")
		wndStatus:SetTooltip(strError or "")
	end
end

function SteadyPlates:OnSettingsColorSetShareClose(wndHandler, wndControl, eMouseButton)
	self.wndColorSetShare:Show(false)
end

function SteadyPlates:OnSettingsMouseButtonDown(wndHandler, wndControl, eMouseButton)
end

function SteadyPlates.UpdateNameplateToSettings(tNameplate, wnds, bReset)
	if bReset then
		if tNameplate then
			local tCluster = tNameplate.unitOwner:GetClusterUnits()
			if tCluster ~= nil then
				tNameplate.nClusterUnits = #tCluster
			else
				tNameplate.nClusterUnits = 0
			end
			tNameplate.bIsCasting = false
		end
		local wnd = wnds.targetMarkerArrow
		wnd:RemoveStyle("TransitionShowHide") -- against possible flashing bug
		--wnd:SetAnchorOffsets()
		wnd = wnds.castBarLabel
		wnd:SetStyle("NoClip", true)
		--wnd:SetStyle("AutoScaleTextOff", true)
		wnd:SetAnchorOffsets(-30, 4, 30, 12)
		wnd:SetFont("Nameplates")
		wnds.castBar:Show(false)
		wnds.vulnerable:Show(false)
		wnds.health:Show(false)
		wnd = wnds.cleanseMarker
		if wnd then
			wnd:Show(false)
		end
		wnd = wnds.threatMarker
		if wnd then
			wnd:Show(false)
		end
	end
	if tNameplate then
		tNameplate.bEnsured = false
		tNameplate.bEnsureHide = false
		local unitOwner = tNameplate.unitOwner
		local unitType = unitOwner:GetType()
		tNameplate.bIsPlayer = unitType == "Player"
		tNameplate.bIsPet = unitType == "Pet" or unitType == "Esper Pet"
		tNameplate.bIsSimple = unitType == "Simple"
		tNameplate.bIsNPC = unitType == "NonPlayer"
		tNameplate.bIsScanner = unitType == "Scanner"
		local unitController = unitOwner:GetUnitOwner() or unitOwner
		tNameplate.bIsGroupMember = unitController:IsInYourGroup()
		tNameplate.bIsPvpFlagged = unitController:IsPvpFlagged()
		local eDisposition = tNameplate.eDisposition
		tNameplate.bIsFriendly = eDisposition == Unit.CodeEnumDisposition.Friendly
		tNameplate.bIsNeutral = eDisposition == Unit.CodeEnumDisposition.Neutral
		tNameplate.bIsHostile = eDisposition == Unit.CodeEnumDisposition.Hostile
		--local nFaction = unitOwner:GetFaction() -- 166 = dominion, 167 = exile; Unit.CodeEnumFaction.ExilesPlayer, Unit.CodeEnumFaction.DominionPlayer
		--tNameplate.nFaction = nFaction or 0
		SteadyPlates.DrawNameHook(aNameplates, tNameplate)
		SteadyPlates.DrawGuildHook(aNameplates, tNameplate)
		--aNameplates:DrawLevel(tNameplate)
		--aNameplates:UpdateNameplateRewardInfo(tNameplate)
	end
	local wnd = wnds.threatMarker
	if wnd then
		if SteadyPlatesInst.bThreatMarkerPositionChanged then
			wnd:Destroy() -- need to destroy it, so that it gets re-created when to show (because due to changed settings another parent has to be used)
			wnds.threatMarker = nil
		else
			wnd:SetSprite(SteadyPlatesInst.strThreatMarkerSprite)
			local nLeft, nTop, nRight, nBottom
			if SteadyPlatesInst.bThreatMarkerLeft then
				local offsets = karThreatMarkerOffsets.Left
				--if tNameplate and tNameplate.bIsTarget and not SteadyPlatesInst.bRemoveHealthBarBackground then
				if tNameplate and tNameplate.bIsTarget then
					nLeft, nTop, nRight, nBottom = offsets.nLeft - knMarkerSideOffset, offsets.nTop, offsets.nRight - knMarkerSideOffset, offsets.nBottom
				else
					nLeft, nTop, nRight, nBottom = offsets.nLeft, offsets.nTop, offsets.nRight, offsets.nBottom
				end
			else
				local offsets = karThreatMarkerOffsets.Top
				nLeft, nTop, nRight, nBottom = offsets.nLeft, offsets.nTop, offsets.nRight, offsets.nBottom
			end
			wnd:SetAnchorOffsets(nLeft, nTop, nRight, nBottom)
		end
	end
	wnd = wnds.cleanseMarker
	if wnd then
		wnd:SetSprite(SteadyPlatesInst.strCleanseMarkerSprite)
	end
	if tNameplate then
		tNameplate.strHealthBarColor = nil
		tNameplate.strShieldBarColor = nil
		tNameplate.strAbsorbBarColor = nil
		tNameplate.nLastHealthCurr = nil
		tNameplate.nLastAbsorbShieldCurr = nil
		tNameplate.nLastHealthShieldTotalMax = nil
		local bIsMe = tNameplate.unitOwner:IsThePlayer()
		tNameplate.bIsMe = bIsMe
		if bIsMe and SteadyPlatesInst.bAnchorMyNameplateToFeet then
			if tNameplate.bMounted then
				tNameplate.wndNameplate:SetUnit(tNameplate.unitOwner:GetUnitMount(), 0)
			else
				tNameplate.wndNameplate:SetUnit(tNameplate.unitOwner, 0)
			end
		end
	end
	if SteadyPlatesInst.bAdjustCertainDeath then
		if tNameplate and tNameplate.bIsTarget then
			wnds.certainDeath:SetAnchorOffsets(-18, -50, 18, -14)
		else
			wnds.certainDeath:SetAnchorOffsets(-18, -36, 18, 0)
		end
	end
	local wndHealth = wnds.health
	if SteadyPlatesInst.bNameplateClickTargetsUnit then
		--wndHealth:AddEventHandler("MouseButtonDown", "OnNameplateNameClick")
		wndHealth:AddEventHandler("MouseButtonDown", "SteadyPlates_OnNameplateClick") 
		aNameplates.SteadyPlates_OnNameplateClick = SteadyPlates.OnNameplateClick
		aNameplates.OnNameplateNameClick = SteadyPlates.OnNameplateClick
		--wndHealth:SetStyle("Overlapped", true)
		--wnds.wndName:SetStyle("Overlapped", true)
		--wnds.wndGuild:SetStyle("Overlapped", true)
	elseif aNameplates.SteadyPlates_OnNameplateClick then
		wndHealth:RemoveEventHandler("MouseButtonDown") 
		aNameplates.SteadyPlates_OnNameplateClick = nil
		aNameplates.OnNameplateNameClick = nil
	end
	wnd = wnds.backgroundPlate
	if wnd then
		wnd:Show(not SteadyPlatesInst.bRemoveHealthBarBackground)
	end
	local wnd = wnds.interruptArmorMarker
	if wnd then
		local strSprite = SteadyPlatesInst.strInterruptArmorMarkerSprite
		if tNameplate and SteadyPlatesInst.bInterruptArmorMarkerUsePlayerClass and tNameplate.bIsPlayer and not tNameplate.bIsMe then
			local nClass = tNameplate.unitOwner:GetClassId() or 0
			strSprite = karClassIcons[nClass]
			wnd:Show(not wnd:IsShown())
			wnd:SetText("")
		else
			wnd:Show(false)
		end
		wnd:SetSprite(strSprite)
	end
end

function SteadyPlates:UpdateExistingNameplatesToSettings(bReset, bIgnorePool)
	for _, tNameplate in pairs(aNameplates.arUnit2Nameplate) do
		self.UpdateNameplateToSettings(tNameplate, tNameplate.wnd, bReset)
	end
	if not bIgnorePool then
		for _, poolEntry in pairs(aNameplates.arWindowPool) do
			local wnds = poolEntry[2]
			if wnds then
				self.UpdateNameplateToSettings(nil, wnds, bReset)
			end
		end
	end
end

-----------------------------------------------------------------------------------------------
-- SteadyPlates Instance
-----------------------------------------------------------------------------------------------
SteadyPlatesInst = SteadyPlates:new()
SteadyPlatesInst:Init()

